;(function ($) {
    /**
     * Manages magic export/import functionality from the admin area.
     */
    var magicExportImport = {
        $magicItemSelect: $('#magic-ex-im-post-type-select'),
        $exportItemsSelect: $('#magic-ex-items-select'),
        exportKeysSelect: '#magic-ex-keys-select',
        exportMediaKeysSelect: '#magic-ex-media-keys-select',
        $selectGroupKeysCheckboxes: $('.magic-ex-group'),
        $selectAllKeysCheckbox: $('#magic-ex-all-keys-checkbox'),
        $exportForm: $('#magic-ex-form'),
        $exportItemsCnt: $('#magic-ex-items-cnt'),
        $exFiltersTogglerCheckbox: $('#magic-ex-advanced-checkbox'),
        $exFilters: $('#magic-ex-advanced-filters'),
        $exApplyFiltersBtn: $('#magic-ex-apply-filters'),
        exAppliedFiltersData: '',
        $importTestModeCheckbox: $('#magic-im-test-mode'),
        $importForm: $('#magic-im-form'),
        $importProgress: $('#magic-im-progress'),
        $importNotices: $('#magic-im-notices'),
        importXHR: null,
        select2IsUnselecting: false,
        timer: 0,

        /**
         * Initializes export and import functionality.
         */
        init: function () {
            // Remove advanced filters toggler if there are no advanced filters.
            if (!this.$exFilters.text().trim()) {
                this.$exFiltersTogglerCheckbox.closest('.form-field').remove()
            }

            let $exportFileLink = $('#magic-ex-generated-file')

            // Download newly generated export file if exists, compatible with Safari.
            if ($exportFileLink.length) {
                setTimeout(() => {
                    let fileURL = $exportFileLink.attr('href'),
                        link = document.createElement('a')

                    link.href = fileURL
                    link.download = ''

                    document.body.appendChild(link)
                    link.click()
                    document.body.removeChild(link)
                }, 500)
            }

            // Init magic item select.
            this.$magicItemSelect.select2({
                escapeMarkup: (markup) => {
                    return markup
                },
                templateResult: (tag) => {
                    return this.select2MagicItemTemplate(tag)
                },
                templateSelection: (tag) => {
                    return this.select2MagicItemTemplate(tag)
                },
            })

            // Init export individual items select.
            this.$exportItemsSelect.select2({
                placeholder: 'Select items to export',
                minimumInputLength: 0,
                closeOnSelect: false,
                width: '100%',
                escapeMarkup: (markup) => {
                    return markup
                },
                templateSelection: (tag, container) => {
                    return tag.text.split('<code>')[0]
                },
                ajax: {
                    type: 'POST',
                    url: WPLocalize.admin_ajax_url,
                    delay: 400,
                    dataType: 'json',
                    data: function (params) {
                        params.action = `get_individual_export_${WPLocalize.magic_type}`
                        params.magic_item = WPLocalize.magic_item
                        return params
                    },
                },
            })

            // Disable advanced filters.
            $('.magic-ex-advanced-filter-type')
                .next('div')
                .find(':input')
                .prop('disabled', true)

            this.initExportKeysSelects()
            this.events()
        },

        /**
         * Event handlers.
         */
        events: function () {
            // Filter by taxonomy checkbox on change.
            $('[name="magic-ex-by-taxonomy"]').on('change', function () {
                $('.magic-ex-taxes-wrapper').toggle()
            })

            // Advanced filters on input.
            this.$exFilters.on('input', ':input', (e) => {
                let $target = $(e.target)

                // Set filters state if related 'export by' checkbox was changed.
                if ($target.is('[name^="magic-ex-by-"]')) {
                    let isChecked = $target.is(':checked'),
                        $relatedFilters = $target
                            .closest('.magic-ex-advanced-filter-type')
                            .next('div')
                            .find(':input:not(.disabled)')

                    $relatedFilters.prop('disabled', !isChecked)
                }

                this.setApplyFiltersBtn()
            })

            // Individual export items select on change.
            this.$exportItemsSelect.on('change', (e) => {
                this.refreshExportKeys()
            })

            // Apply filters button on click.
            this.$exApplyFiltersBtn.on('click', (e) => {
                e.preventDefault()
                this.refreshExportKeys()
            })

            // SWIFT togggler on click.
            $('[href="#swift-details-wrapper"]').on('click', (e) => {
                $('#swift-details-wrapper').toggle()
            })

            // Magic item select on change - init new magic item with page relaod.
            this.$magicItemSelect.on('change', (e) => {
                e.preventDefault()
                this.$exportForm.add(this.$importForm).addClass('js-processing')
                let url = new URL(window.location)
                url.searchParams.set('magic_item', this.$magicItemSelect.val())
                window.history.pushState({}, null, url.toString())
                window.location.reload()
            })

            // Export form on submit.
            this.$exportForm.on('submit', () => {
                this.$exportForm.addClass('js-processing')
            })

            // Stop import button on click.
            $('body').on('click', '#magic-im-stop', (e) => {
                e.preventDefault()
                if (this.importXHR) {
                    this.importXHR.abort()
                }
            })

            // Import test mode checkbox on change - disable download media field.
            this.$importTestModeCheckbox.on('change', () => {
                $('#magic-im-download-media').prop(
                    'disabled',
                    this.$importTestModeCheckbox.is(':checked')
                )
            })

            // Import form on submit.
            this.$importForm.on('submit', (e) => {
                e.preventDefault()
                this.$importProgress.html('')
                this.$importNotices.html('')
                this.toggleImportFormState()
                this.sendImportAjax()
            })

            // Select2 option on unselecting.
            $('body').on('select2:unselecting', 'select', (e) => {
                let $option = $(e.params.args.data.element)
                if ($option.hasClass('mandatory')) {
                    e.preventDefault()
                } else {
                    this.select2IsUnselecting = true
                }
            })

            // Prevent opening Select2 dropdown on unselecting its values (clicking on cross).
            // Explanation: https://forums.select2.org/t/how-to-not-open-dropdown-when-deselecting-option-in-multiselect/1132/2.
            $('body').on('select2:opening', 'select', (e) => {
                if (this.select2IsUnselecting) {
                    this.select2IsUnselecting = false
                    return false
                }
            })

            // Export keys select on change - check selected keys to set related checkboxes UI.
            $('body').on('change', this.exportKeysSelect, (e) => {
                // Check if 'Select all' checkbox should be checked.
                this.$selectAllKeysCheckbox.prop(
                    'checked',
                    $(this.exportKeysSelect).find('option:not(:selected)')
                        .length == 0
                )

                // Check if 'Select by group' checkboxes should be checked.
                this.$selectGroupKeysCheckboxes
                    .filter(':not(:disabled)')
                    .each((index, element) => {
                        let $groupCheckbox = $(element),
                            group = $groupCheckbox.val()

                        $groupCheckbox.prop(
                            'checked',
                            $(this.exportKeysSelect).find(
                                `option[data-group="${group}"]:not(:selected)`
                            ).length == 0
                        )
                    })
            })

            // 'Select by group' export keys checkboxes on change.
            this.$selectGroupKeysCheckboxes.on('change', (e) => {
                let $target = $(e.target),
                    group = $target.val()

                $(this.exportKeysSelect)
                    .find(`option[data-group=${group}]:not(.mandatory)`)
                    .prop('selected', $target.is(':checked'))

                $(this.exportKeysSelect).trigger('change')
            })

            // 'Select all' export keys checkbox on change.
            this.$selectAllKeysCheckbox.on('change', (e) => {
                let isChecked = this.$selectAllKeysCheckbox.is(':checked')

                $(this.exportKeysSelect)
                    .find('option:not(.mandatory)')
                    .prop('selected', isChecked)

                this.$selectGroupKeysCheckboxes
                    .filter(':not(:disabled)')
                    .prop('checked', isChecked)

                $(this.exportKeysSelect).trigger('change')
            })

            // Advanced export filters checkbox on change.
            this.$exFiltersTogglerCheckbox.on('change', () => {
                $('#magic-ex-advanced-wrapper').toggleClass('js-active')
                this.exAppliedFiltersData
                    ? this.refreshExportKeys()
                    : this.toggleExportFormState()
            })
        },

        /**
         * Sends AJAX request to import items, updating the progress.
         *
         * @param {Object} [progress={}] - Progress data to be included back in the request to recurse.
         */
        sendImportAjax: function (progress = {}) {
            let ajaxData = new FormData(this.$importForm[0])
            ajaxData.append(
                'action',
                `process_import_${WPLocalize.magic_type}_form`
            )
            ajaxData.append('progress', JSON.stringify(progress))

            this.importXHR = $.ajax({
                url: WPLocalize.admin_ajax_url,
                type: 'POST',
                processData: false,
                contentType: false,
                data: ajaxData,
                success: (response) => {
                    response = JSON.parse(response)

                    this.$importProgress.html(response.progress_html)
                    this.$importNotices.append(response.notices_html)

                    if (response.error) {
                        alert(response.error)
                        this.toggleImportFormState()
                        return
                    }

                    response.finish
                        ? this.toggleImportFormState()
                        : this.sendImportAjax(response)
                },
                error: (xhr, status, error) => {
                    this.toggleImportFormState()

                    if ('error' === status && 0 === xhr.status) {
                        let possibleError =
                            'File upload failed. Possible file change during upload.'
                        alert(possibleError)
                    } else if ('abort' !== status) {
                        alert(`${xhr.status} ${status} ${error}`)
                    }
                },
            })
        },

        /**
         * Inits export keys selects.
         */
        initExportKeysSelects: function () {
            // Init export keys select.
            $(this.exportKeysSelect).select2(
                this.select2ExportKeysParams($(this.exportKeysSelect))
            )

            // Init export media keys select.
            $(this.exportMediaKeysSelect).select2(
                this.select2ExportKeysParams($(this.exportMediaKeysSelect))
            )
        },

        /**
         * Refreshes export keys based on current form state.
         */
        refreshExportKeys: function () {
            let ajaxData = new FormData(this.$exportForm[0]),
                action = `refresh_export_${WPLocalize.magic_type}_keys`

            // Do not include export keys in the request, as we are refreshing them.
            ajaxData.delete($(this.exportKeysSelect).attr('name'))
            ajaxData.delete($(this.exportMediaKeysSelect).attr('name'))

            ajaxData.append('action', action)
            this.$exportForm.addClass('js-refreshing')

            // Send Ajax
            $.ajax({
                url: WPLocalize.admin_ajax_url,
                type: 'POST',
                processData: false,
                contentType: false,
                data: ajaxData,
                success: (response) => {
                    response = JSON.parse(response)
                    let exportKeysHTML = response.export_keys_html,
                        exportMediaKeysHTML = response.export_media_keys_html,
                        exportItemsCnt = response.export_items_cnt

                    this.$exportItemsCnt
                        .text(exportItemsCnt)
                        .addClass('flash')
                        .one('animationend', function () {
                            $(this).removeClass('flash')
                        })

                    $(this.exportKeysSelect)
                        .select2('destroy')
                        .replaceWith(exportKeysHTML)

                    this.$selectAllKeysCheckbox.prop('checked', true)

                    this.$selectGroupKeysCheckboxes.each((index, element) => {
                        let $groupCheckbox = $(element),
                            group = $groupCheckbox.val(),
                            isDisabled =
                                $(this.exportKeysSelect).find(
                                    `option[data-group="${group}"]`
                                ).length == 0

                        $groupCheckbox.prop('checked', !isDisabled)
                        $groupCheckbox.prop('disabled', isDisabled)
                    })

                    $(this.exportMediaKeysSelect)
                        .select2('destroy')
                        .replaceWith(exportMediaKeysHTML)

                    this.setAppliedFiltersData()
                    this.initExportKeysSelects()
                },
                error: (xhr, status, error) => {
                    alert(`${xhr.status} ${status} ${error}`)
                },
                complete: () => {
                    this.$exportForm.removeClass('js-refreshing')
                },
            })
        },

        /**
         * Processes template callbacks for Select2 magic item field.
         *
         * @param {Object} tag Select2 tag.
         * @returns {string} Select2 option label.
         */
        select2MagicItemTemplate: function (tag) {
            let $option = this.$magicItemSelect.find(
                    `option[value="${tag.id}"]`
                ),
                label = $option.text(),
                value = $option.attr('value')

            tag.text = `${label} <code>${value}</code>`
            return tag.text
        },

        /**
         * Processes template callbacks for Select2 export keys fields.
         *
         * @param {Object} tag Select2 tag.
         * @param {HTMLElement} container Select2 option DOM element.
         * @param {jQuery} $select Select field.
         * @returns {string} Select2 option label.
         */
        select2ExportKeysTemplate: function (tag, container, $select) {
            let $option = $select.find(`option[value="${tag.id}"]`),
                keyGroup = $option.attr('data-group'),
                key = $option.attr('data-key')

            if ($option.hasClass('mandatory')) {
                $(container).addClass('mandatory')
            }

            if (keyGroup && key) {
                tag.text = `${key} <code>${keyGroup}</code>`
            }

            return tag.text
        },

        /**
         * Gets parameters for the Select2 export keys fields.
         *
         * @param {jQuery} $select Select element.
         * @returns {Object} Select2 initialization parameters.
         */
        select2ExportKeysParams: function ($select) {
            let params = {
                closeOnSelect: false,
                width: '100%',
                escapeMarkup: (markup) => {
                    return markup
                },
                templateSelection: (tag, container) => {
                    return this.select2ExportKeysTemplate(
                        tag,
                        container,
                        $select
                    )
                },
                templateResult: (tag, container) => {
                    return this.select2ExportKeysTemplate(
                        tag,
                        container,
                        $select
                    )
                },
            }
            return params
        },

        /**
         * Toggles state of the export form.
         */
        toggleExportFormState: function () {
            let exportItemsCnt = parseInt(this.$exportItemsCnt.text())

            this.$exportForm
                .find('p.submit input')
                .prop(
                    'disabled',
                    exportItemsCnt < 1 ||
                        this.$exApplyFiltersBtn.is(':visible:not(:disabled)')
                )
        },

        /**
         * Toggles state of the import form.
         */
        toggleImportFormState: function () {
            this.$importForm.toggleClass('js-processing')
            let isProcessing = this.$importForm.hasClass('js-processing')

            if (!isProcessing) {
                $('#magic-im-stop').remove()
            }

            this.$importForm
                .find('p.submit input')
                .prop('disabled', isProcessing)
        },

        /**
         * Sets applied filters data.
         */
        setAppliedFiltersData: function () {
            this.exAppliedFiltersData = this.getFiltersData()
            this.setApplyFiltersBtn()
        },

        /**
         * Sets apply filters button state based on applied and actual filters data.
         */
        setApplyFiltersBtn: function () {
            let isDisabled = this.$exApplyFiltersBtn.prop('disabled'),
                shouldBeDisabled =
                    this.exAppliedFiltersData == this.getFiltersData()

            this.$exApplyFiltersBtn.prop('disabled', shouldBeDisabled)

            if (isDisabled && !shouldBeDisabled) {
                this.$exApplyFiltersBtn
                    .addClass('flash')
                    .one('animationend', function () {
                        $(this).removeClass('flash')
                    })
            }

            this.toggleExportFormState()
        },

        /**
         * Gets serialized advanced filters data.
         *
         * @returns {string}
         */
        getFiltersData: function () {
            return this.$exFilters
                .find(':input:not([name^="magic-ex-by-"])')
                .serialize()
        },
    }

    $(document).ready(function () {
        magicExportImport.init()
    })
})(jQuery)
