<?php
/**
 * Class Magic_EX_IM_Data.
 *
 * @package Magic_Export_Import
 */

defined( 'ABSPATH' ) || exit;

/**
 * Static class to manage common data during export or import.
 */
final class Magic_EX_IM_Data {

	/**
	 * Processing magic type.
	 *
	 * @var string
	 */
	private static $magic_type = '';

	/**
	 * Processing magic item.
	 *
	 * @var string
	 */
	private static $magic_item = '';

	/**
	 * Current process: export or import.
	 *
	 * @var string
	 */
	private static $current_process = 'export';

	/**
	 * User parameters.
	 *
	 * @var array
	 */
	private static $user_params = array();

	/**
	 * Import file parameters.
	 *
	 * @var array
	 */
	private static $import_file_params = array();

	/**
	 * Processing item file row.
	 *
	 * @var int
	 */
	private static $importing_item_file_row = 0;

	/**
	 * Importing item name.
	 *
	 * @var string
	 */
	private static $importing_item_name = '';

	/**
	 * The ID of the processing item.
	 *
	 * @var int
	 */
	private static $processing_item_id = 0;

	/**
	 * Determines if item was created during the import.
	 *
	 * @var bool
	 */
	private static $is_importing_item_new = false;

	/**
	 * Import notices.
	 *
	 * @var array
	 */
	private static $import_notices = array();

	/**
	 * Sets processing magic type and magic item.
	 *
	 * @param string $magic_type Magic type.
	 * @param string $magic_item Magic item.
	 */
	public static function set_magic( $magic_type, $magic_item ) {
		self::$magic_type = $magic_type;
		self::$magic_item = $magic_item;
	}

	/**
	 * Gets processing magic type.
	 */
	public static function get_magic_type() {
		return self::$magic_type;
	}

	/**
	 * Gets processing magic item.
	 */
	public static function get_magic_item() {
		return self::$magic_item;
	}

	/**
	 * Gets current process: export or import.
	 *
	 * @return string
	 */
	public static function get_current_process() {
		return self::$current_process;
	}

	/**
	 * Sets file row number of the processing item.
	 *
	 * @param int    $item_file_row First item starts from the row #2 since first row is filled with data keys.
	 * @param string $item_name Item name.
	 */
	public static function set_importing_item( $item_file_row, $item_name ) {
		self::$importing_item_file_row = $item_file_row;
		self::$importing_item_name     = $item_name;
	}

	/**
	 * Gets file row number of the processing item.
	 */
	public static function get_importing_item_file_row() {
		return self::$importing_item_file_row;
	}

	/**
	 * Gets importing item name.
	 */
	public static function get_importing_item_name() {
		return self::$importing_item_name;
	}

	/**
	 * Sets processing item data.
	 *
	 * @param array $item_id Item ID.
	 * @param bool  $is_new_item Whether item was created during the import.
	 */
	public static function set_processing_item( $item_id, $is_new_item = false ) {

		self::$processing_item_id = $item_id;

		if ( 'import' === self::get_current_process() ) {
			self::$is_importing_item_new = $is_new_item;
		}
	}

	/**
	 * Gets processing item ID.
	 */
	public static function get_processing_item_id() {
		return self::$processing_item_id;
	}

	/**
	 * Checks if item was created during the import.
	 */
	public static function is_importing_item_new() {
		return self::$is_importing_item_new;
	}

	/**
	 * Resets processing item data.
	 */
	public static function reset_processing_item() {

		self::set_processing_item( 0 );

		if ( 'import' === self::get_current_process() ) {
			self::set_importing_item( 0, '' );
		}
	}

	/**
	 * Sets user parameters.
	 *
	 * @param array  $params Parameters to set.
	 * @param string $current_process Export or import.
	 */
	public static function set_user_params( $params, $current_process = '' ) {

		if ( is_array( $params ) ) {
			self::$user_params = array_merge( self::$user_params, $params );
		}

		if ( 'export' === $current_process || 'import' === $current_process ) {
			self::$current_process = $current_process;
		}
	}

	/**
	 * Sets import file parameters.
	 *
	 * @param array $params Parameters to set.
	 */
	public static function set_import_file_params( $params ) {

		if ( is_array( $params ) ) {
			self::$import_file_params = array_merge( self::$import_file_params, $params );
		}
	}

	/**
	 * Gets import file parameters.
	 *
	 * @param string $param Parameter name to get instead of all parameters.
	 * @return mixed
	 */
	public static function get_import_file_params( $param = '' ) {
		return $param ? ( self::$import_file_params[ $param ] ?? null ) : self::$import_file_params;
	}

	/**
	 * Gets user parameters.
	 *
	 * @param string $param Parameter name to get instead of all parameters.
	 * @return mixed
	 */
	public static function get_user_params( $param = '' ) {
		return $param ? ( self::$user_params[ $param ] ?? null ) : self::$user_params;
	}

	/**
	 * Appends import notice to the notices array.
	 *
	 * @param string $notice Notice content.
	 */
	public static function add_import_notice( $notice ) {
		self::$import_notices[] = $notice;
	}

	/**
	 * Gets all import notices.
	 *
	 * @return array
	 */
	public static function get_import_notices() {
		return array_filter( self::$import_notices );
	}
}
