<?php
/**
 * Class Magic_EX_IM_Type.
 *
 * @package Magic_Export_Import
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 * phpcs:disable WordPress.Security.NonceVerification.Missing
 * phpcs:disable WordPress.WP.AlternativeFunctions
 * phpcs:disable WordPress.DB.SlowDBQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Abstract class to manage export and import functionality for a particular data type.
 */
abstract class Magic_EX_IM_Type {

	use Magic_EX_IM_Singleton;

	const FILE_UPLOAD_ERRORS = array(
		1 => 'The uploaded file exceeds the upload_max_filesize directive in php.ini',
		2 => 'The uploaded file exceeds the MAX_FILE_SIZE directive that was specified in the HTML form',
		3 => 'The uploaded file was only partially uploaded',
		4 => 'No file was uploaded',
		6 => 'Missing a temporary folder',
		7 => 'Failed to write file to disk.',
		8 => 'A PHP extension stopped the file upload.',
	);

	const FILE_PARAMS_KEY = 'magic_ex_im_file_params';
	const EX_NOTICE_KEY   = 'magic_ex_notice';

	/**
	 * Menu name.
	 *
	 * @var string
	 */
	private $menu_name = '';

	/**
	 * Export/import magic type.
	 *
	 * @var string
	 */
	private $magic_type = '';

	/**
	 * Magic item of the current type.
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Current item labels.
	 *
	 * @var stdClass
	 */
	protected $magic_item_labels = null;

	/**
	 * Screen ID for a current magic type.
	 *
	 * @var string
	 */
	protected $screen_id = '';

	/**
	 * Export file data having 'name', 'path' and 'url' keys.
	 *
	 * @var array
	 */
	protected $export_file = array();

	/**
	 * Importing progress data.
	 *
	 * @var array
	 */
	protected $import_progress = array();

	/**
	 * Getes allowed magic items for the current magic type.
	 *
	 * @return array
	 */
	abstract protected function get_allowed_magic_items();

	/**
	 * Initializes magic item for the current type.
	 */
	abstract protected function init_magic_item();

	/**
	 * Gets array of the related magic type keys to export.
	 *
	 * @param array $item_ids Item IDs to get export keys for.
	 * @return array
	 */
	abstract protected function get_export_keys( $item_ids = array() );

	/**
	 * Gets related group by provided key (post, meta, taxonomy etc.).
	 *
	 * @param string $key Item key.
	 * @return string Related group name.
	 */
	abstract protected function get_key_group( $key );

	/**
	 * Gets list of individual items to be exported based on provided arguments.
	 *
	 * @param array $args Quering arguments.
	 * @return array
	 */
	abstract protected function get_individual_export_items( $args );

	/**
	 * Gets mandatory keys for the current magic type.
	 *
	 * @return array
	 */
	abstract protected function get_mandatory_keys();

	/**
	 * Gets filtered mandatory keys for the current magic type.
	 *
	 * @return array
	 */
	protected function get_filtered_mandatory_keys() {
		return apply_filters( 'magic_ex_im_get_mandatory_keys', $this->get_mandatory_keys() );
	}

	/**
	 * Gets item IDs for the export.
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 * @return int[]
	 */
	abstract protected function get_export_items( $posted_data = array(), $include_related_items = true );

	/**
	 * Gets data from the database by provided group, key and item ID.
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 * @return mixed
	 */
	abstract protected function get_key_data( $group, $key, $item_id );

	/**
	 * Gets existing item ID based on provided item data.
	 *
	 * @param array $item_data Item importing data.
	 * @return int|string|false
	 */
	abstract protected function get_existing_item( $item_data );

	/**
	 * Updates existing item based on provided item data.
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 * @return int|false|WP_Error Updated item ID, false or WP_Error.
	 */
	abstract protected function update_existing_item( $item_data, $item_id );

	/**
	 * Creates new item based on provided item data.
	 *
	 * @param array $item_data Item importing data.
	 * @return int|false|WP_Error Created item ID, false or WP_Error.
	 */
	abstract protected function create_new_item( $item_data );

	/**
	 * Updates importing item with provided data.
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	abstract protected function update_importing_item( $item_data, $item_id );

	/**
	 * Gets importing item name based on provided item data.
	 *
	 * @param array $item_data Item data.
	 * @return string
	 */
	abstract protected function get_importing_item_name( $item_data );

	/**
	 * Gets count of exporting magic items without related items.
	 *
	 * @param array $posted_data Export form data.
	 * @return int
	 */
	protected function get_export_items_count( $posted_data = array() ) {
		$export_item_ids = $this->get_export_items( $posted_data, false );
		return count( $export_item_ids );
	}

	/**
	 * Validates item data before importing.
	 *
	 * @param array $item_data Item data to validate. Unused in base class.
	 * @return string[] Validation errors or empty array if valid.
	 */
	protected function validate_importing_item_data( $item_data ) { // phpcs:ignore Generic.CodeAnalysis
		return array();
	}

	/**
	 * Adjusts item value during export or import.
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Key of the provided value.
	 * @param string $key_group Group name $key belongs to.
	 *
	 * @return mixed
	 */
	public function adjust_value( $value, $key, $key_group ) { // phpcs:ignore Generic.CodeAnalysis
		return $value;
	}

	/**
	 * Determines if magic items are allowed to be updated, not only created.
	 *
	 * @return bool
	 */
	protected function allow_to_update_items() {
		return true;
	}

	/**
	 * Determines if magic items are allowed to be created, not only updated.
	 *
	 * @return bool
	 */
	protected function allow_to_create_items() {
		return true;
	}

	/**
	 * Gets current plugin version.
	 *
	 * @return string
	 */
	protected function get_plugin_version() {
		return get_plugin_data( MAGIC_EX_IM_FILE )['Version'];
	}

	/**
	 * Gets export admin notice based on processing items data.
	 *
	 * @param array $items_export_data Exporting items data.
	 * @return string
	 */
	protected function get_export_admin_notice( $items_export_data ) {

		// Skip exporting items keys (file header).
		array_shift( $items_export_data );

		$export_notice = sprintf(
			'Export file successfully generated with <b>%d</b> %s',
			count( $items_export_data ),
			count( $items_export_data ) > 1 ? $this->magic_item_labels->name : $this->magic_item_labels->singular_name,
		);

		return $export_notice;
	}

	/**
	 * Processes filter hook 'http_request_timeout'.
	 *
	 * @param float $timeout_value Timeout value.
	 * @return int
	 */
	public function http_request_timeout( $timeout_value ) {

		// Set minimum 30 seconds timeout.
		$timeout_value = intval( $timeout_value );
		$timeout_value = $timeout_value >= 30 ? $timeout_value : 30;

		return apply_filters( 'magic_im_http_request_timeout', $timeout_value );
	}

	/**
	 * Gets sanitized post data.
	 *
	 * @return array
	 */
	protected function get_sanitized_post_data() {
		return map_deep( wp_unslash( $_POST ), 'sanitize_text_field' );
	}

	/**
	 * Class constructor.
	 *
	 * @param string $magic_type Type for export and import UI.
	 * @param string $menu_name Name for the related menu item.
	 */
	public function __construct( $magic_type = '', $menu_name = '' ) {

		$this->magic_type = $magic_type;
		$this->menu_name  = $menu_name;

		add_action( 'admin_menu', array( $this, 'admin_menu' ) );
		add_action( 'current_screen', array( $this, 'current_screen' ) );

		add_action( "wp_ajax_get_individual_export_{$this->magic_type}", array( $this, 'ajax_get_individual_export_items' ) );
		add_action( "wp_ajax_refresh_export_{$this->magic_type}_keys", array( $this, 'ajax_refresh_export_keys' ) );
		add_action( "wp_ajax_process_import_{$this->magic_type}_form", array( $this, 'ajax_process_import_form' ) );
	}

	/**
	 * Processes action hook 'admin_notices'.
	 */
	public function admin_notices() {

		// Clear all third party admin notices.
		remove_all_actions( 'admin_notices' );
		remove_all_actions( 'all_admin_notices' );

		$export_notice = get_transient( self::EX_NOTICE_KEY );

		// Show export notice if present.
		if ( $export_notice ) {
			wp_admin_notice( $export_notice, array( 'type' => 'success' ) );
			delete_transient( self::EX_NOTICE_KEY );
		}
	}

	/**
	 * Sets current magic item.
	 */
	protected function set_magic_item() {

		// phpcs:ignore WordPress.Security.NonceVerification.Recommended
		$this->magic_item = sanitize_key( wp_unslash( $_REQUEST['magic_item'] ?? '' ) );

		if ( ! $this->magic_item || ! array_key_exists( $this->magic_item, $this->get_allowed_magic_items() ) ) {
			$this->magic_item = array_key_first( $this->get_allowed_magic_items() );
		}

		$this->init_magic_item();
		Magic_EX_IM_Data::set_magic( $this->magic_type, $this->magic_item );

		add_action( 'magic_ex_before_submit_button', array( $this, 'magic_ex_before_submit_button' ) );
		add_action( 'magic_im_media_downloaded', array( $this, 'magic_im_media_downloaded' ) );

		do_action( 'magic_ex_im_init' );
	}

	/**
	 * Gets filtered and grouped keys for the export form.
	 *
	 * @param array $posted_data Form posted data.
	 * @return array
	 */
	protected function get_grouped_export_keys( $posted_data = array() ) {

		$export_item_ids = $this->get_export_items( $posted_data );

		if ( ! $export_item_ids ) {
			return array();
		}

		return apply_filters(
			'magic_ex_keys',
			$this->group_export_keys( $this->get_export_keys( $export_item_ids ) )
		);
	}

	/**
	 * Processes action hook 'magic_im_media_downloaded'.
	 *
	 * @param int $media_id Downloaded media file ID.
	 */
	public function magic_im_media_downloaded( $media_id ) {

		if ( $media_id ) {
			++$this->import_progress['media_downloaded'];
		}
	}

	/**
	 * Processes action hook 'admin_menu'.
	 *
	 * @global array $admin_page_hooks
	 */
	public function admin_menu() {

		// Add menu page linked to the current export/import type and set screen ID.

		global $admin_page_hooks;

		$menu_name = $this->menu_name ? $this->menu_name : ucfirst( $this->magic_type );

		// Add parent menu page refering to the first submenu page if not yet.
		if ( ! isset( $admin_page_hooks['magic-export-import-posts'] ) ) {

			add_menu_page(
				'Export & Import ' . $menu_name,
				'Magic Export & Import',
				'manage_options',
				"magic-export-import-$this->magic_type",
				array( $this, 'admin_menu_page_content' ),
				'dashicons-database'
			);

			// Add first submenu page.
			$this->screen_id = add_submenu_page(
				"magic-export-import-$this->magic_type",
				'Export & Import ' . $menu_name,
				$menu_name,
				'manage_options',
				"magic-export-import-$this->magic_type",
			);

			return;
		}

		// Add submenu page.
		$this->screen_id = add_submenu_page(
			'magic-export-import-posts',
			'Export & Import ' . $menu_name,
			$menu_name,
			'manage_options',
			"magic-export-import-$this->magic_type",
			array( $this, 'admin_menu_page_content' ),
		);
	}

	/**
	 * Outputs admin menu page content.
	 */
	public function admin_menu_page_content() {

		$args = array(
			'magic_items_count'       => $this->get_export_items_count(),
			'magic_item_labels'       => $this->magic_item_labels,
			'allowed_magic_items'     => $this->get_allowed_magic_items(),
			'mandatory_keys'          => $this->get_filtered_mandatory_keys(),
			'export_file'             => $this->export_file,
			'allow_to_update_items'   => $this->allow_to_update_items(),
			'allow_to_create_items'   => $this->allow_to_create_items(),
			'magic_item_grouped_keys' => $this->get_grouped_export_keys(),
		);

		load_template( MAGIC_EX_IM_ABSPATH . 'template-parts/page-content.php', false, $args, );
	}

	/**
	 * Processes action hook 'current_screen'.
	 */
	public function current_screen() {

		// Init export/import functionality only if any registered magic type is loading.
		if ( get_current_screen()->id !== $this->screen_id ) {
			return;
		}

		$this->set_magic_item();

		// Set export file data based on the current magic type.

		$export_file_name = sprintf(
			'magic-export-%s%s-%s.csv',
			$this->magic_type,
			$this->magic_item === $this->magic_type ? '' : '-' . $this->magic_item,
			wp_parse_url( home_url() )['host']
		);

		$this->export_file = array(
			'name' => $export_file_name,
			'path' => MAGIC_EX_IM_ABSPATH . 'export/' . $export_file_name,
			'url'  => plugin_dir_url( MAGIC_EX_IM_FILE ) . 'export/' . $export_file_name,
		);

		set_time_limit( 0 ); // Set max execution time with no limit.

		add_action( "load-$this->screen_id", array( $this, 'load' ) );
		add_action( 'admin_notices', array( $this, 'admin_notices' ), 1 );
		add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ) );
	}

	/**
	 * Processes action hook 'magic_ex_before_submit_button'.
	 */
	public function magic_ex_before_submit_button() {

		// Output advanced filters for the current magic type.

		$filters_name  = str_replace( '_', '-', $this->magic_type );
		$template_file = MAGIC_EX_IM_ABSPATH . "template-parts/advanced-filters/advanced-filters-{$filters_name}.php";

		if ( file_exists( $template_file ) ) {

			load_template( MAGIC_EX_IM_ABSPATH . 'template-parts/advanced-filters/advanced-filters-toggler.php' );
			echo '<div id="magic-ex-advanced-wrapper">';
			echo '<div id="magic-ex-advanced-filters">';

			load_template(
				$template_file,
				false,
				array(
					'magic_items'       => $this->get_export_items(),
					'magic_items_count' => $this->get_export_items_count(),
					'magic_item_labels' => $this->magic_item_labels,
				)
			);

			do_action( 'magic_ex_advanced_filters' );
			echo '</div>';
			echo '<button id="magic-ex-apply-filters" class="button-secondary" disabled>Apply filters</button>';
			echo '</div>';
		}
	}

	/**
	 * Processes dynamic action hook "load-$this->screen_id".
	 *
	 * @see Magic_EX_IM_Type::admin_menu()
	 */
	public function load() {

		$posted_data = $this->get_sanitized_post_data();

		if ( isset( $posted_data['magic-ex-nonce'] ) && wp_verify_nonce( $posted_data['magic-ex-nonce'], 'magic-ex' ) ) {
			$this->process_export_form( $posted_data );
			return;
		}
	}

	/**
	 * Processes action hook 'admin_enqueue_scripts'.
	 */
	public function admin_enqueue_scripts() {

		// Styles.

		wp_enqueue_style(
			'magic-export-import-select2',
			plugin_dir_url( MAGIC_EX_IM_FILE ) . '/assets/select2/select2.min.css',
			null,
			'4.0.13'
		);

		wp_enqueue_style(
			'magic-export-import',
			plugin_dir_url( MAGIC_EX_IM_FILE ) . 'assets/magic-export-import.min.css',
			null,
			microtime()
		);

		// Scripts.

		wp_enqueue_script(
			'magic-export-import-select2',
			plugin_dir_url( MAGIC_EX_IM_FILE ) . 'assets/select2/select2.min.js',
			array( 'jquery' ),
			'4.0.13',
			true,
		);

		wp_enqueue_script(
			'magic-export-import',
			plugin_dir_url( MAGIC_EX_IM_FILE ) . 'assets/magic-export-import.min.js',
			array( 'jquery' ),
			microtime(),
			true,
		);

		// PHP data for the main script.
		wp_localize_script(
			'magic-export-import',
			'WPLocalize',
			array(
				'admin_ajax_url' => admin_url( 'admin-ajax.php' ),
				'magic_type'     => $this->magic_type,
				'magic_item'     => $this->magic_item,
			)
		);
	}

	/**
	 * Ouputs refreshed export keys selects HTML on AJAX.
	 */
	public function ajax_refresh_export_keys() {

		$this->set_magic_item();
		$posted_data = $this->get_sanitized_post_data();

		$result = array(
			'export_items_cnt'       => $this->get_export_items_count( $posted_data ),
			'export_keys_html'       => '',
			'export_media_keys_html' => '',
		);

		$args = array(
			'magic_item_grouped_keys' => $this->get_grouped_export_keys( $posted_data ),
			'mandatory_keys'          => $this->get_filtered_mandatory_keys(),
		);

		ob_start();
		load_template( MAGIC_EX_IM_ABSPATH . 'template-parts/select-export-keys.php', false, $args );
		$result['export_keys_html'] = ob_get_clean();

		ob_start();
		load_template( MAGIC_EX_IM_ABSPATH . 'template-parts/select-export-media-keys.php', false, $args );
		$result['export_media_keys_html'] = ob_get_clean();

		echo wp_json_encode( $result );
		wp_die();
	}

	/**
	 * Processes export form.
	 *
	 * @param array $posted_data Posted form data.
	 */
	protected function process_export_form( $posted_data ) {

		$user_params = array(
			'magic-ex-keys'     => $posted_data['magic-ex-keys'],
			'export-media-keys' => $posted_data['magic-ex-media-keys'] ?? array(),
		);

		Magic_EX_IM_Data::set_user_params( $user_params, 'export' );

		$items_export_data = array( array() );
		$export_keys       = magic_ex_im_get_user_params( 'magic-ex-keys' );

		// Set file header with keys.
		array_push( $items_export_data[0], ...$export_keys );
		array_push( $items_export_data[0], self::FILE_PARAMS_KEY );

		// Group data by keys.

		$export_keys_grouped = array();

		foreach ( $export_keys as $key ) {

			list($group, $key) = array_values( magic_ex_im_parse_key( $key ) );

			$export_keys_grouped[ $group ] ??= array();
			$export_keys_grouped[ $group ][] = $key;
		}

		// Get item IDs to export.
		$export_items = $this->get_export_items( $posted_data );

		foreach ( $export_items as $index => $item_id ) {

			Magic_EX_IM_Data::set_processing_item( $item_id );
			$item_export_data = array();

			foreach ( $export_keys_grouped as $group => $keys ) {
				foreach ( $keys as $key ) {

					$key_data = apply_filters(
						'magic_ex_get_key_data',
						$this->get_key_data( $group, $key, $item_id ),
						$key,
						$group,
					);

					$item_export_data[ $key ] = magic_ex_maybe_adjust_value( $key_data, $key, $group );
				}
			}

			// Set export parameters within first item.
			if ( ! $index ) {

				$params = array(
					'old-domain'     => wp_parse_url( home_url() )['host'],
					'plugin-version' => $this->get_plugin_version(),
				);

				$params = apply_filters( 'magic_ex_file_params', $params );

				$item_export_data[] = $params;
			} else {
				$item_export_data[] = '';
			}

			$items_export_data[] = array_map( 'maybe_serialize', $item_export_data );
		}

		Magic_EX_IM_Data::reset_processing_item();

		// Open file in write mode.
		$fp = fopen( $this->export_file['path'], 'w' );

		if ( false === $fp ) {
			wp_admin_notice( error_get_last()['message'], array( 'type' => 'error' ) );
			return;
		}

		// Write UTF-8 Byte Order Mark at the file beginning.
		$bom = pack( 'H*', 'EFBBBF' );
		fwrite( $fp, $bom );

		// Write data to the file.
		foreach ( $items_export_data as $row ) {
			fputcsv( $fp, $row );
		}

		fclose( $fp );

		// Generate and save export notice as transient.

		$export_notice = sprintf(
			'%s <a id="magic-ex-generated-file" download href="%s">Download</a>',
			$this->get_export_admin_notice( $items_export_data ),
			$this->export_file['url'] . '?ver=' . filemtime( $this->export_file['path'] ),
		);

		set_transient( self::EX_NOTICE_KEY, $export_notice );

		// Redirect to the same page to prevent form resubmission on page reload in Safari.
		wp_safe_redirect( esc_url_raw( add_query_arg( null, null ) ) );
		exit;
	}

	/**
	 * Processes import form via AJAX.
	 *
	 * @throws Exception Error during the import.
	 */
	public function ajax_process_import_form() {

		try {
			$this->set_magic_item();
			$posted_data = $this->get_sanitized_post_data();

			add_filter( 'http_request_timeout', array( $this, 'http_request_timeout' ), PHP_INT_MAX - 1 );

			$this->import_progress = array_merge(
				array(
					'first_entry'      => true,
					'user_params'      => array(),
					'file_params'      => array(),
					'finish'           => false,
					'total'            => 0,
					'processed'        => 0,
					'updated'          => 0,
					'created'          => 0,
					'skipped'          => 0,
					'media_downloaded' => 0,
					'error'            => '',
					'progress_html'    => '',
					'notices_html'     => '',
				),
				$posted_data['progress'] ? json_decode( $posted_data['progress'], true ) : array()
			);

			// Validate import file and nonce.

			if ( ! isset( $_FILES['magic-im-file'] ) || ! wp_verify_nonce( $posted_data['magic-im-nonce'], 'magic-im' ) ) {

				$error = isset( $_FILES['magic-im-file'] ) ? 'Security error' : 'Import file not found';
				throw new Exception( $error );
			}

			$import_file = $_FILES['magic-im-file']; //phpcs:ignore WordPress.Security

			if ( UPLOAD_ERR_OK !== $import_file['error'] ) {
				throw new Exception( self::FILE_UPLOAD_ERRORS[ $import_file['error'] ] );
			}

			// Read import file content.

			$fp                  = fopen( $import_file['tmp_name'], 'r' );
			$csv_delimiter       = $posted_data['magic-im-csv-delimiter'];
			$import_file_content = array();

			if ( false !== $fp ) {

				while ( true ) {
					$file_data_row = fgetcsv( $fp, null, $csv_delimiter );

					if ( false === $file_data_row ) {
						break;
					}

					if ( $import_file_content && count( $file_data_row ) !== count( $import_file_content[0] ) ) {
						throw new Exception( 'Failed to read the import file. Might be invalid <a href="#magic-im-csv-delimiter">CSV delimiter</a>.' );
					}

					$import_file_content[] = $file_data_row;
				}

				fclose( $fp );
			}

			if ( false === $fp || count( $import_file_content ) <= 1 ) {
				throw new Exception( 'Failed to read the import file.' );
			}

			// Set import file keys and remove UTF-8 BOM if present.
			$import_file_keys    = array_map( 'trim', array_shift( $import_file_content ) );
			$bom                 = pack( 'H*', 'EFBBBF' );
			$import_file_keys[0] = preg_replace( "/^$bom/", '', $import_file_keys[0] );

			// First entry, validate file content and set import parameters.
			if ( $this->import_progress['first_entry'] ) {

				$this->import_progress['total'] = count( $import_file_content );

				// Validate import file keys.
				foreach ( $import_file_keys as $index => $prefixed_key ) {
					$key = magic_ex_im_parse_key( $prefixed_key, 'key' );

					if ( ! $key || sanitize_text_field( $key ) !== $key ) {

						$error = sprintf(
							'%s key %s in the column <b>#%d</b>.',
							$key ? 'Inappropriate' : 'Empty',
							$key ? "<code>{$key}</code>" : '',
							$index + 1
						);

						$import_keys_errors[] = $error;
					}
				}

				if ( isset( $import_keys_errors ) ) {
					array_map( 'magic_im_add_notice', $import_keys_errors );
					throw new Exception( '' );
				}

				// Set and validate import file parameters, stored within the first item.

				$first_item_data = array_combine( $import_file_keys, $import_file_content[0] );

				if ( ! isset( $first_item_data[ self::FILE_PARAMS_KEY ] ) ) {
					throw new Exception( sprintf( 'Mandatory column <code>%s</code> is missing.', self::FILE_PARAMS_KEY ) );
				}

				$file_params = maybe_unserialize( $first_item_data[ self::FILE_PARAMS_KEY ] );
				$old_domain  = $file_params['old-domain'] ?? '';

				// Validate old domain.
				if ( ! filter_var( $old_domain, FILTER_VALIDATE_DOMAIN, FILTER_FLAG_HOSTNAME ) ) {

					$error = sprintf(
						'%s <code>old-domain</code> in the column <code>%s</code>.',
						$old_domain ? 'Invalid' : 'Missing',
						self::FILE_PARAMS_KEY
					);

					throw new Exception( $error );
				}

				// Validate file parameters.
				$file_params_errors = apply_filters( 'magic_im_validate_file_params', array(), $file_params );

				if ( $file_params_errors ) {
					array_map( 'magic_im_add_notice', $file_params_errors );
					throw new Exception( '' );
				}

				// Set user parameters from the import form.
				$user_params = array(
					'skip-new-items'      => isset( $posted_data['magic-im-skip-new-items'] ) || ! $this->allow_to_create_items(),
					'skip-existing-items' => isset( $posted_data['magic-im-skip-existing-items'] ) || ! $this->allow_to_update_items(),
					'download-media'      => isset( $posted_data['magic-im-download-media'] ),
					'test-mode'           => isset( $posted_data['magic-im-test-mode'] ),
				);

				$this->import_progress['user_params'] = $user_params;
				$this->import_progress['file_params'] = $file_params;
				$this->import_progress['first_entry'] = false;

				// Not a first entry, process import.
			} else {

				Magic_EX_IM_Data::set_user_params( $this->import_progress['user_params'], 'import' );
				Magic_EX_IM_Data::set_import_file_params( $this->import_progress['file_params'] );

				// Determine next slice to process.
				$next_slice_cnt = magic_ex_im_get_user_params( 'download-media' ) ? rand( 3, 5 ) : rand( 5, 18 );

				if ( $next_slice_cnt > ceil( $this->import_progress['total'] / 2 ) ) {
					$next_slice_cnt = ceil( $this->import_progress['total'] / 2 );
				}

				$next_slice = array_slice( $import_file_content, $this->import_progress['processed'], $next_slice_cnt );

				// Determine if the last slice is being processed.
				$finish_cnt                      = $this->import_progress['processed'] + count( $next_slice );
				$this->import_progress['finish'] = count( $import_file_content ) <= $finish_cnt;

				// Set import items data.
				foreach ( $next_slice as $import_file_row ) {

					$import_item_data = array_combine( $import_file_keys, $import_file_row );
					unset( $import_item_data[ self::FILE_PARAMS_KEY ] );

					$import_items_data[] = $import_item_data;
				}

				$this->import_items( $import_items_data );
			}
		} catch ( Exception $e ) {

			magic_im_add_notice( $e->getMessage() );
			$this->import_progress['error'] = 'Import cannot be initiated.';
		}

		// Set import progress HTML.
		ob_start();
		load_template( MAGIC_EX_IM_ABSPATH . 'template-parts/import-progress.php', false, $this->import_progress );
		$this->import_progress['progress_html'] = ob_get_clean();

		$import_notices = array_unique( Magic_EX_IM_Data::get_import_notices() );
		$notices_html   = '';

		// Set import notices HTML.
		foreach ( $import_notices as $notice ) {
			$notices_html .= sprintf( '<li>%s</li>', $notice );
		}

		$this->import_progress['notices_html'] = $notices_html;

		echo wp_json_encode( $this->import_progress );
		wp_die();
	}

	/**
	 * Imports items based on the provided data.
	 *
	 * @param array $items_data Items data to import.
	 */
	protected function import_items( $items_data ) {

		$validation_errors = array();

		// Loop through all items. Trim, unserialize and validate data before import start.
		foreach ( $items_data as $index => &$item_data ) {
			$item_data = array_map( 'trim', (array) $item_data );

			// Check for an empty file row.
			if ( ! array_filter( $item_data ) ) {
				$validation_errors[ $index ][] = 'Empty row.';
				continue;
			}

			// Check for the mandatory data.
			foreach ( $this->get_filtered_mandatory_keys() as $mandatory_key ) {
				$mandatory_key = magic_ex_im_build_key( $mandatory_key, $this->get_key_group( $mandatory_key ) );

				if ( ! $item_data[ $mandatory_key ] ?? '' ) {
					$item_error = sprintf( 'Mandatory data <code>%s</code> is missing.', $mandatory_key, );

					$validation_errors[ $index ][] = $item_error;
					continue 2;
				}
			}

			// Unserialize item data if needed.
			foreach ( $item_data as $key => $value ) {
				$items_data[ $index ][ $key ] = maybe_unserialize( $value );

				if ( false === $items_data[ $index ][ $key ] ) {
					$item_error = sprintf( 'Incorrectly serialized data in the column <code>%s</code>.', $key );

					$validation_errors[ $index ][] = $item_error;
					continue 2;
				}
			}

			$item_errors = apply_filters(
				'magic_im_validate_item_data',
				$this->validate_importing_item_data( $item_data ),
				$item_data,
			);

			$item_errors = array_filter( $item_errors );

			if ( $item_errors ) {
				$validation_errors[ $index ] ??= array();
				$validation_errors[ $index ]   = array_merge( $validation_errors[ $index ], $item_errors );
			}
		}

		$items_data_grouped = $this->group_import_items_data( $items_data );

		// Start import looping through all items.
		foreach ( $items_data_grouped as $index => $item_data ) {

			$item_file_row = $this->import_progress['processed'] + 2;
			Magic_EX_IM_Data::set_importing_item( $item_file_row, $this->get_importing_item_name( $item_data ) );

			if ( ! $item_data || isset( $validation_errors[ $index ] ) ) {

				foreach ( $validation_errors[ $index ] ?? array() as $error ) {
					magic_im_add_notice( $error, 'skipped' );
				}

				++$this->import_progress['skipped'];
				++$this->import_progress['processed'];
				continue;
			}

			$item_data = apply_filters( 'magic_im_item_data_before_adjust', $item_data );

			// Adjust importing data before creating or updating the item.
			foreach ( $item_data as $group => $data ) {
				foreach ( $data as $key => $key_data ) {
					$item_data[ $group ][ $key ] = $this->adjust_value( $key_data, $key, $group );
				}
			}

			$existing_item_id = $this->get_existing_item( $item_data );

			// Item exists.
			if ( $existing_item_id ) {

				$is_new_item = false;

				// Skip updating existing items in the test mode.
				if ( magic_ex_im_get_user_params( 'test-mode' ) ) {

					++$this->import_progress['processed'];

					magic_ex_im_get_user_params( 'skip-existing-items' )
						? ++$this->import_progress['skipped']
						: ++$this->import_progress['updated'];

					$notice = sprintf( 'Already exists (id: %s).', $existing_item_id );

					magic_im_add_notice(
						$notice,
						magic_ex_im_get_user_params( 'skip-existing-items' ) ? 'skipped' : 'updated'
					);

					continue;
				}

				// Skip item if user hasn't allowed to update existing items.
				if ( magic_ex_im_get_user_params( 'skip-existing-items' ) ) {

					++$this->import_progress['skipped'];
					++$this->import_progress['processed'];

					$notice = sprintf( 'Already exists (id: %d).', $existing_item_id );
					magic_im_add_notice( $notice, 'skipped' );
					continue;
				}

				$item_id = $this->update_existing_item( $item_data, $existing_item_id );

				// Item doesn't exist.
			} else {

				$is_new_item = true;

				// Skip creating new items in the test mode.
				if ( magic_ex_im_get_user_params( 'test-mode' ) ) {

					++$this->import_progress['processed'];

					magic_ex_im_get_user_params( 'skip-new-items' )
						? ++$this->import_progress['skipped']
						: ++$this->import_progress['created'];

					magic_im_add_notice(
						'Doesn\'t exist.',
						magic_ex_im_get_user_params( 'skip-new-items' ) ? 'skipped' : 'created'
					);

					continue;
				}

				// Skip item if user hasn't allowed to create new items.
				if ( magic_ex_im_get_user_params( 'skip-new-items' ) ) {

					++$this->import_progress['skipped'];
					++$this->import_progress['processed'];

					$notice = sprintf( 'Doesn\'t exist.' );
					magic_im_add_notice( $notice, 'skipped' );
					continue;
				}

				$item_id = $this->create_new_item( $item_data );
			}

			// Check for an error on updating or creating the item.
			if ( ! $item_id || is_wp_error( $item_id ) ) {

				$notice = sprintf(
					'Error to %s. %s',
					$is_new_item ? 'create' : 'update',
					is_wp_error( $item_id ) ? ( '<code>' . $item_id->get_error_message() . '</code>' ) : '',
				);

				magic_im_add_notice( $notice, 'skipped' );

				++$this->import_progress['skipped'];
				++$this->import_progress['processed'];
				continue;
			}

			Magic_EX_IM_Data::set_processing_item( $item_id, $is_new_item );
			do_action( 'magic_im_item_id', $item_data );

			// Adjust importing data after item ID was determined.
			foreach ( $item_data as $group => $data ) {
				foreach ( $data as $key => $key_data ) {
					$item_data[ $group ][ $key ] = magic_im_maybe_adjust_value( $key_data, $key, $group, );
				}
			}

			// Replace remaining old domain with the current in all strings.
			$item_data = magic_im_replace_domain( $item_data );

			// Update importing item.

			$item_data = apply_filters( 'magic_im_item_data_before_update', $item_data );
			$this->update_importing_item( $item_data, $item_id );

			// Count processed items.

			$is_new_item ? $this->import_progress['created']++ : $this->import_progress['updated']++;
			++$this->import_progress['processed'];

			do_action( 'magic_im_item_processed', $item_data );
		}

		Magic_EX_IM_Data::reset_processing_item();
	}

	/**
	 * Processes dynamic action hook "wp_ajax_get_individual_export_{$this->magic_type}".
	 *
	 * @return void Ouputs JSON with individual items for export, ajusted for select2 JS library.
	 */
	public function ajax_get_individual_export_items() {

		$items_per_request = 8;
		$posted_data       = $this->get_sanitized_post_data();
		$current_page      = absint( $posted_data['page'] ?? 1 );

		$args = array(
			'magic_item'        => $posted_data['magic_item'],
			'items_per_request' => $items_per_request,
			's'                 => $posted_data['term'] ?? '',
			'offset'            => ( $current_page - 1 ) * $items_per_request,
		);

		$result = $this->get_individual_export_items( $args );

		// Select2 expects the 'id' parameter to be of type string to function properly.
		// Explanation: https://github.com/select2/select2/issues/6128#issuecomment-996407492.
		foreach ( $result['results'] as &$item ) {
			$item['id'] = (string) $item['id'];
		}

		echo wp_json_encode( $result );
		wp_die();
	}

	/**
	 * Groups export keys by type: post, meta, taxonomy etc.
	 *
	 * @param array $export_keys Keys to group.
	 * @return array
	 */
	protected function group_export_keys( $export_keys ) {
		$grouped_keys = array();

		foreach ( $export_keys as $key ) {

			$group = apply_filters(
				'magic_ex_get_key_group',
				$this->get_key_group( $key ),
				$key,
				$export_keys
			);

			$grouped_keys[ $group ][] = $key;
		}

		return $grouped_keys;
	}

	/**
	 * Groups importing items data by type: post, meta, taxonomy etc.
	 *
	 * @param array $items_data Importing items data.
	 * @return array
	 */
	protected function group_import_items_data( $items_data ) {

		static $groups      = array();
		$grouped_items_data = array();

		foreach ( $items_data as $index => $data ) {

			if ( empty( $data ) || ! is_array( $data ) ) {
				$grouped_items_data[ $index ] = array();
				continue;
			}

			foreach ( $data as $key => $value ) {

				list($predefined_group, $key) = array_values( magic_ex_im_parse_key( $key ) );

				// Set key group if not yet.
				if ( ! isset( $groups[ $key ] ) ) {
					$group = apply_filters( 'magic_im_get_key_group', $this->get_key_group( $key ), $key, $predefined_group );

					// Predefined group should match the defined one.
					if ( $group !== $predefined_group ) {

						// Show group related error only once.
						if ( 0 === $this->import_progress['processed'] ) {

							$error = sprintf(
								'- Key <code>%s</code> associated with the group <code>%s</code> is missing, expected group is <code>%s</code>. %s.',
								$key,
								$predefined_group ? $predefined_group : 'undefined',
								$group,
								magic_ex_im_get_user_params( 'test-mode' ) ? 'Key will be <b>ignored</b>' : 'Key <b>ignored</b>',
							);

							magic_im_add_notice( $error );
						}

						continue;
					}

					$groups[ $key ] = $group;
				}

				$group = $groups[ $key ];

				$grouped_items_data[ $index ][ $group ][ $key ] = $value;
			}
		}

		return $grouped_items_data;
	}
}
