<?php
/**
 * Plugin related functions.
 *
 * @package Magic_Export_Import
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Gets processing magic type: 'posts', 'terms', 'users', 'comments', 'acf_options_pages', 'shop_orders'.
 */
function magic_ex_im_get_magic_type() {
	return Magic_EX_IM_Data::get_magic_type();
}

/**
 * Gets processing magic item: post type name, taxonomy name, 'users', 'comments', 'acf_options_pages', 'shop_orders'.
 *
 * Different for 'posts' and 'terms' magic types, same as magic type for the rest.
 */
function magic_ex_im_get_magic_item() {
	return Magic_EX_IM_Data::get_magic_item();
}

/**
 * Gets processing magic type and magic item seperated by column.
 *
 * @return string
 */
function magic_ex_im_get_magic() {
	return magic_ex_im_get_magic_type() . ':' . magic_ex_im_get_magic_item();
}

/**
 * Gets file row number of the importing item.
 *
 * @return int
 */
function magic_im_get_item_file_row() {
	return Magic_EX_IM_Data::get_importing_item_file_row();
}

/**
 * Gets importing item name.
 *
 * @return string
 */
function magic_im_get_item_name() {
	return Magic_EX_IM_Data::get_importing_item_name();
}

/**
 * Gets processing item ID.
 *
 * @return int
 */
function magic_ex_im_get_item_id() {
	return Magic_EX_IM_Data::get_processing_item_id();
}

/**
 * Checks if importing item was newly created, otherwise is updating as existing.
 *
 * @return bool
 */
function magic_im_is_new_item() {
	return Magic_EX_IM_Data::is_importing_item_new();
}

/**
 * Gets user parameters.
 *
 * @param string $param Specific parameter name to get.
 * @return mixed
 */
function magic_ex_im_get_user_params( $param = '' ) {
	return Magic_EX_IM_Data::get_user_params( $param );
}

/**
 * Gets import file parameters.
 *
 * @param string $param Specific parameter name to get.
 * @return mixed
 */
function magic_im_get_file_params( $param = '' ) {
	return Magic_EX_IM_Data::get_import_file_params( $param );
}

/**
 * Determines if exporting process is running.
 *
 * @return bool
 */
function magic_ex_im_is_export() {
	return 'export' === Magic_EX_IM_Data::get_current_process();
}

/**
 * Determines if importing process is running.
 *
 * @return bool
 */
function magic_ex_im_is_import() {
	return 'import' === Magic_EX_IM_Data::get_current_process();
}

/**
 * Gets current domain name during import.
 *
 * @return string
 */
function magic_im_get_current_domain() {

	static $current_domain = null;
	$current_domain      ??= wp_parse_url( home_url() )['host'];

	return $current_domain;
}

/**
 * Adds log notice during import.
 *
 * @param string $notice Notice content.
 * @param string $state Item state, e.g. 'skipped', 'updated', 'created'.
 */
function magic_im_add_notice( $notice, $state = '' ) {

	if ( ! $notice ) {
		return;
	}

	if ( magic_im_get_item_file_row() ) {

		$item_name = magic_im_get_item_name();
		$state_str = '';

		if ( $state ) {

			$state_str = magic_ex_im_get_user_params( 'test-mode' )
				? sprintf( 'Item will be <b>%s</b>.', $state )
				: sprintf( 'Item <b>%s</b>.', $state );
		}

		$notice = sprintf(
			'- Row <b>#%d</b>%s: %s %s',
			magic_im_get_item_file_row(),
			$item_name ? ' <code>' . $item_name . '</code>' : '',
			$notice,
			$state_str
		);
	}

	Magic_EX_IM_Data::add_import_notice( $notice );
}

/**
 * Gets possible post statuses for export or import.
 *
 * @return array
 */
function magic_ex_im_get_post_statuses() {

	$post_statuses = array_keys( get_post_stati() );
	$post_statuses = array_diff( $post_statuses, array( 'auto-draft' ) );

	return $post_statuses;
}

/**
 * Adjusts value for the export.
 *
 * @param mixed  $value Value to adjust.
 * @param string $key Key of the provided value.
 * @param string $key_group Group name $key belongs to.
 *
 * @return mixed
 */
function magic_ex_maybe_adjust_value( $value, $key, $key_group ) {

	$new_value = apply_filters( 'magic_ex_maybe_adjust_value', $value, $key, $key_group );

	// Early return value in case it is empty or already adjusted within the filter.
	if ( ! $new_value || $new_value !== $value ) {
		return $new_value;
	}

	$value = magic_ex_im_adjust_media_value( $value, $key, $key_group );

	return $value;
}

/**
 * Adjusts user ID for export (converts to login by default) or import (converts back to user ID).
 *
 * @param mixed  $value Value to adjust.
 * @param string $adjust_as Adjust value as: 'login', 'email' etc.
 *
 * @return mixed
 */
function magic_ex_im_adjust_user_value( $value, $adjust_as = 'login' ) {

	if ( ! $value ) {
		return $value;
	}

	// Convert user ID into login during export.
	if ( magic_ex_im_is_export() && is_numeric( $value ) ) {
		$value = get_user_by( 'ID', $value )->{"user_$adjust_as"} ?? '';
	}

	// Convert user login into ID during import.
	if ( magic_ex_im_is_import() && is_string( $value ) ) {
		$value = get_user_by( $adjust_as, $value )->ID ?? '';
	}

	return $value;
}

/**
 * Processes media value during export (converts to URL) or import (converts back to attachment ID).
 *
 * @param mixed  $value Value to process.
 * @param string $key Key of the provided value.
 * @param string $key_group Group name $key belongs to.
 *
 * @return mixed
 */
function magic_ex_im_adjust_media_value( $value, $key, $key_group ) {

	if ( is_array( $value ) ) {
		foreach ( array_keys( $value ) as $index ) {
			$value[ $index ] = magic_ex_im_adjust_media_value( $value[ $index ], $key, $key_group );
		}
	}

	$media_id_prefix = 'magic-media-id-url:';

	// Process numeric value during export.
	if ( magic_ex_im_is_export() && is_numeric( $value ) ) {

		$export_key        = magic_ex_im_build_key( $key, $key_group );
		$export_media_keys = magic_ex_im_get_user_params( 'export-media-keys' );

		// Convert media files IDs into URLs.
		if ( in_array( $export_key, $export_media_keys, true ) ) {

			$value = wp_get_attachment_url( $value );

			if ( $value ) {
				$value = $media_id_prefix . $value;
			}
		}
	}

	// Process string during import.
	if ( magic_ex_im_is_import() && is_string( $value ) ) {

		// Convert media files URLs into local IDs/URLs.
		if ( str_starts_with( $value, $media_id_prefix ) ) {

			$media_url = str_replace( $media_id_prefix, '', $value );
			$value     = magic_im_media_url_to_local_id( $media_url );

		} else {
			$value = magic_im_media_url_to_local_url( $value );
		}
	}

	return $value;
}

/**
 * Gets post ID based on the provided post name.
 *
 * @param string $post_name Post name.
 * @param string $post_type Post type.
 * @param array  $args Additional query arguments.
 *
 * @return int|false Post ID on success, false if post not found or $post_name is not provided.
 */
function magic_im_post_id_by_name( $post_name, $post_type = 'any', $args = array() ) {

	if ( ! $post_name ) {
		return false;
	}

	$args = array_merge(
		array(
			'post_type'      => $post_type,
			'post_name__in'  => magic_ex_im_get_post_name__in( $post_name ),
			'post_status'    => magic_ex_im_get_post_statuses(),
			'posts_per_page' => 1,
			'fields'         => 'ids',
			'no_found_rows'  => true,
		),
		$args
	);

	$args       = apply_filters( 'magic_im_post_id_by_name_args', $args, $post_type );
	$post_query = new WP_Query( $args );

	return reset( $post_query->posts );
}

/**
 * Gets possible post name variations for one post including trashed state.
 *
 * @param string $post_name The post name (slug).
 * @return string[]
 */
function magic_ex_im_get_post_name__in( $post_name ) {

	if ( ! $post_name ) {
		return array();
	}

	// WP appends '__trashed' to the end of trashed post name.
	$post_name    = str_ends_with( $post_name, '__trashed' ) ? substr( $post_name, 0, -9 ) : $post_name;
	$post_name_in = array( $post_name, $post_name . '__trashed' );

	return $post_name_in;
}

/**
 * Adjusts value for import.
 *
 * Converts external URLs into local attachment IDs or URLs.
 *
 * @param mixed  $value Value to adjust.
 * @param string $key Key of the provided value.
 * @param string $key_group Group name $key belongs to.
 *
 * @return mixed
 */
function magic_im_maybe_adjust_value( $value, $key, $key_group ) {

	$new_value = apply_filters( 'magic_im_maybe_adjust_value', $value, $key, $key_group );

	// Early return value in case it is empty or already adjusted within the filter.
	if ( ! $new_value || $new_value !== $value ) {
		return $new_value;
	}

	// Convert media files URLs into local IDs/URLs.
	$value = magic_ex_im_adjust_media_value( $value, $key, $key_group );

	return $value;
}

/**
 * Converts media file URL into local URL, downloading media file if needed.
 *
 * @param string $media_url Media file URL.
 * @return string|false
 *
 * @see magic_im_media_url_to_local_id().
 */
function magic_im_media_url_to_local_url( $media_url ) {

	if ( ! wp_http_validate_url( $media_url ) ) {
		return $media_url;
	}

	$media_id = magic_im_media_url_to_local_id( $media_url );

	if ( $media_id && is_numeric( $media_id ) ) {
		return wp_get_attachment_url( $media_id );
	}

	return $media_url;
}

/**
 * Converts media file URL into local attachment post ID, downloading media file if needed.
 *
 * @param string $media_url Media file URL.
 * @param array  $media_post_data Importing attachment post data.
 * @param array  $prevent_download Whether to prevent media file to be downloaded despite user parameters.
 * @return int|string|false Attachment ID on successs, URL without change if not a valid old domain file URL,
 * false otherwise.
 */
function magic_im_media_url_to_local_id( $media_url, $media_post_data = array(), $prevent_download = false ) {

	// Check if URL is a valid media file URL from the old or current domain, otherwise return it unchanged.

	if ( ! wp_http_validate_url( $media_url ) ) {
		return $media_url;
	}

	$media_url_parsed = wp_parse_url( rtrim( $media_url, '/' ) );
	$media_url_domain = $media_url_parsed['host'] ?? null;

	if ( ! isset( $media_url_parsed['path'], $media_url_domain ) ) {
		return $media_url;
	}

	$old_domain     = magic_im_get_file_params( 'old-domain' );
	$current_domain = magic_im_get_current_domain();

	if ( $old_domain !== $media_url_domain && $current_domain !== $media_url_domain ) {
		return $media_url;
	}

	$media_url_extension = strtolower( pathinfo( $media_url, PATHINFO_EXTENSION ) );
	$media_url_extension = explode( '?', $media_url_extension, 2 )[0];

	if ( ! $media_url_extension || in_array( $media_url_extension, array( 'html', 'htm', 'php' ), true ) ) {
		return $media_url;
	}

	// Check if URL refers to an existing media.
	if ( $current_domain === $media_url_domain ) {

		$media_file_id = attachment_url_to_postid( $media_url );

		// Try to find full size image by URL.
		// Explanation: https://core.trac.wordpress.org/ticket/41816#comment:7.
		if ( ! $media_file_id ) {
			$media_file_id = attachment_url_to_postid( preg_replace( '/-\d+[Xx]\d+\./', '.', $media_url ) );
		}

		if ( $media_file_id ) {
			return $media_file_id;
		}
	}

	// Check if attachment with the same name and mime type already exists in the database.

	global $wpdb;

	$media_post_data = array_filter( $media_post_data );
	$media_file_name = pathinfo( $media_url, PATHINFO_FILENAME );
	$media_file_type = wp_check_filetype( $media_url );

	$attachments = $wpdb->get_results(
		$wpdb->prepare(
			"SELECT ID from $wpdb->posts where post_type = 'attachment' 
			AND post_mime_type = %s 
			AND (post_name = %s OR post_title = %s)",
			$media_file_type['type'],
			$media_post_data['post_name'] ?? sanitize_title( $media_file_name ),
			$media_post_data['post_title'] ?? $media_file_name,
		),
		ARRAY_A
	);

	// Check if attachment media file exists in the upload directory.
	foreach ( $attachments as $attachment ) {
		if ( file_exists( get_attached_file( $attachment['ID'] ) ) ) {
			return $attachment['ID'];
		}
	}

	// Media file not found, check whether it should be downloaded.
	if ( $current_domain === $media_url_domain || ! magic_ex_im_get_user_params( 'download-media' ) || $prevent_download ) {
		return false;
	}

	// Check if file type is allowed to be downloaded, wp_check_filetype() already made it with get_allowed_mime_types().
	if ( ! $media_file_type['type'] || ! $media_file_type['ext'] ) {

		$notice = sprintf(
			'Error downloading media from the URL <code>%s</code>. Forbidden file extension <code>%s</code>.',
			$media_url,
			$media_url_extension
		);

		magic_im_add_notice( $notice );
		return false;
	}

	// Set existing attachment ID only if single attachment post without attached file was found.
	// Otherwise, it might be unclear which attachment to update.
	if ( count( $attachments ) === 1 ) {
		$existing_attachment_id = $attachments[0]['ID'];
	}

	// Download media from the provided URL.
	return magic_im_download_media_from_url( $media_url, $media_post_data, $existing_attachment_id ?? 0 );
}

/**
 * Downloads media from the provided URL inserting it as new attachment post or updating as existing.
 *
 * @param string $url Media URL.
 * @param array  $media_post_data Predefined attachment post data.
 * @param int    $attachment_id Existing attachment ID to update.
 * @return int|false Downloaded media ID on success, false otherwise.
 *
 * @throws Exception Downloading media file errors.
 */
function magic_im_download_media_from_url( $url, $media_post_data = array(), $attachment_id = 0 ) {

	try {

		// Get the file name with extension.
		$file_name = basename( $url );

		// Set the upload directory.
		$upload_dir  = wp_upload_dir();
		$upload_path = $upload_dir['path'] . '/' . $file_name;

		// Make HTTP request with the URL.
		$response = wp_remote_get( $url );

		// Check if the request was successful.
		if ( is_wp_error( $response ) || wp_remote_retrieve_response_code( $response ) !== 200 ) {
			throw new Exception( is_wp_error( $response ) ? $response->get_error_message() : '' );
		}

		$response_headers = wp_remote_retrieve_headers( $response );

		if ( ! isset( $response_headers['content-type'] ) ) {
			throw new Exception();
		}

		// Save the file to the upload directory.
		// phpcs:ignore WordPress.WP.AlternativeFunctions
		$file_saved = file_put_contents( $upload_path, wp_remote_retrieve_body( $response ) );

		// Check if the file was successfully saved.
		if ( false === $file_saved ) {
			throw new Exception( 'Error on calling <code>file_put_contents</code>.' );
		}

		// Update existing attachment.
		if ( $attachment_id ) {

			update_attached_file( $attachment_id, $upload_path );
			$media_file_id = $attachment_id;

			// Insert new attachment.
		} else {

			// Prepare the attachment post data.
			$media_post_data = array_merge(
				array(
					'post_title'     => $media_post_data['post_title'] ?? pathinfo( $file_name, PATHINFO_FILENAME ),
					'post_type'      => 'attachment',
					'post_status'    => 'inherit',
					'post_mime_type' => mime_content_type( $upload_path ),
				),
				$media_post_data
			);

			$media_file_id = wp_insert_attachment( $media_post_data, $upload_path );

			if ( is_wp_error( $media_file_id ) ) {
				throw new Exception( $media_file_id->get_error_message() );
			}
		}

		// Generate and update attachment metadata.

		require_once ABSPATH . 'wp-admin/includes/image.php';

		$attachment_metadata = wp_generate_attachment_metadata( $media_file_id, $upload_path );

		// Manually set missing image meta (for SVG in some scenarios).
		if ( wp_attachment_is_image( $media_file_id ) ) {

			$attachment_metadata['file']   ??= str_replace( $upload_dir['basedir'] . '/', '', $upload_path );
			$attachment_metadata['width']  ??= 0;
			$attachment_metadata['height'] ??= 0;
		}

		wp_update_attachment_metadata( $media_file_id, $attachment_metadata );
		do_action( 'magic_im_media_downloaded', $media_file_id );

	} catch ( Exception $e ) {
		$error = sprintf( 'Error downloading media from the URL <code>%s</code>. %s', $url, $e->getMessage() );
		magic_im_add_notice( $error );
		$media_file_id = false;
	}

	return $media_file_id;
}

/**
 * Checks if provided date is valid to store it in database.
 *
 * @param string $date_str Date string.
 * @param string $date_key Date key.
 * @return string
 */
function magic_im_validate_date_str( $date_str, $date_key ) {

	$error = '';

	if ( ! $date_str ) {
		return $error;
	}

	// Check the date format. Should be as following: '2023-10-31 07:14:58'.
	if ( ! is_string( $date_str )
		|| ! preg_match( '/^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$/', $date_str )
		|| ! wp_resolve_post_date( $date_str ) ) {

			$error = sprintf(
				'Invalid date format <code>%s</code> in the column <code>%s</code>. Valid format is <code>Y-m-d H:i:s</code>. Editors like Excel might distort dates.',
				$date_str,
				$date_key,
			);
	}

	return $error;
}

/**
 * Adjusts export content for the editor fields (Gutenberg or WYSIWYG).
 *
 * @param string $value Editor content.
 * @return string
 */
function magic_ex_adjust_editor_value( $value ) {

	if ( ! $value || ! is_string( $value ) ) {
		return $value;
	}

	// Gutenberg blocks.
	if ( str_contains( $value, '<!-- wp:' ) ) {

		// Convert block pattern IDs into names.
		$value = preg_replace_callback(
			'/<!--\s*wp:block\s*{"ref":(\d+)}\s*\/-->/',
			function ( $matches ) {

				$ref_id   = $matches[1];
				$ref_name = get_post_field( 'post_name', $ref_id );

				// Replace the reference ID with name.
				return sprintf( '<!-- wp:block {"ref":%s} /-->', $ref_name );
			},
			$value
		);

		$value = apply_filters( 'magic_ex_adjust_editor_value_gutenberg', $value );
	}

	return $value;
}

/**
 * Adjusts import content for the editor fields (Gutenberg or WYSIWYG).
 *
 * @param string $value Editor content.
 * @return string
 */
function magic_im_adjust_editor_value( $value ) {

	if ( ! $value || ! is_string( $value ) ) {
		return $value;
	}

	// Gutenberg blocks.
	if ( str_contains( $value, '<!-- wp:' ) ) {

		$new_content = $value;

		// Convert block patterns names into IDs.
		$new_content = preg_replace_callback(
			'/<!--\s*wp:block\s*{"ref":([a-zA-Z0-9-_]+)}\s*\/-->/',
			function ( $matches ) {

				$ref_name = $matches[1];

				// Get pattern ID by its 'post_name'.
				$ref_id = get_posts(
					array(
						'post_type'      => 'wp_block',
						'name'           => $ref_name,
						'posts_per_page' => 1,
						'fields'         => 'ids',
					)
				);

				// Replace the reference ID with name.
				return sprintf( '<!-- wp:block {"ref":%d} /-->', reset( $ref_id ) );
			},
			$new_content
		);

		// Replace images.

		$pattern = '/<!-- wp:(image|media-text) (.*?) -->(.*?)<!-- \/wp:(image|media-text) -->/s';
		preg_match_all( $pattern, $value, $matches, PREG_SET_ORDER );

		foreach ( $matches as $match ) {

			$src_pattern = '/<img\s+([^>]*?)src="(.*?)"/i'; // Match image src attribute.
			preg_match( $src_pattern, $match[3], $src_matches );
			$old_src = $src_matches[2] ?? false;

			// Get new image id.
			$new_image_id = magic_im_media_url_to_local_id( $old_src );

			if ( $new_image_id && is_numeric( $new_image_id ) ) {

				// Extract JSON data.
				$json_data = json_decode( $match[2], true );

				// Get new image src.
				$new_src = wp_get_attachment_image_url( $new_image_id, $json_data['sizeSlug'] ?? 'full' );

				// Match wp-image-[id] class.
				$class_pattern = '/wp-image-(\d+)/';
				preg_match( $class_pattern, $match[3], $class_matches );
				$old_class = $class_matches[0];

				// Modify the class with new image id.
				$new_class = 'wp-image-' . $new_image_id;

				// Block 'Image'.
				if ( isset( $json_data['id'] ) ) {

					$new_content = str_replace(
						'"id":' . $json_data['id'],
						'"id":' . $new_image_id,
						$new_content
					);
				}

				// Block 'Media & Text'.
				if ( isset( $json_data['mediaId'] ) ) {

					$new_content = str_replace(
						'"mediaId":' . $json_data['mediaId'],
						'"mediaId":' . $new_image_id,
						$new_content
					);
				}

				$new_content = str_replace(
					'src="' . $old_src . '"',
					'src="' . $new_src . '"',
					$new_content
				);

				$new_content = str_replace( $old_class, $new_class, $new_content );
			}
		}

		$new_content = apply_filters( 'magic_im_adjust_editor_value_gutenberg', $new_content );

		// Classic editor (WYSIWYG).
	} else {

		$new_content = preg_replace_callback(
			'/<img\s+([^>]*?)src="(.*?)"/i', // Match image with src attribute.
			function ( $matches ) {

				$image_size = '1536x1536'; // Fallback image size.

				// Get image size.
				if ( preg_match( '/size-([^\s]+)/', $matches[1], $size_matches ) ) {
					$image_size = $size_matches[1];
				}

				// Get new image ID from the old image src.
				$old_image_src = $matches[2] ?? false;
				$new_image_id  = magic_im_media_url_to_local_id( $old_image_src );

				// Set new image src and class containing new image id.
				if ( $new_image_id && is_numeric( $new_image_id ) ) {

					$new_image_src = wp_get_attachment_image_url( $new_image_id, $image_size );
					$matches[1]    = preg_replace( '/wp-image-\d+/', 'wp-image-' . $new_image_id, $matches[1] );

				} else {
					$new_image_src = $old_image_src;
				}

				// Reconstruct img tag with new 'src' and 'class'.
				return '<img ' . $matches[1] . 'src="' . $new_image_src . '"';
			},
			$value
		);
	}

	// Replace remaining media files URLs into local URLs.
	$new_content = preg_replace_callback(
		'/https?:\/\/[^\s"\'<>]+/i', // Matches URLs starting with http or https.
		function ( $matches ) {
			$old_url = $matches[0];
			return magic_im_media_url_to_local_url( $old_url );
		},
		$new_content
	);

	// Replace remaining old domain into current domain.
	$new_content = magic_im_replace_domain( $new_content );

	return $new_content;
}

/**
 * Replaces old domain into current in the provided data.
 *
 * @param mixed $data Data.
 * @return mixed
 */
function magic_im_replace_domain( $data ) {

	$old_domain     = magic_im_get_file_params( 'old-domain' );
	$current_domain = magic_im_get_current_domain();

	if ( $old_domain === $current_domain ) {
		return $data;
	}

	if ( is_array( $data ) ) {

		foreach ( $data as $key => $value ) {
			$data[ $key ] = magic_im_replace_domain( $value );
		}
	} elseif ( is_string( $data ) ) {

		$data = str_replace( $old_domain, $current_domain, $data, $replaced_cnt );

		if ( $replaced_cnt ) {

			$data = is_ssl()
				? str_replace( "http://{$current_domain}", "https://{$current_domain}", $data )
				: str_replace( "https://{$current_domain}", "http://{$current_domain}", $data );
		}
	}

	return $data;
}

/**
 * Builds key based on the provided key and its group.
 *
 * @param string $key Data key.
 * @param string $key_group Its group.
 * @return string
 */
function magic_ex_im_build_key( $key, $key_group ) {
	return $key_group . ':' . $key;
}

/**
 * Parses provided key into group and key.
 *
 * @param string $key Key to parse.
 * @param string $parse_as Parse into: 'key' or its 'group'. Default to both.
 * @return string|array Key or its group. Both as associative array if $parse_as not provided.
 */
function magic_ex_im_parse_key( $key, $parse_as = '' ) {

	$result = array(
		'group' => '',
		'key'   => $key,
	);

	$exploded_key = explode( ':', $key, 2 );

	if ( count( $exploded_key ) > 1 ) {
		$result['group'] = $exploded_key[0];
		$result['key']   = $exploded_key[1];
	}

	return $parse_as ? $result[ $parse_as ] : $result;
}
