<?php
/**
 * Class Magic_EX_IM_Type_ACF_Options_Pages.
 *
 * @package Magic_Export_Import\ACF_Options_Pages
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for ACF Options Pages.
 */
final class Magic_EX_IM_Type_ACF_Options_Pages extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'acf_options_pages';
	const KEY_GROUP  = 'ACF-page';

	/**
	 * Current magic item ('acf_options_pages' by default).
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE, 'ACF Options Pages' );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {
		return array( 'acf_options_pages' => array( 'label' => 'ACF Options Pages' ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$page_slug = $item_data[ self::KEY_GROUP ]['page_slug'] ?? '';

		return $page_slug;
	}

	/**
	 * Gets all registered ACF option pages.
	 *
	 * @param string $acf_page_slug Particular ACF Options Page slug to get.
	 * @return array
	 */
	public function get_items( $acf_page_slug = false ) {

		$acf_options_pages = function_exists( 'acf_get_options_pages' ) ? acf_get_options_pages() : array();

		return $acf_page_slug ? ( $acf_options_pages[ $acf_page_slug ] ?? array() ) : $acf_options_pages;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function allow_to_create_items() {
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_mandatory_keys() {
		return array( 'page_title', 'page_slug', 'post_id' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {
		return $posted_data['magic-ex-items'] ?? array_keys( $this->get_items() );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $acf_page_slug Item ID.
	 */
	protected function get_key_data( $group, $key, $acf_page_slug ) {

		$key_data = '';

		if ( self::KEY_GROUP === $group ) {

			switch ( $key ) {

				case 'page_title':
					$key_data = $this->get_items( $acf_page_slug )['page_title'] ?? '';
					break;

				case 'post_id':
					$key_data = $this->get_items( $acf_page_slug )['post_id'] ?? '';
					break;

				case 'page_slug':
					$key_data = $acf_page_slug;
					break;
			}
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.

		$this->magic_item_labels                = new stdClass();
		$this->magic_item_labels->name          = __( 'ACF Options Pages', 'magic-export-import' );
		$this->magic_item_labels->singular_name = __( 'ACF Options Page', 'magic-export-import' );
		$this->magic_item_labels->all_items     = __( 'All ACF Options Pages', 'magic-export-import' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		foreach ( $this->get_items() as $acf_page_slug => $acf_page ) {

			if ( ! $args['s'] ||
				( false !== stripos( $acf_page['page_title'], $args['s'] )
				|| false !== stripos( $acf_page['post_id'], $args['s'] ) )
			) {

				$result['results'][] = array(
					'id'   => $acf_page_slug,
					'text' => $acf_page['page_title'] . ' <code>' . $acf_page['post_id'] . '</code>',
				);
			}
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $page_slugs Specific ACF page slugs to get export keys for.
	 */
	protected function get_export_keys( $page_slugs = array() ) {
		return $this->get_acf_pages_export_keys( $page_slugs );
	}

	/**
	 * Gets ACF Options Pages keys for the export.
	 *
	 * @param string|string[] $page_slugs ACF page slugs to retrieve export keys for.
	 * @return array
	 */
	public function get_acf_pages_export_keys( $page_slugs = array() ) {

		$page_slugs  = (array) $page_slugs;
		$export_keys = $this->get_mandatory_keys();

		foreach ( array_keys( $this->get_items() ) as $acf_page_slug ) {

			if ( $page_slugs && ! in_array( $acf_page_slug, $page_slugs, true ) ) {
				continue;
			}

			$acf_field_groups = acf_get_field_groups( array( 'options_page' => $acf_page_slug ) );

			foreach ( $acf_field_groups as $acf_field_group ) {
				foreach ( acf_get_fields( $acf_field_group['key'] ) as $acf_field ) {
					$export_keys[] = $acf_field['name'];
				}
			}
		}

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key User key.
	 */
	protected function get_key_group( $key ) {
		return in_array( $key, $this->get_mandatory_keys(), true ) ? self::KEY_GROUP : 'ACF';
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {
		return $this->get_items( $item_data[ self::KEY_GROUP ]['page_slug'] ) ? $item_data[ self::KEY_GROUP ]['page_slug'] : false;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 *
	 * @see Magic_EX_IM_Adapter_ACF
	 */
	protected function update_existing_item( $item_data, $item_id ) {
		return $item_id; // Do nothing, everything happens within ACF adapter class.
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {
		return false; // Do nothing. Creating ACF Options pages is not allowed.
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 *
	 * @see Magic_EX_IM_Adapter_ACF
	 */
	protected function update_importing_item( $item_data, $item_id ) {
		// Do nothing.
	}
}

if ( class_exists( 'ACF' ) ) {
	Magic_EX_IM_Type_ACF_Options_Pages::instance();
}
