<?php
/**
 * Class Magic_EX_IM_Type_Comments.
 *
 * @package Magic_Export_Import\Comments
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for comments.
 */
final class Magic_EX_IM_Type_Comments extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'comments';

	/**
	 * Current magic item ('comments' by default).
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {
		return array( 'comments' => array( 'label' => 'comments' ) );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_mandatory_keys() {
		return array( 'comment_post_ID', 'comment_content', 'comment_type', 'comment_date', 'comment_author' );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.

		$this->magic_item_labels                = new stdClass();
		$this->magic_item_labels->name          = __( 'Comments', 'magic-export-import' );
		$this->magic_item_labels->singular_name = __( 'Comment', 'magic-export-import' );
		$this->magic_item_labels->all_items     = __( 'All Comments', 'magic-export-import' );

		add_filter( 'magic_ex_individual_items_title', array( $this, 'magic_ex_individual_items_title' ) );
		add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'adjust_value' ), 5, 3 );

		add_action( 'magic_im_item_processed', array( $this, 'magic_im_item_processed' ) );
	}

	/**
	 * Processes action hook 'magic_im_item_processed'.
	 *
	 * @param array $item_data Processed item data.
	 * @see WC_Comments::clear_transients()
	 */
	public function magic_im_item_processed( $item_data ) {
		do_action( 'wp_update_comment_count', $item_data['comment']['comment_post_ID'] );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$comment_type = $item_data['comment']['comment_type'] ?? '';

		return $comment_type;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {

		// Initial comment query arguments.
		$args = array(
			'post__in' => $posted_data['magic-ex-items'] ?? array(),
			'fields'   => 'ids',
			'status'   => array_keys( get_comment_statuses() ),
		);

		// Advanced filters.
		if ( isset( $posted_data['magic-ex-advanced'] ) ) {

			// Query by comment type.
			if ( isset( $posted_data['magic-ex-by-comment-types'] ) ) {
				$args['type'] = $posted_data['magic-ex-comment-types'];
			}

			// Query by comment status.
			if ( isset( $posted_data['magic-ex-by-comment-statuses'] ) && $posted_data['magic-ex-comment-statuses'] ) {
				$args['status'] = $posted_data['magic-ex-comment-statuses'];
			}

			// Query by comment date.
			if ( isset( $posted_data['magic-ex-by-date'] ) ) {

				$args['date_query'] = array(
					'after'     => $posted_data['magic-ex-date-from'] ?? '',
					'before'    => $posted_data['magic-ex-date-to'] ?? '',
					'inclusive' => true,
				);
			}
		}

		$args = apply_filters( 'magic_ex_get_export_items_args', $args, $posted_data );

		$export_comments_query = new WP_Comment_Query( $args );
		return $export_comments_query->get_comments();
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 */
	protected function get_key_data( $group, $key, $item_id ) {

		$key_data = '';

		switch ( $group ) {

			case 'comment':
				$key_data = get_comment( $item_id )->{$key};
				break;

			case 'meta':
				$key_data = get_comment_meta( $item_id, $key, true );
				break;
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Key of the provided value.
	 * @param string $key_group Group name $key belongs to.
	 */
	public function adjust_value( $value, $key, $key_group ) {

		if ( ! $value ) {
			return $value;
		}

		// Adjust value during export.
		if ( magic_ex_im_is_export() ) {

			if ( 'comment' === $key_group ) {

				switch ( $key ) {

					// Convert post ID into post_type:post_name.
					case 'comment_post_ID':
						$comment_post = get_post( $value );
						$value        = $comment_post ? ( get_post_type( $value ) . ':' . $comment_post->post_name ) : '';
						break;

					// Convert comment author ID into login.
					case 'user_id':
						$value = magic_ex_im_adjust_user_value( $value );
						break;

					// Convert comment ID into 'commnet type|comment author|comment date'.
					case 'comment_parent':
						$comment = get_comment( $value );
						$value   = sprintf( '%s|%s|%s', $comment->comment_type, $comment->comment_author, $comment->comment_date );
						break;
				}
			}
		}

		// Adjust value during import.
		if ( magic_ex_im_is_import() ) {

			if ( 'comment' === $key_group ) {

				switch ( $key ) {

					// Convert post_type:post_name into post ID.
					case 'comment_post_ID':
						list($post_type, $post_name) = explode( ':', $value );

						$value = magic_im_post_id_by_name( $post_name, $post_type );
						break;

					// Convert comment author login into ID.
					case 'user_id':
						$value = magic_ex_im_adjust_user_value( $value );
						break;

					// Convert comment 'commnet type|comment author|comment date' into ID.
					case 'comment_parent':
						list($c_type, $c_author, $c_date) = explode( '|', $value );

						$value = $this->get_existing_comment( $c_type, $c_author, $c_date );
						break;
				}
			}
		}

		return $value;
	}

	/**
	 * Processes filter hook 'magic_ex_individual_items_title'.
	 *
	 * @param string $title Current title.
	 * @return string
	 */
	public function magic_ex_individual_items_title( $title ) {

		$title = 'Individual Posts';
		return $title;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		// Query posts having comments considering offset.
		$posts_with_comments_query = new WP_Query(
			array(
				'post_type'      => 'any',
				'post_status'    => magic_ex_im_get_post_statuses(),
				'posts_per_page' => $args['items_per_request'],
				's'              => $args['s'],
				'search_columns' => array( 'post_title' ),
				'offset'         => $args['offset'],
				'sentence'       => true,
				'comment_count'  => array(
					'value'   => 1,
					'compare' => '>=',
				),
			)
		);

		$result['pagination']['more'] = $posts_with_comments_query->found_posts > $args['items_per_request'] + $args['offset'];

		foreach ( $posts_with_comments_query->posts as $post_obj ) {
			$text = $post_obj->post_title;

			if ( $post_obj->post_name ) {
				$text .= ' <code>' . $post_obj->post_name . '</code>';
			}

			$text .= ' <code>' . $post_obj->post_type . '</code>';

			$result['results'][] = array(
				'id'   => $post_obj->ID,
				'text' => $text,
			);
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $comment_ids Specific comment IDs to get export keys for.
	 */
	protected function get_export_keys( $comment_ids = array() ) {

		global $wpdb;

		// Get comment keys.
		$comment_keys = $wpdb->get_col( "DESC {$wpdb->comments}", 0 );

		// Exclude some comment keys from the export.
		$comment_keys = array_diff( $comment_keys, array( 'comment_ID' ) );

		// Get meta keys.

		$comment_ids_where = '';
		$query_params      = $comment_ids;

		if ( $comment_ids ) {
			$comment_ids_in    = implode( ', ', array_fill( 0, count( $comment_ids ), '%d' ) );
			$comment_ids_where = "AND {$wpdb->commentmeta}.comment_id IN ($comment_ids_in)";
		}

		$query = "SELECT DISTINCT(meta_key)
			FROM $wpdb->comments
			LEFT JOIN $wpdb->commentmeta
			ON $wpdb->comments.comment_ID = $wpdb->commentmeta.comment_id
			WHERE $wpdb->commentmeta.meta_key IS NOT NULL
			AND $wpdb->commentmeta.meta_key != ''
			$comment_ids_where";

		// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
		$meta_keys = $wpdb->get_col( $wpdb->prepare( $query, $query_params ) );

		// Combine all keys.
		$export_keys = array_values( array_merge( $comment_keys, $meta_keys ) );

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key comment key.
	 */
	protected function get_key_group( $key ) {

		global $wpdb;
		static $comment_keys = null;

		// Get comment keys.
		$comment_keys ??= $wpdb->get_col( "DESC {$wpdb->comments}" );

		if ( in_array( $key, $comment_keys, true ) ) {
			$group = 'comment';

		} else {
			$group = 'meta';
		}

		return apply_filters( 'magic_ex_im_get_key_group', $group, $key );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to validate.
	 */
	protected function validate_importing_item_data( $item_data ) {

		$errors = array();

		// Check for invalid dates.
		foreach ( array( 'comment_date', 'comment_date_gmt' ) as $date_key ) {

			$date_key = magic_ex_im_build_key( $date_key, $this->get_key_group( $date_key ) );
			$errors[] = magic_im_validate_date_str( $item_data[ $date_key ] ?? '', $date_key );
		}

		// Make sure related post for the importing comment exists.

		$comment_post_id_key = magic_ex_im_build_key( 'comment_post_ID', $this->get_key_group( 'comment_post_ID' ) );

		list($post_type, $post_name) = explode( ':', $item_data[ $comment_post_id_key ], 2 );

		if ( ! magic_im_post_id_by_name( $post_name, $post_type ) ) {
			$errors[] = sprintf( 'No <b>post</b> found for the %s.', magic_im_get_item_name() );
		}

		return $errors;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {

		// Identify existing importing comment.
		return $this->get_existing_comment(
			$item_data['comment']['comment_type'],
			$item_data['comment']['comment_author'],
			$item_data['comment']['comment_date']
		);
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 */
	protected function update_existing_item( $item_data, $item_id ) {

		$item_data['comment']['comment_ID'] = $item_id;

		$updated = wp_update_comment( $item_data['comment'] );
		return is_numeric( $updated ) ? $item_id : $updated;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {

		$comment_id = wp_insert_comment( $item_data['comment'] );

		return $comment_id;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	protected function update_importing_item( $item_data, $item_id ) {

		// Update comment meta.
		foreach ( $item_data['meta'] ?? array() as $meta_key => $value ) {

			// Skip meta if value is empty and meta doesn't exist (don't create new empty meta).
			if ( $value || is_numeric( $value ) || metadata_exists( 'comment', $item_id, $meta_key ) ) {
				update_comment_meta( $item_id, $meta_key, $value );
			}
		}
	}

	/**
	 * Finds existing comment by the provided data.
	 *
	 * @param string $c_type Comment type.
	 * @param string $c_author Comment author.
	 * @param string $c_date Comment date.
	 *
	 * @return int|false Comment ID or false if not found.
	 */
	protected function get_existing_comment( $c_type, $c_author, $c_date ) {

		global $wpdb;

		$existing_comment_id = $wpdb->get_var(
			$wpdb->prepare(
				"SELECT comment_ID from {$wpdb->prefix}comments
				WHERE comment_type = %s
				AND comment_author = %s
				AND comment_date = %s
				LIMIT 1",
				$c_type,
				$c_author,
				$c_date
			)
		);

		return $existing_comment_id ? intval( $existing_comment_id ) : false;
	}
}

Magic_EX_IM_Type_Comments::instance();
