<?php
/**
 * Class Magic_EX_IM_Type_Posts.
 *
 * @package Magic_Export_Import\Posts
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery, WordPress.DB.SlowDBQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for posts.
 */
final class Magic_EX_IM_Type_Posts extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'posts';

	/**
	 * Current post type.
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE );
	}

	/**
	 * Checks if current processing magic item is 'attachment' post type.
	 *
	 * @return bool
	 */
	private function is_attachment() {
		return 'attachment' === $this->magic_item;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {

		$allowed_post_types = array_diff_key(
			get_post_types( array( 'can_export' => true ), 'objects' ),
			array_flip(
				array(
					'custom_css',
					'wp_template',
					'custom_css',
					'wp_template',
					'wp_template_part',
					'wp_global_styles',
					'wp_navigation',
					'wp_font_family',
					'wp_font_face',
					'nav_menu_item',
					'patterns_ai_data',
				)
			)
		);

		return apply_filters( 'magic_ex_im_get_allowed_magic_items', $allowed_post_types );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @see Magic_EX_IM_Type_Posts::validate_importing_item_data()
	 */
	protected function get_mandatory_keys() {

		$mandatory_keys = array( 'post_type' );

		// Post name might be empty in drafts etc, post title also might be absent in posts.
		// But one of them should be present to process the import (refer to @see tag).
		if ( magic_ex_im_is_export() ) {
			$mandatory_keys = array_merge( $mandatory_keys, array( 'post_name', 'post_title' ) );
		}

		if ( $this->is_attachment() ) {
			$mandatory_keys[] = 'attachment_url';
		}

		return $mandatory_keys;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$post_name  = $item_data['post']['post_name'] ?? '';
		$post_title = $item_data['post']['post_title'] ?? '';

		return $post_name ? $post_name : $post_title;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {

		// Initial posts query arguments.
		$args = array(
			'post_type'      => $this->magic_item,
			'posts_per_page' => -1,
			'order'          => 'DESC',
			'post_status'    => magic_ex_im_get_post_statuses(),
			'post__in'       => $posted_data['magic-ex-items'] ?? array(),
			'fields'         => 'ids',
		);

		// Advanced filters.
		if ( isset( $posted_data['magic-ex-advanced'] ) ) {

			// Query by partial export.
			if ( isset( $posted_data['magic-ex-by-parts'] ) ) {

				$args['posts_per_page'] = $posted_data['magic-ex-cnt'] ?? $args['posts_per_page'];
				$args['offset']         = $posted_data['magic-ex-offset'] ?? 0;
			}

			// Query by date.
			if ( isset( $posted_data['magic-ex-by-date'] ) ) {

				$args['date_query'] = array(
					'after'     => $posted_data['magic-ex-date-from'] ?? '',
					'before'    => $posted_data['magic-ex-date-to'] ?? '',
					'inclusive' => true,
				);
			}

			// Query by post status.
			if ( isset( $posted_data['magic-ex-by-statuses'] ) ) {
				$args['post_status'] = $posted_data['magic-ex-statuses'] ?? $args['post_status'];
			}

			// Query by taxonomies.
			if ( isset( $posted_data['magic-ex-by-taxonomy'] ) ) {
				$args['tax_query'] = array( 'relation' => 'AND' );

				foreach ( $posted_data['magic-ex-taxonomy'] ?? array() as $tax_slug => $term_ids ) {

					$args['tax_query'][] = array(
						'taxonomy' => $tax_slug,
						'terms'    => $term_ids,
						'operator' => 'IN',
					);
				}
			}
		}

		$args = apply_filters( 'magic_ex_get_export_items_args', $args, $posted_data );

		$export_posts_query = new WP_Query( $args );

		if ( ! $include_related_items ) {
			return $export_posts_query->posts;
		}

		return apply_filters( 'magic_ex_get_items', $export_posts_query->posts );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $items_export_data Exporting items data.
	 */
	protected function get_export_admin_notice( $items_export_data ) {

		// Calculate extported post types and form appropriate export notice.
		// Might be muptiple post types since some of them are related to each other (e.g. 'product_variation' to 'product').

		$export_post_types_cnts = array();
		$export_notice          = '';

		// Skip exporting items keys (file header).
		array_shift( $items_export_data );

		// Calculate processed post types.
		foreach ( $items_export_data as $item_export_data ) {
			$post_type = $item_export_data['post_type'];

			$export_post_types_cnts[ $post_type ] ??= 0;
			++$export_post_types_cnts[ $post_type ];
		}

		// Form the export notice.
		foreach ( $export_post_types_cnts as $post_type => $cnt ) {

			if ( $this->magic_item === $post_type ) {
				// Processing magic item always goes first.

				$export_notice .= sprintf(
					'Export file successfully generated with <b>%d</b> %s',
					$cnt,
					$cnt > 1 ? $this->magic_item_labels->name : $this->magic_item_labels->singular_name
				);

			} else {
				$post_type_labels = get_post_type_object( $post_type )->labels;

				$export_notice .= sprintf(
					' and <b>%d</b> related %s.',
					$cnt,
					$cnt > 1 ? $post_type_labels->name : $post_type_labels->singular_name
				);
			}
		}

		return $export_notice;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 */
	protected function get_key_data( $group, $key, $item_id ) {

		$key_data = '';

		switch ( $group ) {

			case 'attachment':
				if ( 'attachment_url' === $key ) {
					$key_data = wp_get_attachment_url( $item_id );
				}
				break;

			case 'post':
				$key_data = get_post_field( $key, $item_id );
				break;

			case 'meta':
				$key_data = get_post_meta( $item_id, $key, true );
				break;

			case 'taxonomy':
				$term_slugs = wp_get_post_terms( $item_id, $key, array( 'fields' => 'slugs' ) );
				$key_data   = implode( ',', $term_slugs );
				break;
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.
		$this->magic_item_labels = get_post_type_object( $this->magic_item )->labels;

		add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'adjust_value' ), 5, 3 );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Key of the provided value.
	 * @param string $key_group Group name $key belongs to.
	 */
	public function adjust_value( $value, $key, $key_group ) {

		if ( ! $value ) {
			return $value;
		}

		// Adjust value during export.
		if ( magic_ex_im_is_export() ) {

			if ( 'post' === $key_group ) {

				switch ( $key ) {

					// Convert post author ID into login.
					case 'post_author':
						$value = magic_ex_im_adjust_user_value( $value );
						break;

					// Convert post ID into name.
					case 'post_parent':
						$value = get_post_field( 'post_name', $value );
						break;

					// Adjust post content.
					case 'post_content':
						$value = magic_ex_adjust_editor_value( $value );
						break;
				}
			}
		}

		// Adjust value during import.
		if ( magic_ex_im_is_import() ) {

			if ( 'post' === $key_group ) {

				switch ( $key ) {

					// Convert post author login into ID.
					case 'post_author':
						$value = magic_ex_im_adjust_user_value( $value );
						break;

					// Convert post parent name into ID.
					case 'post_parent':
						$value = magic_im_post_id_by_name( $value, $this->magic_item );
						break;

					// Adjust post content.
					case 'post_content':
						$value = magic_im_adjust_editor_value( $value );
						break;
				}
			}

			// Convert term names into IDs.
			if ( 'taxonomy' === $key_group && is_string( $value ) ) {

				$term_ids       = array();
				$term_slugs_arr = array_map( 'trim', explode( ',', $value ) );

				foreach ( $term_slugs_arr as $term_slug ) {

					$term_id    = get_term_by( 'slug', $term_slug, $key )->term_id ?? '';
					$term_ids[] = apply_filters( 'magic_im_item_data_taxonomy', $term_id, $key );
				}

				$value = array_filter( $term_ids );
			}
		}

		return $value;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		// Query posts considering offset.
		$posts_query = new WP_Query(
			array(
				'post_type'      => $args['magic_item'],
				'post_status'    => magic_ex_im_get_post_statuses(),
				'posts_per_page' => $args['items_per_request'],
				's'              => $args['s'],
				'search_columns' => array( 'post_title' ),
				'offset'         => $args['offset'],
				'sentence'       => true,
			)
		);

		$result['pagination']['more'] = $posts_query->found_posts > $args['items_per_request'] + $args['offset'];

		foreach ( $posts_query->posts as $post_obj ) {

			$text       = $post_obj->post_title;
			$extra_text = $post_obj->post_mime_type ? $post_obj->post_mime_type : $post_obj->post_name;

			if ( $extra_text ) {
				$text .= ' <code>' . $extra_text . '</code>';
			}

			$result['results'][] = array(
				'id'   => $post_obj->ID,
				'text' => $text,
			);
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $post_ids Specific post IDs to get export keys for.
	 */
	protected function get_export_keys( $post_ids = array() ) {

		global $wpdb;

		// Set post keys.
		$post_keys = $wpdb->get_col( "DESC {$wpdb->posts}" );

		// Exclude some post keys from the export.

		$post_keys_to_exclude = array( 'ID', 'guid', 'comment_count', 'post_content_filtered' );

		if ( $this->is_attachment() ) {

			$post_keys_to_exclude = array_merge(
				$post_keys_to_exclude,
				array( 'post_date', 'post_date_gmt', 'post_mime_type', 'post_modified', 'post_modified_gmt' )
			);
		}

		$post_keys = array_diff( $post_keys, $post_keys_to_exclude );

		// Get meta keys for the export.

		$post_types    = apply_filters( 'magic_ex_im_processing_post_types', array( $this->magic_item ) );
		$post_types_in = implode( ', ', array_fill( 0, count( $post_types ), '%s' ) );

		$post_ids_where = '';

		if ( $post_ids ) {
			$post_ids_in    = implode( ', ', array_fill( 0, count( $post_ids ), '%d' ) );
			$post_ids_where = "AND $wpdb->posts.ID IN ($post_ids_in)";
		}

		$query = "SELECT DISTINCT(meta_key)
			FROM $wpdb->posts
			LEFT JOIN $wpdb->postmeta
			ON $wpdb->posts.ID = $wpdb->postmeta.post_id
			WHERE $wpdb->posts.post_type IN ($post_types_in)
			$post_ids_where
			AND $wpdb->postmeta.meta_key != ''";

		$query_params = $post_types;

		if ( $post_ids ) {
			$query_params = array_merge( $query_params, $post_ids );
		}

		// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
		$meta_keys = $wpdb->get_col( $wpdb->prepare( $query, $query_params ) );

		// Exclude some meta keys from the export.

		$meta_keys = array_diff(
			$meta_keys,
			array(
				'_edit_last',
				'_edit_lock',
				'_wp_old_slug',
				'_wp_old_date',
				'_post_version',
				'_wp_attached_file',
				'_wp_attachment_metadata',
			)
		);

		foreach ( $meta_keys as $index => $meta_key ) {
			if ( str_starts_with( $meta_key, '_oembed_' ) ) {
				unset( $meta_keys[ $index ] );
			}
		}

		// Set taxonomies keys.

		$taxonomy_keys = get_object_taxonomies( $this->magic_item );

		if ( $post_ids ) {

			$post_ids_terms = wp_get_object_terms( $post_ids, $taxonomy_keys );
			$taxonomy_keys  = array();

			foreach ( $post_ids_terms as $term_obj ) {
				$taxonomy_keys[] = $term_obj->taxonomy;
			}
		}

		// Set custom keys.

		$magic_keys = array();

		if ( $this->is_attachment() ) {
			$magic_keys[] = 'attachment_url';
		}

		// Combine all export keys.
		$export_keys = array_values( array_merge( $magic_keys, $post_keys, $meta_keys, $taxonomy_keys ) );

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key Post key.
	 */
	protected function get_key_group( $key ) {

		global $wpdb;
		static $post_keys = null;

		$post_keys    ??= $wpdb->get_col( "DESC {$wpdb->posts}" );
		$taxonomy_names = get_object_taxonomies( $this->magic_item );

		if ( $this->is_attachment() && 'attachment_url' === $key ) {
			$group = 'attachment';

		} elseif ( in_array( $key, $taxonomy_names, true ) ) {
			$group = 'taxonomy';

		} elseif ( in_array( $key, $post_keys, true ) ) {
			$group = 'post';

		} else {
			$group = 'meta';
		}

		return apply_filters( 'magic_ex_im_get_key_group', $group, $key );
	}

	/**
	 * {@inheritDoc}.
	 */
	protected function allow_to_update_items() {
		return 'shop_order' !== $this->magic_item;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to validate.
	 */
	protected function validate_importing_item_data( $item_data ) {

		$errors = array();

		$post_name_key  = magic_ex_im_build_key( 'post_name', $this->get_key_group( 'post_name' ) );
		$post_title_key = magic_ex_im_build_key( 'post_title', $this->get_key_group( 'post_title' ) );

		// Check if post name or post title is present and not empty.
		if ( ! ( $item_data[ $post_name_key ] ?? '' ) && ! ( $item_data[ $post_title_key ] ?? '' ) ) {

			$errors[] = 'Both <code>post_name</code> and <code>post_title</code> are missing. At least one should be present to process import.';
		}

		$post_type_key = magic_ex_im_build_key( 'post_type', $this->get_key_group( 'post_type' ) );
		$post_types    = apply_filters( 'magic_ex_im_processing_post_types', array( $this->magic_item ) );

		// Check if importing post type is valid.
		if ( isset( $item_data[ $post_type_key ] ) && ! in_array( $item_data[ $post_type_key ], $post_types, true ) ) {

			$errors[] = sprintf(
				'Incorrect importing <b>post type</b> <code>%s</code>, current <b>post type</b> is <code>%s</code>.',
				$item_data[ $post_type_key ] ? $item_data[ $post_type_key ] : 'undefined',
				$this->magic_item,
			);
		}

		$post_status_key = magic_ex_im_build_key( 'post_status', $this->get_key_group( 'post_status' ) );

		// Check if post status is valid.
		if ( $item_data[ $post_status_key ] ?? '' ) {

			$item_post_status = $item_data[ $post_status_key ];

			// Check if post status is valid for attachment post type.
			if ( $this->is_attachment() && 'inherit' !== $item_post_status ) {

				$errors[] = sprintf(
					'Invalid attachment <b>post status</b> <code>%s</code>. Required <b>post status</b> is <code>inherit</code>.',
					$item_post_status,
				);

				// Check if post status is valid for other post types except attachment.
			} elseif ( ! $this->is_attachment() && ! in_array( $item_post_status, magic_ex_im_get_post_statuses(), true ) ) {
				$errors[] = sprintf( 'Non-existing <b>post status</b> <code>%s</code>.', $item_post_status );
			}
		}

		// Check for invalid dates.
		foreach ( array( 'post_date', 'post_date_gmt', 'post_modified', 'post_modified_gmt' ) as $date_key ) {

			$date_key = magic_ex_im_build_key( $date_key, $this->get_key_group( $date_key ) );
			$errors[] = magic_im_validate_date_str( $item_data[ $date_key ] ?? '', $date_key );
		}

		return $errors;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {

		// Identify existing attachment post.
		if ( $this->is_attachment() ) {

			$attachment_id = magic_im_media_url_to_local_id( $item_data['attachment']['attachment_url'], array(), true );
			return is_numeric( $attachment_id ) ? $attachment_id : false;
		}

		// Identify existing post by post name or post title.

		$args = array(
			'post_type'      => $item_data['post']['post_type'],
			'posts_per_page' => 1,
			'post_status'    => magic_ex_im_get_post_statuses(),
			'no_found_rows'  => true,
			'fields'         => 'ids',
		);

		$post_name = $item_data['post']['post_name'] ?? '';

		if ( $post_name ) {
			$args['post_name__in'] = magic_ex_im_get_post_name__in( $post_name );
		} else {
			$args['title'] = $item_data['post']['post_title'];
		}

		$existing_post_query = new WP_Query(
			apply_filters( 'magic_im_existing_item_query_args', $args, $item_data )
		);

		return reset( $existing_post_query->posts );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 */
	protected function update_existing_item( $item_data, $item_id ) {

		$item_data['post']['ID'] = $item_id;

		return wp_update_post( $item_data['post'] );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {

		if ( $this->is_attachment() ) {

			$attachment_id = magic_im_media_url_to_local_id( $item_data['attachment']['attachment_url'], $item_data['post'] );
			return is_numeric( $attachment_id ) ? $attachment_id : false;
		}

		return wp_insert_post( $item_data['post'] );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	protected function update_importing_item( $item_data, $item_id ) {

		// Update post meta.
		foreach ( $item_data['meta'] ?? array() as $meta_key => $value ) {

			// Skip meta if value is empty and meta doesn't exist (don't create new empty meta).
			if ( $value || is_numeric( $value ) || metadata_exists( 'post', $item_id, $meta_key ) ) {
				update_post_meta( $item_id, $meta_key, $value );
			}
		}

		// Update post terms.
		foreach ( $item_data['taxonomy'] ?? array() as $tax_slug => $term_ids ) {
			wp_set_post_terms( $item_id, $term_ids, $tax_slug );
		}
	}
}

Magic_EX_IM_Type_Posts::instance();
