<?php
/**
 * Class Magic_EX_IM_Type_Shop_Orders.
 *
 * @package Magic_Export_Import\Shop_Orders
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for WooCommerce shop orders.
 */
final class Magic_EX_IM_Type_Shop_Orders extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'shop_orders';

	/**
	 * Current magic item ('shop_orders' by default).
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE, 'Shop Orders' );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {
		return array( 'shop_orders' => array( 'label' => __( 'Shop Orders', 'magic-export-import' ) ) );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_mandatory_keys() {
		return $this->get_custom_tables_names();
	}

	/**
	 * Gets WooCommerce custom tables names for shop orders.
	 *
	 * @return array
	 */
	protected function get_custom_tables_names() {
		return array( 'wc_orders', 'wc_order_addresses', 'wc_order_operational_data' );
	}

	/**
	 * Gets all order statuses.
	 *
	 * @return array
	 */
	protected function get_order_statuses() {
		$all_statuses = array_keys( wc_get_order_statuses() );
		array_push( $all_statuses, 'trash' );

		return $all_statuses;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$billing_email = $item_data['wc_order']['wc_orders']['billing_email'] ?? '';

		return $billing_email;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {

		$args = array(
			'type'     => 'shop_order',
			'limit'    => -1,
			'status'   => $this->get_order_statuses(),
			'post__in' => $posted_data['magic-ex-items'] ?? array(),
			'return'   => 'ids',
			'parent'   => 0, // Exclude refunds.
		);

		// Advanced filters.
		if ( isset( $posted_data['magic-ex-advanced'] ) ) {

			// Query by partial export.
			if ( isset( $posted_data['magic-ex-by-parts'] ) ) {

				$args['limit']  = $posted_data['magic-ex-cnt'] ?? $args['limit'];
				$args['offset'] = $posted_data['magic-ex-offset'] ?? 0;
			}

			// Query by date.
			if ( isset( $posted_data['magic-ex-by-date'] ) ) {
				$args['date_created'] = $posted_data['magic-ex-date-from'] . '...' . $posted_data['magic-ex-date-to'];
			}

			// Query by order status.
			if ( isset( $posted_data['magic-ex-by-statuses'] ) ) {
				$args['status'] = $posted_data['magic-ex-statuses'] ?? $args['status'];
			}
		}

		$args = apply_filters( 'magic_ex_get_export_items_args', $args, $posted_data );

		$export_orders_query = new WC_Order_Query( $args );
		$export_order_ids    = $export_orders_query->get_orders();

		// Include related refunds.
		if ( $include_related_items ) {
			foreach ( $export_order_ids as $order_id ) {

				$refunds_query = new WC_Order_Query(
					array(
						'parent' => $order_id,
						'type'   => 'shop_order_refund',
						'limit'  => -1,
						'return' => 'ids',
					)
				);

				$export_order_ids = array_merge( $export_order_ids, $refunds_query->get_orders() );
			}
		}

		return array_unique( $export_order_ids );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $items_export_data Exporting items data.
	 */
	protected function get_export_admin_notice( $items_export_data ) {
		$refunds_cnt = 0;

		// Skip exporting items keys (file header).
		array_shift( $items_export_data );

		// Count refunds.
		foreach ( $items_export_data as $item_data ) {

			$wc_orders_data = maybe_unserialize( $item_data['wc_orders'] ?? '' );
			$order_type     = $wc_orders_data['type'] ?? '';

			if ( 'shop_order_refund' === $order_type ) {
				++$refunds_cnt;
			}
		}

		$orders_cnt = count( $items_export_data ) - $refunds_cnt;

		$export_notice = sprintf(
			'Export file successfully generated with <b>%d</b> %s %s',
			$orders_cnt,
			$orders_cnt > 1 ? $this->magic_item_labels->name : $this->magic_item_labels->singular_name,
			$refunds_cnt ? ' and <b>' . $refunds_cnt . '</b> related Refunds.' : '.',
		);

		return $export_notice;
	}

	/**
	 * Builds order ID as string based on its billing email and date created.
	 *
	 * @param int $order_id Order id.
	 * @return string
	 */
	protected function build_exporting_order_id( $order_id ) {

		$wc_orders_row                 = $this->get_order_table_rows( $order_id, 'wc_orders', true );
		$wc_order_operational_data_row = $this->get_order_table_rows( $order_id, 'wc_order_operational_data', true );

		if ( ! isset( $wc_orders_row['billing_email'], $wc_order_operational_data_row['cart_hash'] ) ) {
			return '';
		}

		return $wc_orders_row['billing_email'] . ':' . $wc_order_operational_data_row['cart_hash'];
	}

	/**
	 * Gets existing order ID based on provided data.
	 *
	 * @param string $billing_email Billing email or both billing email and cart hash separated by column.
	 * @param string $cart_hash Cart hash.
	 * @return int|false
	 */
	protected function get_existing_order_id( $billing_email, $cart_hash = '' ) {

		if ( ! $cart_hash && str_contains( $billing_email, ':' ) ) {
			list($billing_email, $cart_hash) = explode( ':', $billing_email, 2 );
		}

		if ( ! $billing_email || ! $cart_hash ) {
			return false;
		}

		global $wpdb;

		$existing_order_id = $wpdb->get_var(
			$wpdb->prepare(
				"SELECT o.id 
				FROM {$wpdb->prefix}wc_orders o
				JOIN {$wpdb->prefix}wc_order_operational_data od ON o.id = od.order_id 
				WHERE o.billing_email = %s 
				AND od.cart_hash = %s 
				LIMIT 1",
				$billing_email,
				$cart_hash
			)
		);

		return $existing_order_id;
	}

	/**
	 * Gets shop order custom table row from provided table name.
	 *
	 * @param int    $order_id Order ID.
	 * @param string $table Table name without prefix.
	 * @param bool   $single_row Whether to return single row.
	 * @return array|false
	 */
	protected function get_order_table_rows( $order_id, $table, $single_row = false ) {

		if ( ! $order_id || ! is_numeric( $order_id ) ) {
			return false;
		}

		global $wpdb;

		$table_name = "{$wpdb->prefix}{$table}";

		// Cache key based on table name and order ID.
		$cache_key  = "wc_order_{$table}_{$order_id}";
		$table_rows = wp_cache_get( $cache_key, MAGIC_EX_IM_CACHE_KEY );

		if ( false === $table_rows ) {

			$query = 'wc_orders' === $table
				? "SELECT * FROM {$table_name} WHERE id = %d"
				: "SELECT * FROM {$table_name} WHERE order_id = %d";

			// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
			$table_rows = $wpdb->get_results( $wpdb->prepare( $query, $order_id ), ARRAY_A );
			$table_rows = array_filter( (array) $table_rows );

			// Cache the result.
			wp_cache_set( $cache_key, $table_rows, MAGIC_EX_IM_CACHE_KEY, 3600 );
		}

		return $single_row ? reset( $table_rows ) : $table_rows;
	}

	/**
	 *  Inserts row into the WooCommerce table.
	 *
	 * @param array  $row Row data.
	 * @param string $table Table name without prefix.
	 * @return int|false
	 */
	protected function insert_wc_table_row( $row, $table ) {

		if ( ! $row || ! is_array( $row ) ) {
			return false;
		}

		global $wpdb;

		$table_name = "{$wpdb->prefix}{$table}";
		$format     = array();

		// Insert a placeholder order for the importing order.
		if ( 'wc_orders' === $table ) {

			$placehold_post_id = wp_insert_post( array( 'post_type' => 'shop_order_placehold' ) );

			if ( ! $placehold_post_id ) {
				return false;
			}

			$row['id'] = $placehold_post_id;
		}

		// Prepare data format for insertion.
		foreach ( $row as $column_value ) {

			if ( is_numeric( $column_value ) ) {

				if ( (string) (int) $column_value !== (string) $column_value ) {
					$format[] = '%f';
				} else {
					$format[] = '%d';
				}
			} else {
				$format[] = '%s';
			}
		}

		$inserted_rows_cnt = $wpdb->insert( $table_name, $row, $format );

		return $inserted_rows_cnt ? ( $row['id'] ?? $wpdb->insert_id ) : false;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 */
	protected function get_key_data( $group, $key, $item_id ) {

		$key_data = '';

		switch ( $group ) {

			case 'wc_order':
				$key_data = $this->get_order_table_rows( $item_id, $key );
				break;

			case 'meta':
				$wc_order = wc_get_order( $item_id );
				$key_data = $wc_order ? $wc_order->get_meta( $key, true, 'edit' ) : '';
				break;
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.

		$this->magic_item_labels                = new stdClass();
		$this->magic_item_labels->name          = __( 'Shop Orders', 'magic-export-import' );
		$this->magic_item_labels->singular_name = __( 'Shop Order', 'magic-export-import' );
		$this->magic_item_labels->all_items     = __( 'All Shop Orders', 'magic-export-import' );

		add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'adjust_value' ), 5, 3 );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		// Query shop orders considering offset.

		global $wpdb;

		$limit  = absint( $args['items_per_request'] );
		$offset = absint( $args['offset'] );
		$search = '%' . $wpdb->esc_like( $args['s'] ) . '%';

		$post_in = $wpdb->get_col(
			$wpdb->prepare(
				"SELECT id FROM {$wpdb->prefix}wc_orders 
				WHERE billing_email LIKE %s 
				OR id LIKE %s
				AND parent_order_id = 0
				LIMIT %d OFFSET %d",
				$search,
				$search,
				$limit,
				$offset
			)
		);

		$post_in      = ( $args['s'] && ! $post_in ) ? array( 0 ) : $post_in;
		$orders_query = new WC_Order_Query( array( 'post__in' => $post_in ) );

		$result['pagination']['more'] = count( $orders_query->get_orders() ) >= $args['items_per_request'];

		foreach ( $orders_query->get_orders() as $order_obj ) {

			$result['results'][] = array(
				'id'   => $order_obj->get_id(),
				'text' => sprintf(
					'%s #%s <code>%s</code>',
					$order_obj->get_billing_email(),
					$order_obj->get_order_number(),
					$order_obj->get_date_created()->date( 'd M Y H:m' )
				),
			);
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $order_ids Specific shop order IDs to get export keys for.
	 */
	protected function get_export_keys( $order_ids = array() ) {

		global $wpdb;

		// Get order keys.
		$order_keys = $this->get_custom_tables_names();

		// Get meta keys.

		$order_ids_where = '';
		$query_params    = $order_ids;

		if ( $order_ids ) {
			$order_ids_in    = implode( ', ', array_fill( 0, count( $order_ids ), ' % d' ) );
			$order_ids_where = "AND {$wpdb->prefix}wc_orders_meta.order_id IN ($order_ids_in)";
		}

		$query = "SELECT DISTINCT(meta_key)
			FROM {$wpdb->prefix}wc_orders
			LEFT JOIN {$wpdb->prefix}wc_orders_meta
			ON {$wpdb->prefix}wc_orders.id = {$wpdb->prefix}wc_orders_meta.order_id
			AND {$wpdb->prefix}wc_orders_meta.meta_key != ''
			$order_ids_where";

		// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
		$meta_keys = $wpdb->get_col( $wpdb->prepare( $query, $query_params ) );

		// Exclude some meta keys from the export.
		$meta_keys = array_diff( $meta_keys, array( '_edit_lock' ) );

		// Combine all keys.
		$export_keys = array_values( array_merge( $order_keys, $meta_keys ) );

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key User key.
	 */
	protected function get_key_group( $key ) {

		if ( in_array( $key, $this->get_custom_tables_names(), true ) ) {
			$group = 'wc_order';

		} else {
			$group = 'meta';
		}

		return apply_filters( 'magic_ex_im_get_key_group', $group, $key );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Key of the provided value.
	 * @param string $key_group Group name $key belongs to.
	 */
	public function adjust_value( $value, $key, $key_group ) {

		if ( ! $value ) {
			return $value;
		}

		// Adjust value during export.
		if ( magic_ex_im_is_export() ) {

			if ( 'wc_order' === $key_group && is_array( $value ) ) {

				foreach ( $value as &$row ) {
					unset( $row['id'], $row['order_id'] );
				}

				if ( 'wc_order_addresses' !== $key ) {
					$value = reset( $value );
				}

				if ( 'wc_orders' === $key ) {
					$value['customer_id']     = magic_ex_im_adjust_user_value( $value['customer_id'] );
					$value['parent_order_id'] = $this->build_exporting_order_id( $value['parent_order_id'] );
				}
			}
		}

		// Adjust value during import.
		if ( magic_ex_im_is_import() ) {

			if ( 'wc_order' === $key_group && is_array( $value ) ) {

				if ( 'wc_orders' === $key ) {
					$value['customer_id']     = magic_ex_im_adjust_user_value( $value['customer_id'] );
					$value['parent_order_id'] = $this->get_existing_order_id( $value['parent_order_id'] );
				}
			}
		}

		return $value;
	}

	/**
	 * {@inheritDoc}.
	 */
	protected function allow_to_update_items() {
		return false;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to validate.
	 */
	protected function validate_importing_item_data( $item_data ) {

		$errors        = array();
		$wc_orders_key = magic_ex_im_build_key( 'wc_orders', $this->get_key_group( 'wc_orders' ) );

		$order_type = $item_data[ $wc_orders_key ]['type'] ?? '';

		// Validate order type.
		if ( ! $order_type || ! in_array( $order_type, wc_get_order_types(), true ) ) {

			$errors[] = sprintf(
				'Invalid order <b>type</b> <code>%s</code> for the %s.',
				$order_type ? $order_type : 'undefined',
				magic_im_get_item_name(),
			);
		}

		$order_status = $item_data[ $wc_orders_key ]['status'] ?? '';

		// Validate order status.
		if ( ! $order_status || ! in_array( $order_status, $this->get_order_statuses(), true ) ) {

			$errors[] = sprintf(
				'Invalid order <b>status</b> <code>%s</code> for the %s.',
				$order_status ? $order_status : 'undefined',
				magic_im_get_item_name(),
			);
		}

		// Check for invalid dates.
		foreach ( array( 'date_created_gmt', 'date_updated_gmt' ) as $date_key ) {

			$date_str = $item_data[ $wc_orders_key ][ $date_key ] ?? '';
			$errors[] = magic_im_validate_date_str( $date_str, $date_key );
		}

		return $errors;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {

		$billing_email = $item_data['wc_order']['wc_orders']['billing_email'] ?? '';
		$cart_hash     = $item_data['wc_order']['wc_order_operational_data']['cart_hash'] ?? '';

		return $this->get_existing_order_id( $billing_email, $cart_hash );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 */
	protected function update_existing_item( $item_data, $item_id ) {
		return false;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {
		return $this->insert_wc_table_row( $item_data['wc_order']['wc_orders'], 'wc_orders' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	protected function update_importing_item( $item_data, $item_id ) {

		// Insert rows in WooCommerce custom tables for the importing order.
		foreach ( $this->get_custom_tables_names() as $table_name ) {

			if ( 'wc_orders' === $table_name || ! isset( $item_data['wc_order'][ $table_name ] ) ) {
				continue;
			}

			$table_data = $item_data['wc_order'][ $table_name ];

			if ( is_array( $table_data ) ) {
				$table_rows = is_array( reset( $table_data ) ) ? $table_data : array( $table_data );

				foreach ( $table_rows as $table_row ) {
					$table_row['order_id'] = $item_id;
					$this->insert_wc_table_row( $table_row, $table_name );
				}
			}
		}

		$wc_order = wc_get_order( $item_id );

		// Update order meta.
		if ( $wc_order && isset( $item_data['meta'] ) ) {

			foreach ( $item_data['meta'] as $meta_key => $value ) {
				$wc_order->update_meta_data( $meta_key, $value );
			}

			$wc_order->save();
		}
	}
}

if ( class_exists( 'woocommerce' ) && wc_string_to_bool( get_option( 'woocommerce_custom_orders_table_enabled' ) ) ) {
	Magic_EX_IM_Type_Shop_Orders::instance();
}
