<?php
/**
 * Class Magic_EX_IM_Type_Terms.
 *
 * @package Magic_Export_Import\Terms
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for terms.
 */
final class Magic_EX_IM_Type_Terms extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'terms';

	/**
	 * Current taxonomy.
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {
		return get_taxonomies( array( 'show_ui' => 1 ), 'objects' );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_mandatory_keys() {
		return array( 'slug', 'taxonomy' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$term_slug = $item_data['term']['slug'] ?? '';

		return $term_slug;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {

		// Initial terms query arguments.
		$args = array(
			'taxonomy'   => $this->magic_item,
			'hide_empty' => false,
			'include'    => $posted_data['magic-ex-items'] ?? array(),
			'fields'     => 'ids',
		);

		$args = apply_filters( 'magic_ex_get_export_items_args', $args, $posted_data );

		$export_terms_query = new WP_Term_Query( $args );
		return $export_terms_query->get_terms();
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 */
	protected function get_key_data( $group, $key, $item_id ) {

		$key_data = '';

		switch ( $group ) {

			case 'term':
				$key_data = get_term_field( $key, $item_id, $this->magic_item );
				break;

			case 'meta':
				$key_data = get_term_meta( $item_id, $key, true );
				break;
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.
		$this->magic_item_labels = get_taxonomy( $this->magic_item )->labels;

		add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'adjust_value' ), 5, 3 );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Key of the provided value.
	 * @param string $key_group Group name $key belongs to.
	 */
	public function adjust_value( $value, $key, $key_group ) {

		if ( ! $value ) {
			return $value;
		}

		// Adjust value during export.
		if ( magic_ex_im_is_export() ) {

			// Convert term ID into slug.
			if ( 'term' === $key_group && 'parent' === $key && $value ) {
				$value = get_term_by( 'term_id', $value, $this->magic_item )->slug ?? '';
			}
		}

		// Adjust value during import.
		if ( magic_ex_im_is_import() ) {

			// Convert term slug into ID.
			if ( 'term' === $key_group && 'parent' === $key && $value ) {
				$term_id = get_term_by( 'slug', $value, $this->magic_item )->term_id ?? '';
				$value   = apply_filters( 'magic_im_item_data_taxonomy', $term_id, $this->magic_item );
			}
		}

		return $value;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		// Query terms considering offset.
		$terms_query = new WP_Term_Query(
			array(
				'taxonomy'   => $args['magic_item'],
				'search'     => $args['s'],
				'hide_empty' => false,
				'number'     => $args['items_per_request'] + 1, // Hack to know there are more terms.
				'offset'     => $args['offset'],
			)
		);

		$result['pagination']['more'] = count( $terms_query->terms ) > $args['items_per_request'];

		foreach ( array_slice( $terms_query->terms, 0, $args['items_per_request'] ) as $term_obj ) {
			$result['results'][] = array(
				'id'   => $term_obj->term_id,
				'text' => sprintf( '%s <code>%s</code>', $term_obj->name, $term_obj->slug ),
			);
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $term_ids Specific term IDs to get export keys for.
	 */
	protected function get_export_keys( $term_ids = array() ) {

		global $wpdb;

		// Get term keys.
		$term_keys = array_merge(
			$wpdb->get_col( "DESC {$wpdb->terms}" ),
			$wpdb->get_col( "DESC {$wpdb->term_taxonomy}" )
		);

		// Exclude some term keys from the export.
		$term_keys = array_diff( $term_keys, array( 'term_id', 'term_taxonomy_id', 'count' ) );

		// Get meta keys.

		$term_ids_where = '';
		$query_params   = array( $this->magic_item );

		if ( ! empty( $term_ids ) ) {
			$term_ids_in    = implode( ', ', array_fill( 0, count( $term_ids ), '%d' ) );
			$term_ids_where = "AND {$wpdb->termmeta}.term_id IN ($term_ids_in)";
			$query_params   = array_merge( $query_params, $term_ids );
		}

		$query = "SELECT DISTINCT meta_key
			FROM {$wpdb->termmeta}
			INNER JOIN {$wpdb->term_taxonomy} ON {$wpdb->termmeta}.term_id = {$wpdb->term_taxonomy}.term_id
			WHERE {$wpdb->term_taxonomy}.taxonomy = %s
			$term_ids_where";

		// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
		$meta_keys = $wpdb->get_col( $wpdb->prepare( $query, $query_params ) );

		// Exclude some meta keys from the export.
		$meta_keys = array_diff( $meta_keys, array( 'product_count_product_cat' ) );

		// Combine all keys.
		$export_keys = array_values( array_merge( $term_keys, $meta_keys ) );

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key Term key.
	 */
	protected function get_key_group( $key ) {

		global $wpdb;
		static $term_keys = null;

		// Get term keys.
		$term_keys ??= array_merge(
			$wpdb->get_col( "DESC {$wpdb->term_taxonomy}" ),
			$wpdb->get_col( "DESC {$wpdb->terms}" )
		);

		if ( in_array( $key, $term_keys, true ) ) {
			$group = 'term';

		} else {
			$group = 'meta';
		}

		return apply_filters( 'magic_ex_im_get_key_group', $group, $key );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to validate.
	 */
	protected function validate_importing_item_data( $item_data ) {

		$errors  = array();
		$tax_key = magic_ex_im_build_key( 'taxonomy', $this->get_key_group( 'taxonomy' ) );

		// Check if importing taxonomy matches current taxonomy.
		if ( isset( $item_data[ $tax_key ] ) && $item_data[ $tax_key ] !== $this->magic_item ) {

			$errors[] = sprintf(
				'Invalid <b>taxonomy</b> <code>%s</code>, current <b>taxonomy</b> is <code>%s</code>.',
				$item_data[ $tax_key ],
				$this->magic_item,
			);
		}

		return $errors;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {

		// Identify importing term by slug.
		$args = array(
			'taxonomy'   => $this->magic_item,
			'hide_empty' => false,
			'fields'     => 'ids',
			'slug'       => $item_data['term']['slug'],
		);

		$existing_term_query = new WP_Term_Query(
			apply_filters( 'magic_im_existing_item_query_args', $args, $item_data )
		);

		$existing_term = $existing_term_query->get_terms();
		return reset( $existing_term );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 */
	protected function update_existing_item( $item_data, $item_id ) {

		$updated_term = wp_update_term( $item_id, $this->magic_item, $item_data['term'] );

		return is_wp_error( $updated_term ) ? $updated_term : $updated_term['term_id'];
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {

		$created_term = wp_insert_term( $item_data['term']['name'], $this->magic_item, $item_data['term'] );

		return is_wp_error( $created_term ) ? $created_term : $created_term['term_id'];
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	protected function update_importing_item( $item_data, $item_id ) {

		// Update term meta.
		foreach ( $item_data['meta'] ?? array() as $meta_key => $value ) {

			// Skip meta if value is empty and meta doesn't exist (don't create new empty meta).
			if ( $value || is_numeric( $value ) || metadata_exists( 'term', $item_id, $meta_key ) ) {
				update_term_meta( $item_id, $meta_key, $value );
			}
		}
	}
}

Magic_EX_IM_Type_Terms::instance();
