<?php
/**
 * Class Magic_EX_IM_Type_Users.
 *
 * @package Magic_Export_Import\Users
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to manage export and import functionality for users.
 */
final class Magic_EX_IM_Type_Users extends Magic_EX_IM_Type {

	const MAGIC_TYPE = 'users';

	/**
	 * Current magic item ('users' by default).
	 *
	 * @var string
	 */
	protected $magic_item = '';

	/**
	 * Class constructor.
	 */
	protected function __construct() {
		parent::__construct( self::MAGIC_TYPE );
	}

	/**
	 * Updates user password with direct database query to prevent hashing (should be already hashed).
	 *
	 * @param string $user_id User ID.
	 * @param string $user_data User data.
	 * @return int|false
	 */
	private function update_user_password( $user_id, $user_data ) {

		if ( ! isset( $user_data['user_pass'] ) ) {
			return false;
		}

		global $wpdb;

		return $wpdb->update(
			$wpdb->users,
			array( 'user_pass' => $user_data['user_pass'] ),
			array( 'ID' => $user_id ),
			array( '%s' ),
			array( '%d' )
		);
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_allowed_magic_items() {
		return array( 'users' => array( 'label' => __( 'Users', 'magic-export-import' ) ) );
	}

	/**
	 * {@inheritDoc}
	 */
	protected function get_mandatory_keys() {
		return array( 'user_email' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data.
	 */
	protected function get_importing_item_name( $item_data ) {

		$user_email = $item_data['user']['user_email'] ?? '';

		return $user_email;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $posted_data Posted export form data.
	 * @param bool  $include_related_items Whether to include related items.
	 */
	protected function get_export_items( $posted_data = array(), $include_related_items = true ) {

		// Initial users query arguments.
		$args = array(
			'number'  => -1,
			'include' => $posted_data['magic-ex-items'] ?? array(),
			'fields'  => 'ids',
		);

		// Advanced filters.
		if ( isset( $posted_data['magic-ex-advanced'] ) ) {

			// Query by user roles.
			if ( isset( $posted_data['magic-ex-by-roles'], $posted_data['magic-ex-roles'] ) ) {
				$args['role__in'] = $posted_data['magic-ex-roles'];
			}

			// Query by registered date.
			if ( isset( $posted_data['magic-ex-by-date'] ) ) {

				$args['date_query'] = array(
					'after'     => $posted_data['magic-ex-date-from'],
					'before'    => $posted_data['magic-ex-date-to'],
					'inclusive' => true,
				);
			}
		}

		$args = apply_filters( 'magic_ex_get_export_items_args', $args, $posted_data );

		$export_users_query = new WP_User_Query( $args );
		return $export_users_query->get_results();
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $group Key related group.
	 * @param string $key The Key.
	 * @param int    $item_id Item ID.
	 */
	protected function get_key_data( $group, $key, $item_id ) {

		$key_data = '';

		switch ( $group ) {

			case 'user':
				$key_data = get_user_by( 'ID', $item_id )->{$key};
				break;

			case 'meta':
				$key_data = get_user_meta( $item_id, $key, true );
				break;
		}

		return $key_data;
	}

	/**
	 * {@inheritDoc}
	 */
	protected function init_magic_item() {

		// Set labels.

		$this->magic_item_labels                = new stdClass();
		$this->magic_item_labels->name          = __( 'Users', 'magic-export-import' );
		$this->magic_item_labels->singular_name = __( 'User', 'magic-export-import' );
		$this->magic_item_labels->all_items     = __( 'All Users', 'magic-export-import' );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $args Quering arguments.
	 */
	protected function get_individual_export_items( $args ) {

		$result = array(
			'results'    => array(),
			'pagination' => array( 'more' => false ),
		);

		// Query users considering offset.
		$users_query = new WP_User_Query(
			array(
				'number' => $args['items_per_request'],
				'offset' => $args['offset'],
				'search' => '*' . $args['s'] . '*',
			)
		);

		$result['pagination']['more'] = $users_query->total_users > $args['items_per_request'] + $args['offset'];

		foreach ( $users_query->get_results() as $user_obj ) {
			$result['results'][] = array(
				'id'   => $user_obj->ID,
				'text' => sprintf( '%s <code>%s</code>', $user_obj->display_name, $user_obj->user_email ),
			);
		}

		return $result;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $user_ids Specific user IDs to get export keys for.
	 */
	protected function get_export_keys( $user_ids = array() ) {

		global $wpdb;

		// Get user keys.
		$user_keys = $wpdb->get_col( "DESC {$wpdb->users}" );

		// Exclude some user keys from the export.
		$user_keys = array_diff( $user_keys, array( 'ID', 'user_url' ) );

		// Get meta keys.

		$user_ids_where = '';
		$query_params   = $user_ids;

		if ( $user_ids ) {
			$user_ids_in    = implode( ', ', array_fill( 0, count( $user_ids ), '%d' ) );
			$user_ids_where = "AND {$wpdb->usermeta}.user_id IN ($user_ids_in)";
		}

		$query = "SELECT DISTINCT(meta_key)
			FROM $wpdb->users
			LEFT JOIN $wpdb->usermeta
			ON $wpdb->users.ID = $wpdb->usermeta.user_id
			AND $wpdb->usermeta.meta_key != ''
			$user_ids_where";

		// phpcs:ignore WordPress.DB.PreparedSQL.NotPrepared
		$meta_keys = $wpdb->get_col( $wpdb->prepare( $query, $query_params ) );

		// Exclude some meta keys from the export.

		$meta_keys = array_diff(
			$meta_keys,
			array(
				'session_tokens',
				'community-events-location',
				'wp_dashboard_quick_press_last_post_id',
				'last_update',
			)
		);

		foreach ( $meta_keys as $index => $meta_key ) {

			if ( str_starts_with( $meta_key, 'closedpostboxes_' )
				|| str_starts_with( $meta_key, 'meta-box-order_' )
				|| str_starts_with( $meta_key, 'screen_layout_' )
				|| str_starts_with( $meta_key, 'metaboxhidden_' ) ) {

				unset( $meta_keys[ $index ] );
			}
		}

		// Exclude meta keys starting with 'wp_' followed by numeric value.
		$meta_keys = array_filter(
			$meta_keys,
			function ( $meta_key ) {
				return ! preg_match( '/^wp_\d+/', $meta_key );
			}
		);

		// Combine all keys.
		$export_keys = array_values( array_merge( $user_keys, $meta_keys ) );

		return array_unique( array_filter( $export_keys ) );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param string $key User key.
	 */
	protected function get_key_group( $key ) {

		global $wpdb;
		static $user_keys = null;

		// Get user keys.
		$user_keys ??= $wpdb->get_col( "DESC {$wpdb->users}", 0 );

		if ( in_array( $key, $user_keys, true ) ) {
			$group = 'user';

		} else {
			$group = 'meta';
		}

		return apply_filters( 'magic_ex_im_get_key_group', $group, $key );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to validate.
	 */
	protected function validate_importing_item_data( $item_data ) {

		$errors = array();

		$email_key = magic_ex_im_build_key( 'user_email', $this->get_key_group( 'user_email' ) );

		// Check user email is valid.
		if ( ( $item_data[ $email_key ] ?? '' ) && ! is_email( $item_data[ $email_key ] ) ) {
			$errors[] = sprintf( 'Invalid <code>user_email</code> for the %s.', magic_im_get_item_name() );
		}

		// Check for the invalid date.
		$date_key = magic_ex_im_build_key( 'user_registered', $this->get_key_group( 'user_registered' ) );
		$errors[] = magic_im_validate_date_str( $item_data[ $date_key ] ?? '', $date_key );

		return $errors;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function get_existing_item( $item_data ) {

		// Identify importing user by its email.

		$args = array(
			'fields'         => 'ids',
			'number'         => 1,
			'search'         => $item_data['user']['user_email'],
			'search_columns' => array( 'user_email' ),
		);

		$existing_user_query = new WP_User_Query(
			apply_filters( 'magic_im_existing_item_query_args', $args, $item_data )
		);

		$existing_user = $existing_user_query->get_results();
		return reset( $existing_user );
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 * @param array $item_id Item ID to update.
	 */
	protected function update_existing_item( $item_data, $item_id ) {

		$item_data['user']['ID'] = $item_id;

		$user_id = wp_update_user( $item_data['user'] );
		$this->update_user_password( $user_id, $item_data['user'] );

		return $user_id;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item importing data.
	 */
	protected function create_new_item( $item_data ) {

		// Column 'user_login' is mandatory to create new user.
		if ( ! isset( $item_data['user']['user_login'] ) || ! $item_data['user']['user_login'] ) {
			$item_data['user']['user_login'] = $item_data['user']['user_email'];
		}

		$user_id = wp_insert_user( $item_data['user'] );
		$this->update_user_password( $user_id, $item_data['user'] );

		return $user_id;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @param array $item_data Item data to update.
	 * @param int   $item_id Item ID.
	 */
	protected function update_importing_item( $item_data, $item_id ) {

		// Update user meta.
		foreach ( $item_data['meta'] ?? array() as $meta_key => $value ) {

			// Skip meta if value is empty and meta doesn't exist (don't create new empty meta).
			if ( $value || is_numeric( $value ) || metadata_exists( 'user', $item_id, $meta_key ) ) {
				update_user_meta( $item_id, $meta_key, $value );
			}
		}
	}
}

Magic_EX_IM_Type_Users::instance();
