<?php
/**
 * Class Magic_EX_IM_Adapter_ACF.
 *
 * @package Magic_Export_Import\Plugin_Adapter
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to provide 'Advanced Custom Fields' plugin support.
 */
class Magic_EX_IM_Adapter_ACF {

	const KEYS_GROUP = 'ACF';

	/**
	 * Constructor.
	 */
	public function __construct() {

		if ( class_exists( 'ACF' ) ) {
			add_action( 'magic_ex_im_init', array( $this, 'magic_ex_im_init' ) );
		}
	}

	/**
	 * Processes action hook 'magic_ex_im_init'.
	 */
	public function magic_ex_im_init() {

			// Export / Import.
			add_filter( 'magic_ex_im_get_allowed_magic_items', array( $this, 'magic_ex_im_get_allowed_magic_items' ) );

			// Export.
			add_filter( 'magic_ex_get_key_group', array( $this, 'magic_ex_get_key_group' ), 10, 3 );
			add_filter( 'magic_ex_adjust_editor_value_gutenberg', array( $this, 'adjust_editor_value_gutenberg' ), 10, 1 );
			add_filter( 'magic_ex_keys', array( $this, 'magic_ex_keys' ) );
			add_filter( 'magic_ex_get_key_data', array( $this, 'magic_ex_get_key_data' ), 10, 3 );
			add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'adjust_acf_value' ), 10, 3 );

			// Import.
			add_filter( 'magic_im_get_key_group', array( $this, 'magic_im_get_key_group' ), 10, 3 );
			add_filter( 'magic_im_adjust_editor_value_gutenberg', array( $this, 'adjust_editor_value_gutenberg' ), 10, 2 );
			add_filter( 'magic_im_maybe_adjust_value', array( $this, 'adjust_acf_value' ), 10, 3 );
			add_filter( 'magic_im_item_data_before_update', array( $this, 'magic_im_item_data_before_update' ) );
			remove_filter( 'content_save_pre', 'acf_parse_save_blocks', 5 );
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_allowed_magic_items'.
	 *
	 * @param array $allowed_magic_items Allowed magic items for the current magic type.
	 * @return array
	 */
	public function magic_ex_im_get_allowed_magic_items( $allowed_magic_items ) {

		// Exclude ACF related post types.
		if ( 'posts' === magic_ex_im_get_magic_type() ) {

			$post_types_to_exlude = array(
				'acf-taxonomy',
				'acf-post-type',
				'acf-ui-options-page',
				'acf-field-group',
				'acf-field',
			);

			$allowed_magic_items = array_diff_key( $allowed_magic_items, array_flip( $post_types_to_exlude ) );
		}

		return $allowed_magic_items;
	}

	/**
	 * Gets ACF Options Page ID based on the provided slug and ACF field name.
	 *
	 * @param string $acf_page_slug ACF Options Page slug.
	 * @param string $key ACF field name.
	 * @return string
	 */
	private function get_acf_options_page_id( $acf_page_slug, $key ) {

		$acf_options_pages = Magic_EX_IM_Type_ACF_Options_Pages::instance();

		if ( in_array( $key, $acf_options_pages->get_acf_pages_export_keys( $acf_page_slug ), true ) ) {
			$acf_page_id = $acf_options_pages->get_items( $acf_page_slug )['post_id'];
		}

		return $acf_page_id ?? false;
	}

	/**
	 * Processes filter hooks 'magic_ex_adjust_editor_value_gutenberg' and 'magic_im_adjust_editor_value_gutenberg'.
	 *
	 * @param string $value Gutenberg content.
	 * @return string
	 *
	 * @see acf_serialize_block_attributes()
	 */
	public function adjust_editor_value_gutenberg( $value ) {

		// Match ACF Gutenberg blocks.
		preg_match_all( '/<!-- wp:acf\/[a-zA-Z0-9_-]+ ({.*?}) \/\-\->/s', $value, $matches );

		// Loop through each ACF block found.
		foreach ( $matches[0] as $index => $acf_block ) {

			// Extract JSON part from the matched block.
			$json_str = $matches[1][ $index ];

			// Decode JSON into array.
			$json_data = json_decode( $json_str, true );

			// Skip ACF block if JSON is not valid.
			if ( json_last_error() !== JSON_ERROR_NONE || ! isset( $json_data['data'] ) ) {
				continue;
			}

			// Loop throgh all ACF fields in the block.
			foreach ( $json_data['data'] as $acf_name => &$acf_value ) {
				$maybe_acf_key = "_$acf_name";

				if ( isset( $json_data['data'][ $maybe_acf_key ] ) ) {

					$acf_key          = $json_data['data'][ $maybe_acf_key ];
					$acf_field_object = get_field_object( $acf_key );

					$acf_value = $this->adjust_acf_value( $acf_value, $acf_name, self::KEYS_GROUP, $acf_field_object );
				}
			}

			// Encode modified JSON back to string (refer to @see tag).
			$modified_json_str = serialize_block_attributes( $json_data );

			// Prepare content for the wp_unslash() using under wp_insert_post().
			if ( magic_ex_im_is_import() ) {
				$modified_json_str = str_replace( '\\', '\\\\', $modified_json_str );
			}

			// Reconstruct the full block with modified JSON.
			$modified_block = str_replace( $json_str, $modified_json_str, $acf_block );

			// Replace the original block in the content with the modified block.
			$value = str_replace( $acf_block, $modified_block, $value );
		}

		return $value;
	}

	/**
	 * Processes filter hook 'magic_ex_keys'.
	 *
	 * @param array $export_keys Item grouped keys for export.
	 * @return array
	 */
	public function magic_ex_keys( $export_keys ) {

		if ( ! isset( $export_keys['meta'] ) ) {
			return $export_keys;
		}

		$meta_keys_to_exclude = array();

		// Exclude some meta keys for users.

		if ( 'users' === magic_ex_im_get_magic_type() ) {
			$meta_keys_to_exclude[] = 'acf_user_settings';
		}

		foreach ( $export_keys['meta'] as $meta_key ) {

			if ( 'users' === magic_ex_im_get_magic_type()
				&& ( str_starts_with( $meta_key, 'manageedit-acf-' ) || str_ends_with( $meta_key, '_acf-field-group' ) ) ) {

				$meta_keys_to_exclude[] = $meta_key;
			}
		}

		// Exclude ACF related keys like '_meta_key', 'meta_key_subfield', 'meta_key_0_subfield' 'meta_key_group_subfield' etc.
		foreach ( $export_keys[ self::KEYS_GROUP ] ?? array() as $meta_key ) {

			$meta_keys_to_exclude[] = "_$meta_key";

			$regex = sprintf( '/^_?%s_.*/', $meta_key );
			array_push( $meta_keys_to_exclude, ...preg_grep( $regex, $export_keys['meta'] ) );
		}

		$export_keys['meta'] = array_diff( $export_keys['meta'], $meta_keys_to_exclude );

		return $export_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_group'.
	 *
	 * @param string   $group Current key group.
	 * @param string   $key The key.
	 * @param string[] $all_export_keys All export keys.
	 * @return string
	 */
	public function magic_ex_get_key_group( $group, $key, $all_export_keys ) {

		if ( 'meta' === $group && $this->get_acf_field_key_by_name( $key ) ) {

			// Check for existence of ACF meta with underscore prefix, where ACF field identifier should be stored.
			if ( in_array( "_$key", $all_export_keys, true ) ) {
				return self::KEYS_GROUP;
			}
		}

		return $group;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_group'.
	 *
	 * @param string   $group Current key group.
	 * @param string   $key The key.
	 * @param string[] $predefined_group Predefined group.
	 * @return string
	 */
	public function magic_im_get_key_group( $group, $key, $predefined_group ) {

		if ( 'meta' === $group && $this->get_acf_field_key_by_name( $key ) ) {

			// Additional check if predefined group matches.
			if ( self::KEYS_GROUP === $predefined_group ) {
				return self::KEYS_GROUP;
			}
		}

		return $group;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_data';
	 *
	 * @param mixed  $key_data Current key data.
	 * @param string $key Group key.
	 * @param string $group Key Group.
	 * @return mixed
	 */
	public function magic_ex_get_key_data( $key_data, $key, $group ) {

		if ( self::KEYS_GROUP !== $group ) {
			return $key_data;
		}

		$item_acf_id      = $this->get_item_acf_id( $key );
		$acf_field_object = get_field_object( $this->get_acf_field_key_by_name( $key ) );

		if ( isset( $acf_field_object['type'] ) && $item_acf_id ) {
			$key_data = get_field( $key, $item_acf_id );
		}

		return $key_data;
	}

	/**
	 * Adjusts ACF value both for export and import.
	 *
	 * The algorithm is suitable for both ACF Gutenberg blocks and classic ACF fields.
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Value key.
	 * @param string $key_group Group name $key belongs to.
	 * @param array  $acf_field_object ACF field object.
	 *
	 * @return mixed Adjusted ACF value.
	 *
	 * @see Magic_EX_IM_Adapter_ACF::adjust_editor_value_gutenberg()
	 */
	public function adjust_acf_value( $value, $key, $key_group, $acf_field_object = null ) {

		if ( self::KEYS_GROUP !== $key_group || ! $value ) {
			return $value;
		}

		$acf_field_object ??= get_field_object( $this->get_acf_field_key_by_name( $key ) );

		if ( ! isset( $acf_field_object['type'] ) ) {
			return $value;
		}

		$acf_field_type = $acf_field_object['type'];

		// Process exporting item.
		if ( magic_ex_im_is_export() ) {

			switch ( $acf_field_type ) {

				case 'textarea':
					$new_lines = $acf_field_object['new_lines'] ?? '';

					if ( 'br' === $new_lines ) {
						$value = str_replace( array( '<br>', '<br/>', '<br />' ), '', $value );

					} elseif ( 'wpautop' === $new_lines ) {
						$value = str_replace( array( '<p>', '</p>' ), array( '', "\n" ), $value );
					}
					break;

				case 'image':
				case 'file':
				case 'gallery':
					if ( is_array( $value ) && ! isset( $value['id'], $value['ID'] ) ) {
						foreach ( array_keys( $value ) as $index ) {
							$value[ $index ] = $this->adjust_acf_value(
								$value[ $index ],
								$key,
								$key_group,
								$acf_field_object
							);
						}
					} elseif ( ! wp_http_validate_url( $value ) ) {
						$value = $value['url'] ?? wp_get_attachment_url( $value ); // Set value as media file URL.
					}

					break;

				case 'post_object':
				case 'taxonomy':
				case 'relationship':
					if ( is_array( $value ) ) {

						foreach ( array_keys( $value ) as $index ) {
							$value[ $index ] = $this->adjust_acf_value(
								$value[ $index ],
								$key,
								$key_group,
								$acf_field_object,
							);
						}
					} else {
						// Convert ID into term slug / post name.
						$value = 'taxonomy' === $acf_field_type
							? get_term_field( 'slug', $value, $acf_field_object['taxonomy'] )
							: get_post_field( 'post_name', $value );
					}

					break;
			}
		}

		// Process importing item.
		if ( magic_ex_im_is_import() ) {

			switch ( $acf_field_type ) {

				case 'wysiwyg':
					$value = magic_im_adjust_editor_value( $value );
					break;

				case 'url':
					$value = magic_im_media_url_to_local_url( $value );
					break;

				case 'link':
					if ( is_scalar( $value ) ) {
						$value = array(
							'title'  => '',
							'url'    => $value,
							'target' => '',
						);
					}

					$value['url'] = magic_im_media_url_to_local_url( $value['url'] );
					break;

				case 'image':
				case 'file':
				case 'gallery':
					if ( is_array( $value ) ) {

						foreach ( array_keys( $value ) as $index ) {
							$value[ $index ] = $this->adjust_acf_value(
								$value[ $index ],
								$key,
								$key_group,
								$acf_field_object
							);
						}
					} else {
						$value = magic_im_media_url_to_local_id( $value );
					}
					break;

				case 'post_object':
				case 'taxonomy':
				case 'relationship':
					if ( is_array( $value ) ) {

						foreach ( array_keys( $value ) as $index ) {
							$value[ $index ] = $this->adjust_acf_value(
								$value[ $index ],
								$key,
								$key_group,
								$acf_field_object,
							);
						}

						// Convert term slug into term ID.
					} elseif ( 'taxonomy' === $acf_field_type ) {

						$term_id = get_term_by( 'slug', $value, $acf_field_object['taxonomy'] )->term_id ?? false;
						$value   = apply_filters( 'magic_im_item_data_taxonomy', $term_id, $acf_field_object['taxonomy'] );

						// Convert post name into Post ID.
					} else {
						$value = magic_im_post_id_by_name( $value, $acf_field_object['post_type'] );
					}

					break;
			}
		}

		// Process exporting or importing item.
		switch ( $acf_field_type ) {

			case 'group':
				if ( is_array( $value ) ) {
					foreach ( array_keys( $value ) as $group_item_key ) {
						$acf_sub_field_object = null;

						foreach ( $acf_field_object['sub_fields'] ?? array() as $index => $sub_field ) {

							if ( isset( $sub_field['name'] ) && $group_item_key === $sub_field['name'] ) {

								$acf_sub_field_object = $sub_field;
								unset( $acf_field_object['sub_fields'][ $index ] );
								break;
							}
						}

						if ( ! $acf_sub_field_object ) {
							continue;
						}

						$value[ $group_item_key ] = $this->adjust_acf_value(
							$value[ $group_item_key ],
							$key,
							$key_group,
							$acf_sub_field_object
						);
					}
				}

				break;

			case 'repeater':
				if ( is_array( $value ) ) {
					foreach ( $value as $repeater_item_index => $repeater_item_value ) {

						if ( is_array( $repeater_item_value ) ) {

							foreach ( array_keys( $repeater_item_value ) as $i => $subfield_key ) {

								if ( ! isset( $acf_field_object['sub_fields'][ $i ] ) ) {
									continue;
								}

								$acf_sub_field_object = $acf_field_object['sub_fields'][ $i ];

								$value[ $repeater_item_index ][ $subfield_key ] = $this->adjust_acf_value(
									$value[ $repeater_item_index ][ $subfield_key ],
									$key,
									$key_group,
									$acf_sub_field_object
								);

							}
						}
					}
				}

				break;

			case 'flexible_content':
				if ( is_array( $value ) && isset( $acf_field_object['layouts'] ) ) {
					foreach ( $value as $fc_item_index => $fc_item_value ) {

						if ( ! isset( $fc_item_value['acf_fc_layout'] ) ) {
							continue;
						}

						$fc_item_layout_name      = $fc_item_value['acf_fc_layout'];
						$fc_item_acf_field_object = null;

						unset( $fc_item_value['acf_fc_layout'] );

						foreach ( $acf_field_object['layouts'] as $acf_layout_object ) {
							if ( $fc_item_layout_name === $acf_layout_object['name'] ) {
								$fc_item_acf_field_object = $acf_layout_object;
								break;
							}
						}

						if ( isset( $fc_item_acf_field_object ) ) {
							foreach ( array_keys( $fc_item_value ) as $i => $fc_subitem_key ) {
								$acf_fc_subitem_object = $fc_item_acf_field_object['sub_fields'][ $i ];

								$value[ $fc_item_index ][ $fc_subitem_key ] = $this->adjust_acf_value(
									$value[ $fc_item_index ][ $fc_subitem_key ],
									$key,
									$key_group,
									$acf_fc_subitem_object
								);
							}
						}
					}
				}
				break;
		}

		return $value;
	}

	/**
	 * Processes filter hook 'magic_im_item_data_before_update'.
	 *
	 * @param array $item_data Importing item data grouped by type.
	 * @return array
	 */
	public function magic_im_item_data_before_update( $item_data ) {

		// Update ACF related data.
		foreach ( $item_data[ self::KEYS_GROUP ] ?? array() as $field_name => $value ) {

			$item_acf_id      = $this->get_item_acf_id( $field_name );
			$acf_field_object = get_field_object( $this->get_acf_field_key_by_name( $field_name ) );

			if ( isset( $acf_field_object['type'] ) && $item_acf_id ) {
				update_field( $field_name, $value, $item_acf_id );
			}
		}

		return $item_data;
	}

	/**
	 * Gets processing item ACF ID.
	 *
	 * @param string $field_name ACF field name.
	 * @return string|int
	 */
	private function get_item_acf_id( $field_name ) {

		$item_id = magic_ex_im_get_item_id();

		switch ( magic_ex_im_get_magic_type() ) {

			case 'terms':
				$item_id = magic_ex_im_get_magic_item() . '_' . $item_id;
				break;

			case 'users':
				$item_id = 'user_' . $item_id;
				break;

			case 'comments':
				$item_id = 'comment_' . $item_id;
				break;

			case 'acf_options_pages':
				$item_id = $this->get_acf_options_page_id( $item_id, $field_name );
				break;
		}

		return $item_id;
	}

	/**
	 * Finds ACF field key by its name.
	 *
	 * @param string $field_name ACF field name.
	 * @return string|false ACF field key on success, false if not found.
	 */
	private function get_acf_field_key_by_name( $field_name ) {

		$acf_field_key      = false;
		$processing_item_id = magic_ex_im_get_item_id();

		// Prepare filter for the function acf_get_field_groups().
		switch ( magic_ex_im_get_magic_type() ) {

			case 'posts':
				if ( $processing_item_id ) {

					$filter = 'attachment' === magic_ex_im_get_magic_item()
						? array( 'attachment' => $processing_item_id )
						: array( 'post_id' => $processing_item_id );
				}
				break;

			case 'terms':
				$filter = array( 'taxonomy' => magic_ex_im_get_magic_item() );
				break;

			case 'users':
				if ( $processing_item_id ) {
					$filter = array( 'user_id' => $processing_item_id );
				}
				break;

			case 'options_page':
				if ( $processing_item_id ) {
					$filter = array( 'options_page' => $processing_item_id );
				}
				break;
		}

		$acf_field_groups = acf_get_field_groups( $filter ?? array() );

		foreach ( $acf_field_groups as $field_group ) {
			$acf_fields = acf_get_fields( $field_group['key'] );

			foreach ( $acf_fields as $acf_field ) {
				if ( $field_name === $acf_field['name'] ) {
					$acf_field_key = $acf_field['key'];
					break;
				}
			}
		}

		return $acf_field_key;
	}
}

return new Magic_EX_IM_Adapter_ACF();
