<?php
/**
 * Class Magic_EX_IM_Adapter_Polylang.
 *
 * @package Magic_Export_Import\Plugin_Adapter
 *
 * phpcs:disable WordPress.DB.SlowDBQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to provide 'Polylang' plugin support.
 */
class Magic_EX_IM_Adapter_Polylang {

	const KEYS_GROUP = 'polylang';

	/**
	 * Polylang language code for a processing item.
	 *
	 * @var string
	 */
	private $lang = '';

	/**
	 * Constructor.
	 */
	public function __construct() {

		if ( class_exists( 'Polylang' ) ) {
			add_action( 'magic_ex_im_init', array( $this, 'magic_ex_im_init' ) );
		}
	}

	/**
	 * Processes action hook 'magic_ex_im_init'.
	 */
	public function magic_ex_im_init() {

		// Export/Import.
		add_filter( 'magic_ex_im_get_mandatory_keys', array( $this, 'magic_ex_im_get_mandatory_keys' ) );
		add_filter( 'magic_ex_im_get_key_group', array( $this, 'magic_ex_im_get_key_group' ), 10, 2 );

		// Export.
		add_filter( 'magic_ex_keys', array( $this, 'magic_ex_keys' ) );
		add_filter( 'magic_ex_get_key_data', array( $this, 'magic_ex_get_key_data' ), 10, 3 );
		add_filter( 'magic_ex_posts_advanced_filter_terms', array( $this, 'magic_ex_posts_advanced_filter_terms' ), 10, 2 );
		add_filter( 'magic_ex_item_data_taxonomy', array( $this, 'translate_term' ), 10, 2 );
		add_filter( 'magic_ex_get_export_items_args', array( $this, 'magic_ex_get_export_items_args' ), 10, 2 );
		add_action( 'magic_ex_before_submit_button', array( $this, 'magic_ex_before_submit_button' ) );
		add_action( 'magic_ex_advanced_filters', array( $this, 'magic_ex_advanced_filters' ) );

		// Import.
		add_filter( 'get_terms_args', array( $this, 'get_terms_args' ), 5 ); // Before Polylang.
		add_filter( 'magic_im_validate_item_data', array( $this, 'magic_im_validate_item_data' ), 10, 2 );
		add_action( 'magic_im_item_id', array( $this, 'magic_im_item_id' ) );
		add_filter( 'magic_im_item_data_before_adjust', array( $this, 'magic_im_item_data_before_adjust' ) );
		add_filter( 'magic_im_existing_item_query_args', array( $this, 'magic_im_existing_item_query_args' ) );
		add_filter( 'magic_im_post_id_by_name_args', array( $this, 'magic_im_post_id_by_name_args' ) );
		add_filter( 'magic_im_item_data_taxonomy', array( $this, 'translate_term' ), 10, 2 );
	}

	/**
	 * Sets processing item language.
	 *
	 * @param string $lang Language slug.
	 */
	private function set_lang( $lang ) {
		$this->lang = $lang;
	}

	/**
	 * Gets processing item language.
	 *
	 * @return string
	 */
	private function get_lang() {
		return $this->lang;
	}

	/**
	 * Gets Polylang term slug separator.
	 *
	 * @return string
	 */
	private function get_term_slug_separator() {

		// Constant isn't available in old versions of Polylang Pro.
		if ( class_exists( 'PLL_Share_Term_Slug' ) ) {
			return defined( 'PLL_Share_Term_Slug::TERM_SLUG_SEPARATOR' ) ? PLL_Share_Term_Slug::TERM_SLUG_SEPARATOR : '___';
		}

		return '';
	}

	/**
	 * Processes filter hoook 'magic_ex_get_export_items_args'.
	 *
	 * @param array $args Quering arguments.
	 * @param array $posted_data Export from posted data.
	 * @return array
	 */
	public function magic_ex_get_export_items_args( $args, $posted_data ) {

		if ( isset( $posted_data['magic-ex-advanced'], $posted_data['magic-ex-by-langs'], $posted_data['magic-ex-langs'] ) ) {
			$args['lang'] = $posted_data['magic-ex-langs'];

		} elseif ( isset( $posted_data['lang'] ) ) {
			$args['lang'] = $posted_data['lang'];
		}

		return $args;
	}

	/**
	 * Processes action hook 'magic_ex_before_submit_button'.
	 */
	public function magic_ex_before_submit_button() {
		printf( '<input type="hidden" name="lang" value="%s">', esc_attr( pll_current_language() ) );
	}

	/**
	 * Processes action hook 'magic_ex_advanced_filters'.
	 */
	public function magic_ex_advanced_filters() {

		$magic_type    = magic_ex_im_get_magic_type();
		$template_file = MAGIC_EX_IM_ABSPATH . "template-parts/advanced-filters/advanced-filters-{$magic_type}-polylang.php";

		if ( file_exists( $template_file ) ) {
			load_template( $template_file );
		}
	}

	/**
	 * Checks if language code belongs to an installed Polylang language.
	 *
	 * @param string $lang_code Language code.
	 * @return bool
	 */
	private function is_installed_lang( $lang_code ) {
		return in_array( $lang_code, pll_languages_list(), true );
	}

	/**
	 * Checks if current processing item is translated by Polylang.
	 *
	 * @param string $magic_item Processing magic item.
	 * @return bool
	 */
	private function is_translated_item( $magic_item = null ) {

		$magic_type   = magic_ex_im_get_magic_type();
		$magic_item ??= magic_ex_im_get_magic_item();

		$is_translated_post = 'posts' === $magic_type && pll_is_translated_post_type( $magic_item );
		$is_translated_term = 'terms' === $magic_type && pll_is_translated_taxonomy( $magic_item );

		return $is_translated_post || $is_translated_term;
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_mandatory_keys'.
	 *
	 * @param array $mandatory_keys Mandatory keys.
	 * @return array
	 */
	public function magic_ex_im_get_mandatory_keys( $mandatory_keys ) {

		if ( $this->is_translated_item() ) {
			$mandatory_keys[] = 'language';
		}

		return $mandatory_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_data';
	 *
	 * @param mixed  $key_data Current key data.
	 * @param string $key Data key.
	 * @param string $group Data key Group.
	 * @return mixed
	 */
	public function magic_ex_get_key_data( $key_data, $key, $group ) {

		if ( self::KEYS_GROUP !== $group ) {
			return $key_data;
		}

		$item_id = magic_ex_im_get_item_id();

		switch ( $key ) {

			// Set item language as code.
			case 'language':
				if ( 'terms' === magic_ex_im_get_magic_type() ) {
					$key_data = pll_get_term_language( $item_id );
				}

				if ( 'posts' === magic_ex_im_get_magic_type() ) {

					$key_data = pll_get_post_language( $item_id );

					if ( ! $key_data && has_post_parent( $item_id ) ) {
						$post_parent_id = get_post_field( 'post_parent', $item_id );
						$key_data       = pll_get_post_language( $post_parent_id );
					}
				}

				break;

			// Set post translations as post names.
			case 'post_translations':
				$key_data = array_map(
					function ( $translation_id ) {
						return get_post_field( 'post_name', $translation_id );
					},
					pll_get_post_translations( $item_id )
				);

				break;

			// Set term translations as term slugs.
			case 'term_translations':
				$key_data = array_map(
					function ( $term_id ) {
						return get_term_field( 'slug', $term_id, magic_ex_im_get_magic_item() );
					},
					pll_get_term_translations( $item_id )
				);

				break;
		}

		return $key_data;
	}

	/**
	 * Processes filter hook 'magic_im_validate_item_data'.
	 *
	 * @param string[] $errors Existing validation error.
	 * @param array    $item_data Item data.
	 *
	 * @return string
	 */
	public function magic_im_validate_item_data( $errors, $item_data ) {

		$lang_key  = magic_ex_im_build_key( 'language', self::KEYS_GROUP );
		$item_lang = $item_data[ $lang_key ] ?? '';

		if ( $item_lang && ! $this->is_installed_lang( $item_lang ) ) {
			$errors[] = sprintf( 'Non-installed Polylang <b>language</b> <code>%s</code>.', $item_data[ $lang_key ] );
		}

		return $errors;
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_key_group'.
	 *
	 * @param string $group Current key group.
	 * @param string $key The key.
	 * @return string
	 */
	public function magic_ex_im_get_key_group( $group, $key ) {

		if ( in_array( $key, array( 'language', 'post_translations', 'term_translations' ), true ) ) {
			$group = self::KEYS_GROUP;
		}

		return $group;
	}

	/**
	 * Processes filter hook 'magic_ex_keys'.
	 *
	 * @param array $export_keys Item grouped keys for export.
	 * @return array
	 */
	public function magic_ex_keys( $export_keys ) {

		// Add polylang keys for translated terms.
		if ( 'terms' === magic_ex_im_get_magic_type() && $this->is_translated_item() ) {
			$export_keys[ self::KEYS_GROUP ][] = 'language';
			$export_keys[ self::KEYS_GROUP ][] = 'term_translations';
		}

		return $export_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_posts_advanced_filter_terms'.
	 *
	 * @param WP_Term[] $terms Terms for the posts advanced filter.
	 * @param string    $tax Terms taxonomy name.
	 *
	 * @return WP_Term[]
	 */
	public function magic_ex_posts_advanced_filter_terms( $terms, $tax ) {

		// Remove some Polylang taxonomies from advanced export filters.
		if ( 'post_translations' === $tax || ( 'language' === $tax && pll_current_language() ) ) {
			return array();
		}

		return $terms;
	}

	/**
	 * Processes filter hook 'get_terms_args'.
	 *
	 * @param array $args WP_Term_Query arguments.
	 * @return array
	 */
	public function get_terms_args( $args ) {

		// Prevent quering by default language during import.
		if ( magic_ex_im_is_import() && ! isset( $args['lang'] ) ) {
			$args['lang'] = '';
		}

		return $args;
	}

	/**
	 * Processes filter hook 'magic_im_item_data_before_adjust'.
	 *
	 * @param array $item_data Importing item data grouped by type.
	 * @return array
	 */
	public function magic_im_item_data_before_adjust( $item_data ) {

		// Set importing item language.

		$item_lang = $item_data[ self::KEYS_GROUP ]['language'] ?? '';

		// Try to set ACF options page language based on its ID.
		// This is a common way to register same ACF options page for multiple languages.
		if ( ! $item_lang && 'acf_options_pages' === magic_ex_im_get_magic_type() ) {
			$item_lang = $item_data['ACF-page']['post_id'];

			if ( $item_lang && ! $this->is_installed_lang( $item_lang ) ) {
				$item_lang = '';
			}
		}

		if ( 'posts' === magic_ex_im_get_magic_type() && ! $item_lang ) {
			if ( ! $this->is_translated_item( $item_data['post']['post_type'] ) ) {

				$this->set_lang( '' );
				return $item_data;
			}
		}

		if ( ! $item_lang && ! $this->is_translated_item() ) {
			$this->set_lang( '' );
		} else {
			$this->set_lang( $item_lang ? $item_lang : pll_default_language() );
		}

		// Set unique term slug before updating or creating the term as Polylang naturally does itself.
		if ( 'terms' === magic_ex_im_get_magic_type() && $this->get_lang() ) {

			if ( isset( $item_data['term']['slug'] ) && $this->get_term_slug_separator() ) {
				$item_data['term']['slug'] .= $this->get_term_slug_separator() . $this->get_lang();
			}
		}

		return $item_data;
	}

	/**
	 * Processes filter hooks 'magic_im_existing_item_query_args'.
	 *
	 * @param array $args Query arguments.
	 * @return array
	 *
	 * @see Magic_EX_IM_Adapter_Polylang::magic_im_item_data_before_adjust()
	 */
	public function magic_im_existing_item_query_args( $args ) {

		// Prevent quering by default language.
		$args['lang'] ??= $this->get_lang();

		if ( 'terms' === magic_ex_im_get_magic_type() ) {

			// Set original term slug to query for an existing item.
			if ( isset( $args['slug'] ) && $this->get_term_slug_separator() ) {
				$args['slug'] = explode( $this->get_term_slug_separator(), $args['slug'] )[0];
			}
		}

		return $args;
	}

	/**
	 * Processes filter hook 'magic_im_post_id_by_name_args'.
	 *
	 * @param int $args Query arguments.
	 * @return array
	 */
	public function magic_im_post_id_by_name_args( $args ) {

		// Prevent quering by default language.
		$args['lang'] ??= $this->get_lang();

		return $args;
	}

	/**
	 * Processes action hook 'magic_im_item_id'.
	 *
	 * @param array $item_data Importing item data grouped by keys type.
	 * @return array
	 *
	 * @see PLL_Share_Post_Slug::wp_unique_post_slug()
	 */
	public function magic_im_item_id( $item_data ) {

		if ( ! $this->get_lang() ) {
			return;
		}

		$item_id = magic_ex_im_get_item_id();

		if ( 'posts' === magic_ex_im_get_magic_type() ) {

			pll_set_post_language( $item_id, $this->get_lang() );

			// Update post after language is set to be sure provided slug remains even if it's not unique.
			wp_update_post(
				array(
					'ID'        => $item_id,
					'post_name' => $item_data['post']['post_name'],
				)
			);

			// Save post translations.
			if ( isset( $item_data[ self::KEYS_GROUP ]['post_translations'] )
				&& is_array( $item_data[ self::KEYS_GROUP ]['post_translations'] ) ) {

				$post_translations = array();

				// Convert post names into IDs skipping non-installed languages.
				foreach ( $item_data[ self::KEYS_GROUP ]['post_translations'] as $lang_code => $post_name ) {

					if ( $lang_code === $this->get_lang() ) {
						$post_translations[ $lang_code ] = $item_id;
						continue;
					}

					if ( ! $this->is_installed_lang( $lang_code ) ) {
						continue;
					}

					$post_translations[ $lang_code ] = magic_im_post_id_by_name(
						$post_name,
						magic_ex_im_get_magic_item(),
						array( 'lang' => $lang_code )
					);
				}

				$post_translations = array_filter( $post_translations );

				if ( count( $post_translations ) > 1 ) {
					$post_translations = array_merge( pll_get_post_translations( $item_id ), $post_translations );
					ksort( $post_translations ); // Keep same order for each item to avoid meta duplications.
					pll_save_post_translations( $post_translations );
				}
			}
		}

		if ( 'terms' === magic_ex_im_get_magic_type() ) {

			pll_set_term_language( $item_id, $this->get_lang() );

			// Save term translations.
			if ( isset( $item_data[ self::KEYS_GROUP ]['term_translations'] )
				&& is_array( $item_data[ self::KEYS_GROUP ]['term_translations'] ) ) {

				$term_translations = array();

				// Convert term slugs into IDs.
				foreach ( $item_data[ self::KEYS_GROUP ]['term_translations'] as $lang_code => $term_slug ) {

					if ( $lang_code === $this->get_lang() ) {
						$term_translations[ $lang_code ] = $item_id;
						continue;
					}

					$term_id = get_term_by( 'slug', $term_slug, magic_ex_im_get_magic_item() )->term_id ?? '';

					if ( $term_id && $this->is_installed_lang( $lang_code ) ) {
						$term_translations[ $lang_code ] = pll_get_term( $term_id, $lang_code );
					}
				}

				$term_translations = array_filter( $term_translations );

				if ( count( $term_translations ) > 1 ) {
					$term_translations = array_merge( pll_get_term_translations( $item_id ), $term_translations );
					ksort( $term_translations ); // Keep same order for each item to avoid meta duplications.
					pll_save_term_translations( $term_translations );
				}
			}
		}
	}

	/**
	 * Processes filter hooks 'magic_ex_item_data_taxonomy', 'magic_im_item_data_taxonomy'.
	 *
	 * @param int $term_id Term ID.
	 * @param int $tax_name Taxonomy name.
	 * @return int|false|array Translated term ID, false if translation not found.
	 */
	public function translate_term( $term_id, $tax_name ) {

		if ( pll_is_translated_taxonomy( $tax_name ) && is_numeric( $term_id ) && $this->get_lang() ) {
			return pll_get_term( $term_id, $this->get_lang() );
		}

		return $term_id;
	}
}

return new Magic_EX_IM_Adapter_Polylang();
