<?php
/**
 * Class Magic_EX_IM_Adapter_Woocommerce.
 *
 * @package Magic_Export_Import\Plugin_Adapter
 *
 * phpcs:disable WordPress.DB.DirectDatabaseQuery, 
 * phpcs:disable WordPress.DB.SlowDBQuery
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to provide 'WooCommerce' plugin support.
 */
class Magic_EX_IM_Adapter_Woocommerce {

	const KEYS_GROUP                = 'woocommerce';
	const CUSTOM_TABLES_FEATURE_KEY = 'woocommerce_custom_orders_table_enabled';
	const ORDER_ITEM_TYPES          = array( 'line_item', 'fee', 'shipping', 'coupon', 'tax' );
	const ORDER_ITEM_GROUP          = 'order_item_meta';

	/**
	 * Constructor.
	 */
	public function __construct() {

		if ( class_exists( 'woocommerce' ) ) {
			add_action( 'magic_ex_im_init', array( $this, 'magic_ex_im_init' ) );
		}
	}

	/**
	 * Processes action hook 'magic_ex_im_init'.
	 */
	public function magic_ex_im_init() {

			// Export / Import.
			add_filter( 'magic_ex_im_get_allowed_magic_items', array( $this, 'magic_ex_im_get_allowed_magic_items' ) );
			add_filter( 'magic_ex_im_get_key_group', array( $this, 'magic_ex_im_get_key_group' ), 10, 2 );
			add_filter( 'magic_ex_im_get_mandatory_keys', array( $this, 'magic_ex_im_get_mandatory_keys' ) );

			// Export.
			add_filter( 'magic_ex_file_params', array( $this, 'magic_ex_file_params' ) );
			add_filter( 'magic_ex_im_processing_post_types', array( $this, 'magic_ex_im_processing_post_types' ) );
			add_filter( 'magic_ex_keys', array( $this, 'magic_ex_keys' ) );
			add_filter( 'magic_ex_media_keys_options', array( $this, 'magic_ex_media_keys_options' ) );
			add_filter( 'magic_ex_media_keys', array( $this, 'magic_ex_media_keys' ) );
			add_filter( 'magic_ex_get_items', array( $this, 'magic_ex_get_items' ) );
			add_filter( 'magic_ex_get_key_data', array( $this, 'magic_ex_get_key_data' ), 10, 3 );
			add_filter( 'magic_ex_maybe_adjust_value', array( $this, 'magic_ex_maybe_adjust_value' ), 10, 2 );

			// Import.
			add_filter( 'magic_im_validate_file_params', array( $this, 'magic_im_validate_file_params' ), 10, 2 );
			add_filter( 'magic_im_validate_item_data', array( $this, 'magic_im_validate_item_data' ), 10, 2 );
			add_filter( 'magic_im_maybe_adjust_value', array( $this, 'magic_im_maybe_adjust_value' ), 10, 2 );
			add_filter( 'magic_im_item_data_before_update', array( $this, 'magic_im_item_data_before_update' ) );
	}

	/**
	 * Checks if the current magic is related to shop order.
	 *
	 * @return bool
	 */
	protected function is_shop_order() {
		return 'posts:shop_order' === magic_ex_im_get_magic() || 'shop_orders' === magic_ex_im_get_magic_type();
	}

	/**
	 * Processes filter hook 'magic_ex_file_params'.
	 *
	 * @param array $file_params Exporting file parameters.
	 * @return array
	 */
	public function magic_ex_file_params( $file_params ) {

		if ( $this->is_shop_order() ) {
			$file_params[ self::CUSTOM_TABLES_FEATURE_KEY ] = get_option( self::CUSTOM_TABLES_FEATURE_KEY );
		}

		return $file_params;
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_allowed_magic_items'.
	 *
	 * @param array $allowed_magic_items Allowed magic items for the current magic type.
	 * @return array
	 */
	public function magic_ex_im_get_allowed_magic_items( $allowed_magic_items ) {

		if ( 'posts' === magic_ex_im_get_magic_type() ) {

			// Exclude secondary post types, which are meant to be processed together with other post types.
			$post_types_to_exlude = array( 'product_variation', 'shop_order_refund' );

			// Exclude shop order if new storing feature within custom tables is enabled.
			if ( wc_string_to_bool( get_option( self::CUSTOM_TABLES_FEATURE_KEY ) ) ) {
				$post_types_to_exlude[] = 'shop_order';
			}

			$allowed_magic_items = array_diff_key( $allowed_magic_items, array_flip( $post_types_to_exlude ) );
		}

		return $allowed_magic_items;
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_mandatory_keys'.
	 *
	 * @param array $mandatory_keys Mandatory keys.
	 * @return array
	 */
	public function magic_ex_im_get_mandatory_keys( $mandatory_keys ) {

		switch ( magic_ex_im_get_magic() ) {

			case 'posts:shop_order':
				$mandatory_keys[] = 'post_status';
				break;

			case 'posts:product':
				$mandatory_keys[] = 'product_type';
				break;
		}

		return $mandatory_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_key_group'.
	 *
	 * @param string $group Current key group.
	 * @param string $key The key.
	 * @return string
	 */
	public function magic_ex_im_get_key_group( $group, $key ) {

		if ( $this->is_shop_order() && ( 'order_items' === $key || 'order_notes' === $key ) ) {
			$group = self::KEYS_GROUP;
		}

		return $group;
	}

	/**
	 * Processes filter hook 'magic_ex_get_items'.
	 *
	 * @param array $item_ids Exporting item IDs.
	 * @return array
	 */
	public function magic_ex_get_items( $item_ids ) {

		// Add related shop order refunds to the exporting shop orders.
		if ( 'posts:shop_order' === magic_ex_im_get_magic() ) {

			$refund_item_ids = array();

			foreach ( $item_ids as $item_id ) {

				if ( 'wc-refunded' === get_post_status( $item_id ) ) {

					// Get related refund post ID.
					$refund_item_id = get_children(
						array(
							'post_parent' => $item_id,
							'post_type'   => 'shop_order_refund',
							'numberposts' => 1,
							'fields'      => 'ids',
						),
						ARRAY_N
					);

					$refund_item_ids[] = reset( $refund_item_id );
				}
			}

			$item_ids = array_merge( $item_ids, array_filter( $refund_item_ids ) );
		}

		// Add related product variations to the exporting products.
		if ( 'posts:product' === magic_ex_im_get_magic() ) {

			$product_variation_ids = array();

			foreach ( $item_ids as $item_id ) {
				$wc_product = wc_get_product( $item_id );

				if ( $wc_product && $wc_product->is_type( 'variable' ) ) {
					array_push( $product_variation_ids, ...$wc_product->get_children() );
				}
			}

			$item_ids = array_merge( $item_ids, array_unique( $product_variation_ids ) );
		}

		return $item_ids;
	}

	/**
	 * Processes filter hook 'magic_ex_im_processing_post_types'.
	 *
	 * @param array $post_types Processing post types for export or import.
	 * @return array
	 */
	public function magic_ex_im_processing_post_types( $post_types ) {

		if ( in_array( 'shop_order', $post_types, true ) ) {
			$post_types[] = 'shop_order_refund';
		}

		if ( in_array( 'product', $post_types, true ) ) {
			$post_types[] = 'product_variation';
		}

		return $post_types;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_data';
	 *
	 * @param mixed  $key_data Current key data.
	 * @param string $key Key.
	 * @param string $group Group.
	 * @return mixed
	 */
	public function magic_ex_get_key_data( $key_data, $key, $group ) {

		if ( self::KEYS_GROUP !== $group || ! $this->is_shop_order() ) {
			return $key_data;
		}

		$item_id = magic_ex_im_get_item_id();

		// Set order items.
		if ( 'order_items' === $key ) {

			global $wpdb;

			$wc_order = wc_get_order( $item_id );

			if ( $wc_order ) {
				$key_data = array();

				foreach ( $wc_order->get_items( self::ORDER_ITEM_TYPES ) as $order_item_id => $order_item ) {
					$item_meta = array();

					// Query all meta related to the current order item.
					$results = $wpdb->get_results(
						$wpdb->prepare(
							"SELECT meta_key, meta_value
							FROM {$wpdb->prefix}woocommerce_order_itemmeta
							WHERE order_item_id = %d",
							$order_item_id
						),
						ARRAY_A
					);

					// Convert database results to an associative array adjusting data for import.
					foreach ( $results as $meta ) {

						$meta_key   = $meta['meta_key'];
						$meta_value = $meta['meta_value'];

						switch ( $meta_key ) {

							// Convert product ID into product name.
							case '_product_id':
							case '_variation_id':
								$meta_value = get_post_field( 'post_name', $meta_value );
								break;

							// Convert refunded order item ID into name with prefixed type.
							case '_refunded_item_id':
								$meta_value = $order_item->get_type() . ':' . $order_item->get_name();
								break;

							// Convert media files IDs into URLs.
							default:
								$meta_value = magic_ex_im_adjust_media_value( $meta_value, $meta_key, self::ORDER_ITEM_GROUP );
								break;
						}

						$item_meta[ $meta_key ] = $meta_value;
					}

					$key_data[] = array(
						'order_item_type' => $order_item->get_type(),
						'order_item_name' => $order_item->get_name(),
						'order_itemmeta'  => $item_meta,
					);
				}
			}
		}

		// Set order notes.
		if ( 'order_notes' === $key ) {

			global $wpdb;

			$order_notes = $wpdb->get_results(
				$wpdb->prepare(
					"SELECT *
					FROM {$wpdb->prefix}comments
					WHERE comment_post_ID = %d
					AND comment_type = 'order_note'",
					$item_id
				),
				ARRAY_A
			);

			// Adjust order notes for export.
			foreach ( $order_notes as &$order_note ) {
				unset( $order_note['comment_ID'] );
				unset( $order_note['comment_post_ID'] );
			}

			$key_data = $order_notes;
		}

		return $key_data;
	}

	/**
	 * Processes filter hook 'magic_ex_keys'.
	 *
	 * @param array $export_keys Item grouped keys for export.
	 * @return array
	 */
	public function magic_ex_keys( $export_keys ) {

		// Exclude some WooCommerce keys by default.
		if ( isset( $export_keys['meta'] ) ) {

			$ex_meta_keys_to_exclude = array_filter(
				$export_keys['meta'],
				function ( $meta_key ) {
					return str_starts_with( $meta_key, '_woocommerce_persistent_cart_' )
						|| str_starts_with( $meta_key, '_maybe_used_by_' );
				}
			);

			$ex_meta_keys_to_exclude = array_merge(
				$ex_meta_keys_to_exclude,
				array( 'wc_last_active', '_wc_rating_count', '_wc_review_count', '_wc_average_rating' )
			);

			$export_keys['meta'] = array_diff( $export_keys['meta'], $ex_meta_keys_to_exclude );
		}

		// Add shop order keys.
		if ( $this->is_shop_order() ) {
			$export_keys[ self::KEYS_GROUP ][] = 'order_items';
			$export_keys[ self::KEYS_GROUP ][] = 'order_notes';
		}

		return $export_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_media_keys_options'.
	 *
	 * @param array $media_keys_options Meta keys to be treated as media files IDs holders.
	 * @return array
	 */
	public function magic_ex_media_keys_options( $media_keys_options ) {

		// Add shop order items meta keys.
		if ( $this->is_shop_order() ) {

			global $wpdb;

			$order_items_meta_keys = $wpdb->get_col(
				"SELECT DISTINCT meta_key
				FROM {$wpdb->prefix}woocommerce_order_itemmeta"
			);

			if ( $order_items_meta_keys ) {
				$media_keys_options[ self::ORDER_ITEM_GROUP ] = $order_items_meta_keys;
			}
		}

		return $media_keys_options;
	}

	/**
	 * Processes filter hook 'magic_ex_media_keys'.
	 *
	 * @param array $media_keys Meta keys to be treated as media files holders.
	 * @return array
	 */
	public function magic_ex_media_keys( $media_keys ) {
		$media_keys[] = '_product_image_gallery';
		return $media_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_maybe_adjust_value'.
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Value key.
	 *
	 * @return mixed
	 */
	public function magic_ex_maybe_adjust_value( $value, $key ) {

		if ( ! $value ) {
			return $value;
		}

		// Convert user ID into user email (shop_coupon).
		if ( '_used_by' === $key ) {
			$value = magic_ex_im_adjust_user_value( $value, 'email' );
		}

		// Convert product gallery image IDs into URLs.
		if ( '_product_image_gallery' === $key ) {

			$tmp_value = array_map( 'trim', explode( ',', $value ) );
			$tmp_value = array_map( 'wp_get_attachment_url', $tmp_value );
			$value     = implode( ',', array_filter( $tmp_value ) );
		}

		// Convert comma separated product IDs into names.
		if ( in_array( $key, array( 'product_ids', 'exclude_product_ids' ), true ) && is_string( $value ) ) {
			$tmp_value = array_map( 'intval', explode( ',', $value ) );

			foreach ( $tmp_value as $index => $product_id ) {
				$tmp_value[ $index ] = get_post_field( 'post_name', $product_id );
			}

			$value = implode( ',', array_filter( $tmp_value ) );
		}

		// Convert product IDs into names.
		if ( in_array( $key, array( '_upsell_ids', '_crosssell_ids', '_wc_url_coupons_product_ids' ), true )
			&& is_array( $value ) ) {

			$value = array_map(
				function ( $product_id ) {
					return get_post_field( 'post_name', $product_id );
				},
				$value
			);

			$value = array_filter( $value );
		}

		// Convert product categories IDs into slugs.
		if ( in_array( $key, array( 'product_categories', 'exclude_product_categories' ), true ) && is_array( $value ) ) {

			$value = array_map(
				function ( $term_id ) {
					return get_term_field( 'slug', $term_id, 'product_cat' );
				},
				$value
			);

			$value = array_filter( $value );
		}

		// Convert user ID into login.
		if ( '_customer_user' === $key || '_refunded_by' === $key ) {
			$value = magic_ex_im_adjust_user_value( $value );
		}

		return $value;
	}

	/**
	 * Processes filter hook 'magic_im_validate_file_params'.
	 *
	 * @param string[] $file_content_errors Existing validation errors.
	 * @param array    $file_params Import file parameters.
	 * @return string[]
	 */
	public function magic_im_validate_file_params( $file_content_errors, $file_params ) {

		// Verify WooCommerce HPOS (custom order tables) status matches between import file and current site.
		if ( $this->is_shop_order() ) {

			$feature_label = __( 'High-Performance order storage', 'woocommerce' );

			if ( ! isset( $file_params [ self::CUSTOM_TABLES_FEATURE_KEY ] ) ) {

				$file_content_errors[] = sprintf(
					'Missing <code>%s</code> feature detection in the import file.',
					$feature_label
				);

				return $file_content_errors;
			}

			$file_param_feature_enabled = wc_string_to_bool( $file_params [ self::CUSTOM_TABLES_FEATURE_KEY ] );
			$theme_feature_enabled      = wc_string_to_bool( get_option( self::CUSTOM_TABLES_FEATURE_KEY ) );

			if ( ! $file_param_feature_enabled && $theme_feature_enabled ) {

				$file_content_errors[] = sprintf(
					'<code>%s</code> is enabled in the current theme, but not in the import file.',
					$feature_label
				);

			} elseif ( $file_param_feature_enabled && ! $theme_feature_enabled ) {

				$file_content_errors[] = sprintf(
					'<code>%s</code> is enabled in the import file, but not in the current theme.',
					$feature_label
				);
			}
		}

		return $file_content_errors;
	}

	/**
	 * Processes filter hook 'magic_im_validate_item_data'.
	 *
	 * @param string[] $errors Existing validation errors.
	 * @param array    $item_data Importing item data.
	 *
	 * @return string
	 */
	public function magic_im_validate_item_data( $errors, $item_data ) {

		// Check product type is valid.
		if ( 'posts:product' === magic_ex_im_get_magic() ) {

			$product_type_key = magic_ex_im_build_key( 'product_type', 'taxonomy' );
			$product_type     = $item_data[ $product_type_key ] ?? '';

			if ( ! $product_type || ! in_array( $product_type, array_keys( wc_get_product_types() ), true ) ) {

				$errors[] = sprintf(
					'Non-existing <b>product type</b> <code>%s</code>.',
					$product_type ? $product_type : 'undefined'
				);
			}
		}

		return $errors;
	}

	/**
	 * Processes filter hook 'magic_im_maybe_adjust_value'.
	 *
	 * @param mixed  $value Value to adjust.
	 * @param string $key Value key.
	 *
	 * @return mixed
	 */
	public function magic_im_maybe_adjust_value( $value, $key ) {

		// Convert product image gallery URLs into IDs.
		if ( '_product_image_gallery' === $key && is_string( $value ) ) {
			$tmp_value = array_map( 'trim', explode( ',', $value ) );

			foreach ( $tmp_value as $index => $url ) {
				$tmp_value[ $index ] = magic_im_media_url_to_local_id( $url );
			}

			$value = implode( ',', array_filter( $tmp_value ) );
		}

		// Convert comma separated product names into IDs.
		if ( in_array( $key, array( 'product_ids', 'exclude_product_ids' ), true ) && is_string( $value ) ) {
			$tmp_value = array_map( 'trim', explode( ',', $value ) );

			foreach ( $tmp_value as $index => $product_name ) {
				$tmp_value[ $index ] = magic_im_post_id_by_name( $product_name, 'product' );
			}

			$value = implode( ',', array_filter( $tmp_value ) );
		}

		// Convert post names into IDs.
		if ( in_array( $key, array( '_upsell_ids', '_crosssell_ids', '_wc_url_coupons_product_ids' ), true )
			&& is_array( $value ) ) {

			foreach ( $value as $index => $product_name ) {
				$value[ $index ] = magic_im_post_id_by_name( $product_name, 'product' );
			}

			$value = array_filter( $value );
		}

		// Convert product categories slugs into IDs.
		if ( in_array( $key, array( 'product_categories', 'exclude_product_categories' ), true ) && is_array( $value ) ) {

			$value = array_map(
				function ( $term_slug ) {
					$term_id = get_term_by( 'slug', $term_slug, 'product_cat' )->term_id ?? '';
					return apply_filters( 'magic_im_item_data_taxonomy', $term_id, 'product_cat' );
				},
				$value
			);

			$value = array_filter( $value );
		}

		// Convert user login into ID.
		if ( '_customer_user' === $key || '_refunded_by' === $key ) {
			$value = magic_ex_im_adjust_user_value( $value );
		}

		return $value;
	}

	/**
	 * Processes filter hook 'magic_im_item_data_before_update'.
	 *
	 * @param array $item_data Processing item data grouped by keys.
	 * @return array
	 */
	public function magic_im_item_data_before_update( $item_data ) {

		$item_id = magic_ex_im_get_item_id();

		// Process shop order.
		if ( $this->is_shop_order() ) {

			global $wpdb;
			$found_refunded_order_item_ids = array();

			// Set order items.
			foreach ( $item_data[ self::KEYS_GROUP ]['order_items'] ?? array() as $order_item_data ) {

				// Insert order item.
				$wpdb->insert(
					$wpdb->prefix . 'woocommerce_order_items',
					array(
						'order_item_name' => $order_item_data['order_item_name'],
						'order_item_type' => $order_item_data['order_item_type'],
						'order_id'        => $item_id,
					),
					array( '%s', '%s', '%d' )
				);

				// Get the newly inserted order item ID.
				$inserted_order_item_id = $wpdb->insert_id;

				if ( ! $inserted_order_item_id ) {
					$error = sprintf( 'Error to create order item for the order with ID <code>%d</code>', intval( $item_id ) );
					magic_im_add_notice( $error );
					continue;
				}

				// Set order item meta.
				foreach ( $order_item_data['order_itemmeta'] ?? array() as $meta_key => $meta_value ) {

					switch ( $meta_key ) {

						// Convert product name into product ID.
						case '_product_id':
						case '_variation_id':
							$meta_value = magic_im_post_id_by_name( $meta_value, 'product' );
							break;

						// Convert order item name with prefixed type into order item ID.
						case '_refunded_item_id':
							// Get related refunded order.
							$parent_order_id = $item_data['post']['post_parent'] ?? $item_data['shop_order']['parent_order_id'] ?? '';
							$wc_order        = wc_get_order( $parent_order_id );

							if ( $wc_order && str_contains( $meta_value, ':' ) ) {

								list($item_type, $item_name) = explode( ':', $meta_value, 2 );

								// Find refunded order item by type and name, skip already found items.
								foreach ( $wc_order->get_items( self::ORDER_ITEM_TYPES ) as $order_item_id => $order_item ) {

									if ( $item_name === $order_item->get_name() && $item_type === $order_item->get_type()
										&& ! in_array( $order_item_id, $found_refunded_order_item_ids, true ) ) {

										$meta_value                      = $order_item_id;
										$found_refunded_order_item_ids[] = $order_item_id;
										break;
									}
								}
							} else {
								$meta_value = '';
							}
							break;

						// Convert media files URLs into IDs.
						default:
							$meta_value = magic_ex_im_adjust_media_value( $meta_value, $meta_key, self::ORDER_ITEM_GROUP );
							break;
					}

					// Update order item meta.
					$wpdb->insert(
						$wpdb->prefix . 'woocommerce_order_itemmeta',
						array(
							'order_item_id' => $inserted_order_item_id,
							'meta_key'      => $meta_key,
							'meta_value'    => $meta_value,
						),
						array( '%d', '%s', '%s' )
					);

					$inserted_itemmeta = $wpdb->insert_id;

					if ( ! $inserted_itemmeta ) {
						$error = sprintf( 'Error to create order itemmeta for the order with ID <code>%d</code>', intval( $item_id ) );
						magic_im_add_notice( $error );
					}
				}
			}

			// Set order notes.
			foreach ( $item_data[ self::KEYS_GROUP ]['order_notes'] ?? array() as $order_note ) {

				$order_note['comment_post_ID'] = $item_id;
				$order_note_id                 = wp_insert_comment( $order_note );

				if ( ! $order_note_id ) {
					$error = sprintf( 'Error to create order note for the order with ID <code>%d</code>', intval( $item_id ) );
					magic_im_add_notice( $error );
				}
			}
		}

		// Process product.
		if ( 'posts:product' === magic_ex_im_get_magic() ) {

			// Always set newly created product total sales meta.
			if ( magic_im_is_new_item() ) {
				$item_data['meta']['total_sales'] = intval( $item_data['meta']['total_sales'] ?? 0 );
			}

			// Set product attributes meta.

			$product_attributes = $item_data['meta']['_product_attributes'] ?? array();

			// Get product attributes meta from the the database.
			if ( ! $product_attributes && ! magic_im_is_new_item() ) {
				$product_attributes = get_post_meta( $item_id, '_product_attributes', true );
			}

			if ( ! is_array( $product_attributes ) ) {
				$product_attributes = array();
			}

			$is_variation = 'variable' === $item_data['taxonomy']['product_type'] ?? '';

			// Set attributes according to the existing attributes terms.
			foreach ( $item_data['taxonomy'] ?? array() as $tax_slug => $term_ids ) {

				if ( in_array( $tax_slug, wc_get_attribute_taxonomy_names(), true ) ) {
					if ( $term_ids && ! isset( $product_attributes[ $tax_slug ] ) ) {

						$product_attributes[ $tax_slug ] = array(
							'name'         => $tax_slug,
							'is_taxonomy'  => 1,
							'is_variation' => $is_variation,
						);
					}

					if ( ! $term_ids ) {
						unset( $product_attributes[ $tax_slug ] );
					}
				}
			}

			$item_data['meta']['_product_attributes'] = $product_attributes;
		}

		return $item_data;
	}
}

return new Magic_EX_IM_Adapter_Woocommerce();
