<?php
/**
 * Class Magic_EX_IM_Adapter_Yoast_SEO.
 *
 * @package Magic_Export_Import\Plugin_Adapter
 */

defined( 'ABSPATH' ) || exit;

/**
 * Singleton class to provide 'Yoast SEO' plugin support.
 */
class Magic_EX_IM_Adapter_Yoast_SEO {

	const KEYS_GROUP           = 'yoast-wpseo';
	const TAXONOMY_META_PREFIX = '_yoast_wpseo_primary_';

	/**
	 * Constructor.
	 */
	public function __construct() {

		if ( class_exists( 'WPSEO_Options' ) ) {
			add_action( 'magic_ex_im_init', array( $this, 'magic_ex_im_init' ) );
		}
	}

	/**
	 * Processes action hook 'magic_ex_im_init'.
	 */
	public function magic_ex_im_init() {

		// Export / Import.
		add_filter( 'magic_ex_im_get_key_group', array( $this, 'magic_ex_im_get_key_group' ), 10, 2 );

		// Export.
		add_filter( 'magic_ex_get_key_data', array( $this, 'magic_ex_get_key_data' ), 10, 3 );
		add_filter( 'magic_ex_keys', array( $this, 'magic_ex_keys' ) );
		add_filter( 'magic_ex_media_keys', array( $this, 'magic_ex_media_keys' ) );

		// Import.
		add_filter( 'magic_im_item_data_before_update', array( $this, 'magic_im_item_data_before_update' ) );
	}

	/**
	 * Gets all plugin related term keys.
	 *
	 * @return array
	 */
	private function get_term_keys() {

		if ( property_exists( 'WPSEO_Taxonomy_Meta', 'defaults_per_term' )
			&& is_array( WPSEO_Taxonomy_Meta::$defaults_per_term ) ) {

			$term_keys = array_keys( WPSEO_Taxonomy_Meta::$defaults_per_term );

			// Exclude 'wpseo_opengraph-image' key since it's just 'wpseo_opengraph-image-id' URL.
			$term_keys = array_diff( $term_keys, array( 'wpseo_opengraph-image' ) );
		}

		return $term_keys ?? array();
	}

	/**
	 * Processes filter hook 'magic_ex_im_get_key_group'.
	 *
	 * @param string $group Current key group.
	 * @param string $key The key.
	 * @return string
	 */
	public function magic_ex_im_get_key_group( $group, $key ) {

		if ( 'terms' === magic_ex_im_get_magic_type() && in_array( $key, $this->get_term_keys(), true ) ) {
			$group = self::KEYS_GROUP;
		}

		return $group;
	}

	/**
	 * Processes filter hook 'magic_ex_get_key_data';
	 *
	 * @param mixed  $key_data Current key data.
	 * @param string $key Data key.
	 * @param string $group Data key Group.
	 * @return mixed
	 */
	public function magic_ex_get_key_data( $key_data, $key, $group ) {

		// Get plugin related term data from the 'wp_options' table.
		if ( self::KEYS_GROUP === $group && 'terms' === magic_ex_im_get_magic_type()
			&& in_array( $key, $this->get_term_keys(), true ) ) {

			if ( is_callable( 'WPSEO_Taxonomy_Meta', 'get_term_meta' ) ) {

				$key_data = WPSEO_Taxonomy_Meta::get_term_meta(
					magic_ex_im_get_item_id(),
					magic_ex_im_get_magic_item(),
					str_replace( 'wpseo_', '', $key )
				);
			}
		}

		// Convert term ID into term slug.
		if ( 'meta' === $group && str_starts_with( $key, self::TAXONOMY_META_PREFIX ) ) {
			$term_id = $key_data;

			if ( $term_id ) {
				$tax_name = str_replace( self::TAXONOMY_META_PREFIX, '', $key );
				$term_id  = apply_filters( 'magic_ex_item_data_taxonomy', $term_id, $tax_name );

				if ( $term_id ) {
					$term_slug = get_term_field( 'slug', $term_id, $tax_name );
				}
			}

			$key_data = $term_slug ?? '';
		}

		return $key_data;
	}

	/**
	 * Processes filter hook 'magic_ex_keys'.
	 *
	 * @param array $export_keys Item grouped keys for export.
	 * @return array
	 */
	public function magic_ex_keys( $export_keys ) {

		$export_keys['meta'] = array_diff(
			$export_keys['meta'] ?? array(),
			array(
				'_yoast_wpseo_opengraph-image', // It's '_yoast_wpseo_opengraph-image-id' URL.
				'wp_yoast_notifications',
				'_yoast_wpseo_profile_updated',
			)
		);

		// Add terms keys stored in 'wp_options' table.
		if ( 'terms' === magic_ex_im_get_magic_type() ) {
			$export_keys[ self::KEYS_GROUP ] ??= array();
			$export_keys[ self::KEYS_GROUP ]   = array_merge( $export_keys[ self::KEYS_GROUP ], $this->get_term_keys() );
		}

		return $export_keys;
	}

	/**
	 * Processes filter hook 'magic_ex_media_keys'.
	 *
	 * @param array $media_keys Meta keys to be treated as media files holders.
	 * @return array
	 */
	public function magic_ex_media_keys( $media_keys ) {

		$media_keys[] = '_yoast_wpseo_opengraph-image-id';
		$media_keys[] = 'wpseo_opengraph-image-id';

		return $media_keys;
	}

	/**
	 * Processes filter hook 'magic_im_item_data_before_update'.
	 *
	 * @param array $item_data Importing item data grouped by type.
	 * @return array
	 */
	public function magic_im_item_data_before_update( $item_data ) {

		$item_id = magic_ex_im_get_item_id();

		// Set image URL to '_yoast_wpseo_opengraph-image' from '_yoast_wpseo_opengraph-image-id'.
		if ( isset( $item_data['meta']['_yoast_wpseo_opengraph-image-id'] ) ) {

			$image_id = $item_data['meta']['_yoast_wpseo_opengraph-image-id'];
			$item_data['meta']['_yoast_wpseo_opengraph-image'] = wp_get_attachment_url( $image_id );
		}

		// Convert post term slug into ID.
		foreach ( array_keys( $item_data['meta'] ?? array() ) as $meta_key ) {

			if ( str_starts_with( $meta_key, self::TAXONOMY_META_PREFIX ) ) {

				$term_slug = $item_data['meta'][ $meta_key ];
				$tax_name  = str_replace( self::TAXONOMY_META_PREFIX, '', $meta_key );
				$term_id   = get_term_by( 'slug', $term_slug, $tax_name )->term_id ?? '';

				$item_data['meta'][ $meta_key ] = apply_filters( 'magic_im_item_data_taxonomy', $term_id, $tax_name );
			}
		}

		// Save term data to the 'wp_options' table using plugin API.
		foreach ( $this->get_term_keys() as $term_key ) {

			if ( ! isset( $item_data[ self::KEYS_GROUP ][ $term_key ] ) || ! is_callable( 'WPSEO_Taxonomy_Meta', 'set_value' ) ) {
				continue;
			}

			WPSEO_Taxonomy_Meta::set_value(
				$item_id,
				magic_ex_im_get_magic_item(),
				str_replace( 'wpseo_', '', $term_key ),
				$item_data[ self::KEYS_GROUP ][ $term_key ]
			);

			// Set image URL to 'wpseo_opengraph-image' from 'wpseo_opengraph-image-id'.
			if ( 'wpseo_opengraph-image-id' === $term_key ) {

				WPSEO_Taxonomy_Meta::set_value(
					$item_id,
					magic_ex_im_get_magic_item(),
					'opengraph-image',
					wp_get_attachment_url( $item_data[ self::KEYS_GROUP ][ $term_key ] )
				);
			}
		}

		return $item_data;
	}
}

return new Magic_EX_IM_Adapter_Yoast_SEO();
